/*  RAxML-HPC, a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright March 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  stamatak@ics.forth.gr
 *
 *  When publishing work that is based on the results from RAxML-HPC please cite:
 *  
 *  Alexandros Stamatakis: "An Efficient Program for phylogenetic Inference Using Simulated Annealing". 
 *  Proceedings of IPDPS2005,  Denver, Colorado, April 2005.
 *  
 *  AND
 *
 *  Alexandros Stamatakis, Thomas Ludwig, and Harald Meier: "RAxML-III: A Fast Program for Maximum Likelihood-based Inference of Large Phylogenetic Trees". 
 *  In Bioinformatics 21(4):456-463.
 */


#define maxcategories 2

/*  Program constants and parameters  */

/*#define maxlogf       1024  maximum number of user trees */
#define smoothings      32  /* maximum smoothing passes through tree */
#define iterations      10  /* maximum iterations of iterations per insert */
#define newzpercycle     1  /* iterations of makenewz per tree traversal */
#define nmlngth         100  /* number of characters in species name */
#define deltaz     0.00001  /* test of net branch length change in update */
#define zmin       1.0E-15  /* max branch prop. to -log(zmin) (= 34) */
#define zmax (1.0 - 1.0E-6) /* min branch prop. to 1.0-zmax (= 1.0E-6) */
#define defaultz       0.9  /* value of z assigned as starting point */
#define unlikely  -1.0E300  /* low likelihood for initialization */

/*  These values are used to rescale the lilelihoods at a given site so that
 *  there is no floating point underflow.
 */
#define twotothe256  \
  115792089237316195423570985008687907853269984665640564039457584007913129639936.0
                                                     /*  2**256 (exactly)  */


#define minlikelihood  (1.0/twotothe256)


#define badEval        1.0
#define badZ           0.0
#define badRear         -1
/*#define badSigma      -1.0*/

#define TRUE             1
#define FALSE            0

#define treeNone         0
#define treeNewick       1
#define treeProlog       2
#define treePHYLIP       3
#define treeMaxType      3
#define treeDefType  treePHYLIP

#define ABS(x)    (((x)<0)   ?  (-(x)) : (x))
#define MIN(x,y)  (((x)<(y)) ?    (x)  : (y))
#define MAX(x,y)  (((x)>(y)) ?    (x)  : (y))
#define LOG(x)    (((x)>0)   ? log(x)  : hang("log domain error"))
#define NINT(x)   ((int) ((x)>0 ? ((x)+0.5) : ((x)-0.5)))

#define programName        "RAxML-HPC"
#define programVersion     "1.0"
#define programDate        "March 2005"

#define TREE_INFERENCE 1
#define TREE_EVALUATION 2
#define FAST_MODE 3
#define SIMULATED_ANNEALING 4
#define TREE_RECURSIVE 5
#define HEAVY_MODE 6
#define PRE_SELECT_MODE 7
#define SPR 8
#define BIG_MODE 9
#define BIG_FAST_MODE 10
#define BIG_RAPID_MODE 11
#define PARALLEL_MODE 12
#define BIG_EXPERIMENTAL 13
#define BIG_SIMULATED_ANNEALING 14
#define NEXUS_CONVERSION 15

#define M_HKY85 1
#define M_HKY85CAT 2
#define M_GTR 3
#define M_GTRCAT 4
#define M_PROTEINS 5
#define M_PAM 6
#define M_PMB 7
#define M_JTT 8
#define M_PROTEINSCAT 9
#define M_HKY85GAMMA 10
#define M_HKY85GAMMA2 11
#define M_GTRGAMMA 12
#define M_GTRMIX 13

#define REARRANGEMENT_SPECIFICATION 0

typedef  int boolean;

typedef struct bInf {
  double likelihood;
  int node;
} bestInfo;

typedef struct iL {
  bestInfo *list;
  int n;
  int valid;
} infoList;


typedef  struct  likelihood_vector 
{
  double      a, c, g, t;
  int         exp;
} 
  likelivector;

typedef  struct  gamma_likelihood_vector 
{
  double     a0, c0, g0, t0, a1, c1, g1, t1, a2, c2, g2, t2, a3, c3, g3, t3;
  int        exp;
} 
  gammalikelivector;

typedef struct pvv 
{
  int  parsimonyScore;
  char parsimonyState; 
} 
  parsimonyVector;




typedef struct ratec 
{
  double accumulatedSiteLikelihood;
  double rate;
} 
  rateCategorize;


typedef  struct noderec 
{
  double           z;
  struct noderec  *next;
  struct noderec  *back;
  int              number;
  void            *x;   
  char            *tip;           
} 
  node, *nodeptr;

typedef  struct 
{
  int              numsp;       /* number of species (also tr->mxtips) */
  int              sites;       /* number of input sequence positions */
  char          **y;           /* sequence data array */
  char            *y0;  
    
  double freqaREV[maxcategories+1], 
    freqcREV[maxcategories+1], 
    freqgREV[maxcategories+1], 
    freqtREV[maxcategories+1], 
    ttratioREV[maxcategories+1];
  boolean matread;
  double EIGV[maxcategories+1][4][4];
  double invfreq[maxcategories+1][4];
  double EIGN[maxcategories+1][4];
 
  double initialRates[5];
  double           freqa, freqc, freqg, freqt, ttratio;
  double          freqr, freqy, invfreqr, invfreqy,
                  freqar, freqcy, freqgr, freqty;
  double           xi, xv, fracchange; /* transition/transversion */
/* End of DNA specific values */
    int             *wgt;         /* weight per sequence pos */
    int             *wgt2;        /* weight per pos (booted) */ 
    } rawdata;

typedef  struct {
  int             *alias;       /* site representing a pattern */
  int             *reAlias;
  int             *aliaswgt;    /* weight by pattern */
  int             *rateCategory;
  int              endsite;     /* # of sequence patterns */
  int              wgtsum;      /* sum of weights of positions */
  double          *patrat;      /* rates per pattern */
  double          *patratStored;
  double          *wr;          /* weighted rate per pattern */
  double          *wr2;         /* weight*rate**2 per pattern */
} cruncheddata;


typedef struct rmsg {
  int removedNumberP;
  int removedNumberQ;
  int removedNumberR;
  int insertNumberQ;
  int insertNumberR;
  double currentZQR;
  double currentLZR;
  double currentLZQ;
  double currentLZS; 
} rearrangementDescriptor;

  
typedef struct {   
  double fxqn;
  double tempi_minus_fxqn;
  double tempi_minus_fxqn2;
  double a_minus_tempi;
  double g_minus_tempi;
  double c_minus_tempi2;
  double t_minus_tempi2;  
} checkdim;


typedef struct {

  double fxpa;
  double fxpg;
  double fxpc;
  double fxpt; 
  double fxpr;
  double fxpy;
  double fxpr_plus_fxpy;
} evaldim;


typedef struct {
  double a, c, g, t;
  double fx2r, fx2y, fx2r_plus_fx2y;
}  likelivectorAcc;

typedef struct {
  double a, c, g, t;  
}  likelivectorGTR;


typedef  struct  {    
  double           startLH;
  double           endLH;
  double           likelihood;   
  node           **nodep;
  node            *start;
  node            *outgrnode;
  int              mxtips;
  int              ntips;
  int              nextnode;
  int              opt_level;   
  int              outgr;       
  int              NumberOfCategories;
  int              parsimonyLength;
  boolean          prelabeled; 
  boolean          smoothed;
  boolean          rooted;
  rawdata         *rdta;        
  cruncheddata    *cdta;
  double          t_value;
  double          *modelData;
  int             modelParameters;
  double          alpha;
  int             gammaCATs;
  double          gammaRates[4]; /* TODO */
  double          gammaProbs[4]; /* TODO */

  char **nameList;
  
  int bestParsimony;
  double bestOfNode;
  nodeptr removeNode;
  nodeptr insertNode;
  double zqr;
  double currentZQR;

  double currentLZR;
  double currentLZQ;
  double currentLZS;
  double currentLZI;
  double lzs;
  double lzq;
  double lzr;
  double lzi;
  int nodeInfo;

  /***********BIG************/

  /********PARALLEL***********/
#ifdef PARALLEL_2
  int removeNumberP;
  int removeNumberQ;
  int removeNumberR;
  int removedNumberP;
  int removedNumberQ;
  int removedNumberR;
  int insertNumberQ;
  int insertNumberR;
  
  rearrangementDescriptor *rDesc;
#endif
  /*************************/

  double EI[12];
  double EV[16];
  likelivectorAcc tip[16];
  likelivectorGTR gtrTip[16];
  likelivector GTRGammaTip[16];
  checkdim spp[16];
  evaldim skk[16];

    } tree;

typedef struct conntyp {
    double           z;           /* branch length */
    node            *p, *q;       /* parent and child sectors */
    void            *valptr;      /* pointer to value of subtree */
    int              descend;     /* pointer to first connect of child */
    int              sibling;     /* next connect from same parent */
    } connect, *connptr;

typedef  struct {
    double           likelihood;
  int              number;
    connect         *links;       /* pointer to first connect (start) */
    node            *start;
    int              nextlink;    /* index of next available connect */
                                  /* tr->start = tpl->links->p */
    int              ntips;
    int              nextnode;
    int              opt_level;   /* degree of branch swapping explored */
    int              scrNum;      /* position in sorted list of scores */
    int              tplNum;      /* position in sorted list of trees */
    
    boolean          prelabeled;  /* the possible tip names are known */
    boolean          smoothed;    /* branch optimization converged? */
    } topol;

typedef struct {
    double           best;        /* highest score saved */
    double           worst;       /* lowest score saved */
    topol           *start;       /* starting tree for optimization */
    topol          **byScore;
    topol          **byTopol;
    int              nkeep;       /* maximum topologies to save */
    int              nvalid;      /* number of topologies saved */
    int              ninit;       /* number of topologies initialized */
    int              numtrees;    /* number of alternatives tested */
    boolean          improved;
    } bestlist;

typedef  struct {  
  double t0;
  int categories;
  int repeats; 
  int model;
  int protmodel;
  int              max_rearrange;
  int              stepwidth;
  int              initial;
  boolean          initialSet;
  int              mode;
  long             boot;        /* bootstrap random number seed */    
  boolean          interleaved; /* input data are in interleaved format */    
  boolean          restart;     /* resume addition to partial tree */  
  boolean          useWeightFile;  
  int checkpoints;
  int startingTreeOnly;
  int rapidParsimony;
  int useMixedModel;
  int multipleRuns;
  int parsimonySeed;
} analdef;

boolean (*newview) (tree *, nodeptr);
double (*evaluate)(tree *, nodeptr);   
double (*makenewz) (tree *, nodeptr, nodeptr, double, int);
double (*evaluatePartial)(tree *, nodeptr, int, double);
void    (*newviewPartial)(tree *, nodeptr, int, double, likelivector *);
boolean (*initravPartial) (tree *, nodeptr, int, double);


 void newviewPartialGTRCAT ( tree *tr, nodeptr p, int i, double ki, likelivector *x3 );
 double evaluatePartialGTRCAT ( tree *tr, nodeptr p, int i, double ki );
 boolean newviewGTRCAT ( tree *tr, nodeptr p );
 double evaluateGTRCAT ( tree *tr, nodeptr p );
 double makenewzGTRCAT ( tree *tr, nodeptr p, nodeptr q, double z0, int maxiter );
 void newviewPartialGTRCAT ( tree *tr, nodeptr p, int i, double ki, likelivector *x3 );
 double evaluatePartialGTRCAT ( tree *tr, nodeptr p, int i, double ki );
 boolean newviewGTRCAT ( tree *tr, nodeptr p );
 double evaluateGTRCAT ( tree *tr, nodeptr p );
 double makenewzGTRCAT ( tree *tr, nodeptr p, nodeptr q, double z0, int maxiter );
 boolean newviewGTRGAMMA ( tree *tr, nodeptr p );
 double evaluateGTRGAMMA ( tree *tr, nodeptr p );
 double makenewzGTRGAMMA ( tree *tr, nodeptr p, nodeptr q, double z0, int maxiter );
 double evaluateGTRGAMMA2 ( tree *tr, nodeptr p );
 boolean newviewGTRGAMMA ( tree *tr, nodeptr p );
 double evaluateGTRGAMMA ( tree *tr, nodeptr p );
 double makenewzGTRGAMMA ( tree *tr, nodeptr p, nodeptr q, double z0, int maxiter );
 double gettime ( void );
 int gettimeSrand ( void );
 void *tipValPtr ( nodeptr p );
 int cmpTipVal ( void *v1, void *v2 );
 topol *setupTopol ( int maxtips, int nsites );
 void freeTopol ( topol *tpl );
 int saveSubtree ( nodeptr p, topol *tpl );
 nodeptr minSubtreeTip ( nodeptr p0 );
 nodeptr minTreeTip ( nodeptr p );
 void saveTree ( tree *tr, topol *tpl );
 void copyTopol ( topol *tpl1, topol *tpl2 );
 boolean restoreTreeRecursive ( topol *tpl, tree *tr );
 boolean restoreTree ( topol *tpl, tree *tr );
 boolean restoreTopology ( topol *tpl, tree *tr );
 int initBestTree ( bestlist *bt, int newkeep, int numsp, int sites );
 int resetBestTree ( bestlist *bt );
 boolean freeBestTree ( bestlist *bt );
 int cmpSubtopol ( connptr p10, connptr p1, connptr p20, connptr p2 );
 int cmpTopol ( void *tpl1, void *tpl2 );
 int cmpTplScore ( void *tpl1, void *tpl2 );
 int findInList ( void *item, void *list[], int n, int (* cmpFunc)() );
 int findTreeInList ( bestlist *bt, tree *tr );
 int saveBestTree ( bestlist *bt, tree *tr );
 int recallBestTreeRecursive ( bestlist *bt, int rank, tree *tr );
 int recallBestTree ( bestlist *bt, int rank, tree *tr );
 int recallBestTopology ( bestlist *bt, int rank, tree *tr );
 void getnums ( rawdata *rdta );
 boolean digitchar ( int ch );
 boolean whitechar ( int ch );
 void uppercase ( int *chptr );
 int base36 ( int ch );
 int itobase36 ( int i );
 int findch ( int c );
 void getyspace ( rawdata *rdta );
 void freeyspace ( rawdata *rdta );
 boolean setupTree ( tree *tr, int nsites, analdef *adef );
 void freeTreeNode ( nodeptr p );
 void freeTree ( tree *tr );
 boolean getdata ( analdef *adef, rawdata *rdta, tree *tr );
 void inputweights ( analdef *adef, rawdata *rdta, cruncheddata *cdta );
 void getinput ( analdef *adef, rawdata *rdta, cruncheddata *cdta, tree *tr );
 double randum ( long *seed );
 void makeboot ( analdef *adef, rawdata *rdta, cruncheddata *cdta );
 void sitesort ( rawdata *rdta, cruncheddata *cdta );
 void sitecombcrunch ( rawdata *rdta, cruncheddata *cdta );
 boolean makeweights ( analdef *adef, rawdata *rdta, cruncheddata *cdta );
 boolean makevalues ( rawdata *rdta, cruncheddata *cdta );
 char *itobase36_str ( int i );
 boolean empiricalfreqs ( analdef *adef, rawdata *rdta, cruncheddata *cdta );
 void alterTT ( tree *tr, double t_value );
 void linkdata2tree ( rawdata *rdta, cruncheddata *cdta, tree *tr );
 void allocNodex ( tree *tr, analdef *adef );
 void freeNodex ( tree *tr );
 void *getxnode ( nodeptr p );
 boolean update ( tree *tr, nodeptr p );
 boolean smooth ( tree *tr, nodeptr p );
 boolean smoothTree ( tree *tr, int maxtimes );
 boolean localSmooth ( tree *tr, nodeptr p, int maxtimes );
 void hookup ( nodeptr p, nodeptr q, double z );
 boolean insert ( tree *tr, nodeptr p, nodeptr q, boolean glob );
 nodeptr removeNode ( tree *tr, nodeptr p );
 boolean initrav ( tree *tr, nodeptr p );
 boolean initravPoutsa ( tree *tr, nodeptr p );
 boolean initravDIST ( tree *tr, nodeptr p, int distance );
 boolean readKeyValue ( char *string, char *key, char *format, void *value );
 void printTree ( FILE *outf, tree *tr, int quoted );
 void printTreeNEXUS ( tree *tr, int quoted );
 void printTreeMrBayes ( tree *tr, int quoted );
 void printResult ( tree *tr, int quoted );
 void printLog ( tree *tr, analdef *adef );
 void initTreeString ( char *treestr, tree *tr, int form );
 char *treeString ( char *treestr, tree *tr, nodeptr p, int form );
 char *treeStringNEXUS ( char *treestr, tree *tr, nodeptr p, int form );
 void treeOut ( FILE *treefile, tree *tr, int form );
 int treeFinishCom ( FILE *fp, char **strp );
 int treeGetCh ( FILE *fp );
 boolean treeLabelEnd ( int ch );
 boolean treeGetLabel ( FILE *fp, char *lblPtr, int maxlen );
 boolean treeFlushLabel ( FILE *fp );
 int treeFindTipByLabel ( char *str, tree *tr );
 int treeFindTipName ( FILE *fp, tree *tr );
 void treeEchoContext ( FILE *fp1, FILE *fp2, int n );
 boolean treeProcessLength ( FILE *fp, double *dptr );
 int treeFlushLen ( FILE *fp );
 boolean treeNeedCh ( FILE *fp, int c1, char *where );
 boolean addElementLen ( FILE *fp, tree *tr, nodeptr p );
 int saveTreeCom ( char **comstrp );
 boolean processTreeCom ( FILE *fp, tree *tr );
 nodeptr uprootTree ( tree *tr, nodeptr p );
 boolean treeReadLen ( FILE *fp, tree *tr );
 int str_treeFinishCom ( char **treestrp, char **strp );
 int str_treeGetCh ( char **treestrp );
 boolean str_treeGetLabel ( char **treestrp, char *lblPtr, int maxlen );
 boolean str_treeFlushLabel ( char **treestrp );
 int str_treeFindTipName ( char **treestrp, tree *tr );
 boolean str_treeProcessLength ( char **treestrp, double *dptr );
 boolean str_treeFlushLen ( char **treestrp );
 boolean str_treeNeedCh ( char **treestrp, int c1, char *where );
 boolean str_processTreeCom ( tree *tr, char **treestrp );
 boolean str_processTreeComMerge ( int *ntaxa, char **treestrp );
 boolean str_addElementLen ( char **treestrp, tree *tr, nodeptr p );
 boolean str_treeReadLen ( char *treestr, tree *tr );
boolean treeEvaluate ( tree *tr, double smoothFactor );
 boolean evalFast ( tree *tr );
 double str_readTreeLikelihood ( char *treestr );
 void initAdef ( analdef *adef );
 void printVersionInfo ( void );
 void printREADME ( void );
 void get_args ( int argc, char *argv[], boolean print_usage, analdef *adef, rawdata *rdta );
 boolean str_treeReadLenSpecial ( char *treestr, tree *tr );
 void optimizeAlpha ( tree *tr );
 void optimizeTT ( tree *tr );
 void baseFrequenciesGTR ( rawdata *rdta, cruncheddata *cdta );
 void initReversibleGTR ( rawdata *rdta, cruncheddata *cdta, tree *tr );
 void alterRates ( tree *tr, int k, analdef *adef );
 void optimizeRates ( tree *tr, analdef *adef );
 void printModelAndProgramInfo ( tree *tr, analdef *adef, int argc, char *argv[] );
 void sortInfoList ( void );
 void resetInfoList ( void );
 void initInfoList ( int n );
 void freeInfoList ( void );
 void insertInfoList ( int node, double likelihood );
 void printInfoList ( void );
 boolean smoothRegion ( tree *tr, nodeptr p, int region );
 boolean regionalSmooth ( tree *tr, nodeptr p, int maxtimes, int region );
 nodeptr removeNodeBIG ( tree *tr, nodeptr p );
 nodeptr removeNodeRestoreBIG ( tree *tr, nodeptr p );
 boolean insertBIG ( tree *tr, nodeptr p, nodeptr q, boolean glob );
 boolean insertRestoreBIG ( tree *tr, nodeptr p, nodeptr q, boolean glob );
 void restoreTopologyOnly ( tree *tr );
 boolean testInsertBIG ( tree *tr, nodeptr p, nodeptr q );
 void addTraverseBIG ( tree *tr, nodeptr p, nodeptr q, int mintrav, int maxtrav );
 int rearrangeBIGEXP ( tree *tr, nodeptr p, int mintrav, int maxtrav );
 int rearrangeBIG ( tree *tr, nodeptr p, int mintrav, int maxtrav );
 boolean testInsertRestoreBIG ( tree *tr, nodeptr p, nodeptr q );
 void restoreTreeFast ( tree *tr );
 void restoreTreeFastEXP ( tree *tr );
 int treeOptimizeBIG ( tree *tr, int _mintrav, int _maxtrav, analdef *adef );
 void computeBIG ( tree *tr, analdef *adef );
 int determineRearrangementSetting ( tree *tr, analdef *adef, bestlist *bestT );
 int determineRearrangementSettingEXP ( tree *tr, analdef *adef, bestlist *bestT );
 double treeOptimizeQuick ( tree *tr, int mintrav, int maxtrav, analdef *adef );
 void computeBIGFAST ( tree *tr, analdef *adef );
 double treeOptimizeRapid ( tree *tr, int mintrav, int maxtrav, analdef *adef );
 void computeBIGRAPID ( tree *tr, analdef *adef );
 void computeBIGEXP ( tree *tr, analdef *adef );
 void computeBIGRAPIDFAST ( tree *tr, analdef *adef );
 void simulatedAnnealingRapid ( tree *tr, analdef *adef );
 double calcPartialLikelihood ( tree *tr, int i, double ki );
 void categorize ( tree *tr, rateCategorize *rc );
 void optimizeRateCategories ( tree *tr, int categorized, int _maxCategories );
 int optimizeModel ( tree *tr, analdef *adef, int finalOptimization );
 void consensusTree ( bestlist *consensusList, tree *tr );
 void treeEvaluation ( tree *tr, analdef *adef );
 double LnGamma ( double alpha );
 double IncompleteGamma ( double x, double alpha, double ln_gamma_alpha );
 double PointNormal ( double prob );
 double PointChi2 ( double prob, double v );
 void makeGammaCats ( tree *tr );
 void initModel ( tree *tr, rawdata *rdta, cruncheddata *cdta, analdef *adef );
 void initModel2 ( tree *tr, rawdata *rdta, cruncheddata *cdta, analdef *adef );
 int filexists ( char *filename );
 void givens ( double (*a)[20], long i, long j, long n, double ctheta, double stheta, boolean left );
 void coeffs ( double x, double y, double *c, double *s, double accuracy );
 void tridiag ( double (*a)[20], long n, double accuracy );
 void shiftqr ( double (*a)[20], long n, double accuracy );
 void qreigen ( double (*prob)[20], long n );
 void printGarli ( tree *tr );
 void generateScripts ( void );
 void initGlobalTip ( tree *tr, analdef *adef );
 void printWeigths ( tree *tr );
 int main ( int argc, char *argv[] );
 void printMultipleTree ( tree *tr, int quoted, char *fileName );
 void printBootstrap ( tree *tr, int quoted, char *fileName );
 void printLogMultiple ( tree *tr, analdef *adef, char *checkName, char *logName );
 void computeBIGRAPIDMULT ( tree *tr, analdef *adef, int printLogFile, char *checkName, char *logName );
 boolean regetdata ( analdef *adef, rawdata *rdta, tree *tr );
 void reMakeValuesAndWeigths ( tree *tr, analdef *adef );
 void doPapa ( void );
 void doBootstrap ( tree *tr, analdef *adef, rawdata *rdta, cruncheddata *cdta, int argc, char *argv[] );
 void doMultiple ( tree *tr, analdef *adef, rawdata *rdta, cruncheddata *cdta, int argc, char *argv[] );
 void printMultipleTree ( tree *tr, int quoted, char *fileName );
 void printBootstrapParallel ( int quoted, char *fileName, char *str );
 void printLogMultiple ( tree *tr, analdef *adef, char *checkName, char *logName );
 void computeBIGRAPIDMULT ( tree *tr, analdef *adef, int printLogFile, char *checkName, char *logName );
 boolean regetdata ( analdef *adef, rawdata *rdta, tree *tr );
 void reMakeValuesAndWeigths ( tree *tr, analdef *adef );
 void doPapa ( void );
 void doBootstrap ( tree *tr, analdef *adef, rawdata *rdta, cruncheddata *cdta, int argc, char *argv[] );
 void doMultiple ( tree *tr, analdef *adef, rawdata *rdta, cruncheddata *cdta, int argc, char *argv[] );
 double evaluateRate ( tree *tr, int i, double rate, analdef *adef );
 double evaluateAlpha ( tree *tr, double alpha );
 int brakAlpha ( double *param, double *ax, double *bx, double *cx, double *fa, double *fb, double *fc, double lim_inf, double lim_sup, tree *tr );
 double brentAlpha ( double fb, double *param, double ax, double bx, double cx, double tol, double *xmin, tree *tr, int n_iter_max );
 int brakRates ( double *param, double *ax, double *bx, double *cx, double *fa, double *fb, double *fc, double lim_inf, double lim_sup, tree *tr, int i, analdef *adef );
 double brentRates ( double fb, double *param, double ax, double bx, double cx, double tol, double *xmin, tree *tr, int n_iter_max, int i, analdef *adef );
 void optAlpha ( tree *tr, analdef *adef, double modelEpsilon );
 void optRates ( tree *tr, analdef *adef, double modelEpsilon );
 void modOpt ( tree *tr, analdef *adef );
 void newviewParsimony ( tree *tr, nodeptr p );
 int evaluateParsimony ( tree *tr, nodeptr p );
 void initravParsimony ( tree *tr, nodeptr p );
 void insertParsimony ( tree *tr, nodeptr p, nodeptr q );
 nodeptr buildNewTip ( tree *tr, nodeptr p );
 boolean buildSimpleTree ( tree *tr, int ip, int iq, int ir );
 void testInsertParsimony ( tree *tr, nodeptr p, nodeptr q );
 void restoreTreeParsimony ( tree *tr, nodeptr p, nodeptr q );
 void addTraverseParsimony ( tree *tr, nodeptr p, nodeptr q, int mintrav, int maxtrav );
 nodeptr findAnyTip ( nodeptr p );
 int randomInt ( int n );
 void makePermutation ( int *perm, int n, analdef *adef );
 void initravDISTParsimony ( tree *tr, nodeptr p, int distance );
 nodeptr removeNodeParsimony ( tree *tr, nodeptr p );
 void rearrangeParsimony ( tree *tr, nodeptr p, int mintrav, int maxtrav );
 void restoreTreeRearrangeParsimony ( tree *tr );
 void allocNodexParsimony ( tree *tr );
 void freeNodexParsimony ( tree *tr );
 void restore ( tree *tr, int *alias );
 void sortInformativeSites ( tree *tr, int *informative, int *alias );
 void determineUninformativeSites ( tree *tr, int *informative, int *alias );
 void makeParsimonyTree ( tree *tr, analdef *adef );
 void tred2 ( double *a, const int n, const int np, double *d, double *e );
 double pythag ( double a, double b );
 void tqli ( double *d, double *e, int n, int np, double *z );
