/*  RAxML-HPC, a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright March 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  stamatak@ics.forth.gr
 *
 *  When publishing work that is based on the results from RAxML-HPC please cite:
 *  
 *  Alexandros Stamatakis: "An Efficient Program for phylogenetic Inference Using Simulated Annealing". 
 *  Proceedings of IPDPS2005,  Denver, Colorado, April 2005.
 *  
 *  AND
 *
 *  Alexandros Stamatakis, Thomas Ludwig, and Harald Meier: "RAxML-III: A Fast Program for Maximum Likelihood-based Inference of Large Phylogenetic Trees". 
 *  In Bioinformatics 21(4):456-463.
 */

#ifdef WIN32
   #include <sys/timeb.h>
   #include "getopt.h"
   #include <direct.h>
#else
   #include <sys/times.h>
   #include <sys/types.h>
   #include <sys/time.h>
   #include <unistd.h> 
   #include <sys/mman.h>
#endif

#include <math.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifdef PROCESSOR_AFFINITY
#include <numa.h>
#endif


#ifdef PARALLEL
#include <mpi.h>
#endif

#include "axml.h"
#include "globalVariables.h"

/*=======================================================================*/
/*                              PROGRAM                                  */
/*=======================================================================*/


double gettime()
{
#ifdef WIN32
   struct _timeb timebuffer;
   _ftime(&timebuffer);
   return timebuffer.time + timebuffer.millitm * 0.000001;
#else
   struct timeval ttime;
    gettimeofday(&ttime , NULL);
    return ttime.tv_sec + ttime.tv_usec * 0.000001;
#endif
}

int gettimeSrand()
{
#ifdef WIN32
   struct _timeb timebuffer;
   _ftime(&timebuffer);
   return timebuffer.time + timebuffer.millitm;
#else
   struct timeval ttime;
    gettimeofday(&ttime , NULL);
    return ttime.tv_sec + ttime.tv_usec;
#endif
}


void  *tipValPtr (nodeptr p)
  { return  (void *) & p->number;
    }


int  cmpTipVal (void *v1, void *v2)
  { /* cmpTipVal */
    int  i1, i2;

    i1 = *((int *) v1);
    i2 = *((int *) v2);
    return  (i1 < i2) ? -1 : ((i1 == i2) ? 0 : 1);
  } /* cmpTipVal */





/*  These are the only routines that need to UNDERSTAND topologies */

topol  *setupTopol (int maxtips, int nsites)
  { /* setupTopol */
    topol   *tpl;

    if (! (tpl = (topol *) malloc(sizeof(topol))) || 
        ! (tpl->links = (connptr) malloc((2*maxtips-3) * sizeof(connect))))
      {
	printf("ERROR: Unable to get topology memory");
	tpl = (topol *) NULL;
      }
    else 
      {
	tpl->likelihood  = unlikely;
	tpl->start       = (node *) NULL;
	tpl->nextlink    = 0;
	tpl->ntips       = 0;
	tpl->nextnode    = 0;
	tpl->opt_level   = 0;     /* degree of branch swapping explored */
	tpl->scrNum      = 0;     /* position in sorted list of scores */
	tpl->tplNum      = 0;     /* position in sorted list of trees */
	tpl->prelabeled  = TRUE;
	tpl->smoothed    = FALSE; /* branch optimization converged? */
      }

    return  tpl;
  } /* setupTopol */


void  freeTopol (topol *tpl)
  { /* freeTopol */
    free(tpl->links);
    free(tpl);
  } /* freeTopol */


int  saveSubtree (nodeptr p, topol *tpl)
    /*  Save a subtree in a standard order so that earlier branches
     *  from a node contain lower value tips than do second branches from
     *  the node.  This code works with arbitrary furcations in the tree.
     */
  { /* saveSubtree */
    connptr  r, r0;
    nodeptr  q, s;
    int      t, t0, t1;

    r0 = tpl->links;
    r = r0 + (tpl->nextlink)++;
    r->p = p;
    r->q = q = p->back;
    r->z = p->z;
    r->descend = 0;                     /* No children (yet) */

    if (q->tip) {
      r->valptr = tipValPtr(q);         /* Assign value */
      }

    else {                              /* Internal node, look at children */
      s = q->next;                      /* First child */
      do {
        t = saveSubtree(s, tpl);        /* Generate child's subtree */

        t0 = 0;                         /* Merge child into list */
        t1 = r->descend;
        while (t1 && (cmpTipVal(r0[t1].valptr, r0[t].valptr) < 0)) {
          t0 = t1;
          t1 = r0[t1].sibling;
          }
        if (t0) r0[t0].sibling = t;  else  r->descend = t;
        r0[t].sibling = t1;

        s = s->next;                    /* Next child */
        } while (s != q);

      r->valptr = r0[r->descend].valptr;   /* Inherit first child's value */
      }                                 /* End of internal node processing */

    return  r - r0;
  } /* saveSubtree */


nodeptr  minSubtreeTip (nodeptr  p0)
  { /* minTreeTip */
    nodeptr  minTip, p, testTip;

    if (p0->tip) return p0;

    p = p0->next;
    minTip = minSubtreeTip(p->back);
    while ((p = p->next) != p0) {
      testTip = minSubtreeTip(p->back);
      if (cmpTipVal(tipValPtr(testTip), tipValPtr(minTip)) < 0)
        minTip = testTip;
      }
    return minTip;
  } /* minTreeTip */


nodeptr  minTreeTip (nodeptr  p)
  { /* minTreeTip */
    nodeptr  minp, minpb;

    minp  = minSubtreeTip(p);
    minpb = minSubtreeTip(p->back);
    return cmpTipVal(tipValPtr(minp), tipValPtr(minpb)) < 0 ? minp : minpb;
  } /* minTreeTip */


void saveTree (tree *tr, topol *tpl)
    /*  Save a tree topology in a standard order so that first branches
     *  from a node contain lower value tips than do second branches from
     *  the node.  The root tip should have the lowest value of all.
     */
  { /* saveTree */
    connptr  r;
    int  i;

    tpl->nextlink = 0;                             /* Reset link pointer */
    r = tpl->links + saveSubtree(minTreeTip(tr->start), tpl);  /* Save tree */
    r->sibling = 0;

    tpl->number     = tr->nodeInfo;
    tpl->likelihood = tr->likelihood;
    tpl->start      = tr->start;
    tpl->ntips      = tr->ntips;
    tpl->nextnode   = tr->nextnode;
    tpl->opt_level  = tr->opt_level;
    tpl->prelabeled = tr->prelabeled;
    tpl->smoothed   = tr->smoothed;

   
  } /* saveTree */


void copyTopol (topol *tpl1, topol *tpl2)
  { /* copyTopol */
    connptr  r1, r2, r10, r20;
   
    int  i;

    r10 = tpl1->links;
    r20 = tpl2->links;
    tpl2->nextlink = tpl1->nextlink; 

    r1 = r10;
    r2 = r20;
    i = 2 * tpl1->ntips - 3;
    while (--i >= 0) {
      r2->z = r1->z;
      r2->p = r1->p;
      r2->q = r1->q;
      r2->valptr = r1->valptr;
      r2->descend = r1->descend; 
      r2->sibling = r1->sibling; 
      r1++;
      r2++;
      }

    
    tpl2->likelihood = tpl1->likelihood;
    tpl2->start      = tpl1->start;
    tpl2->ntips      = tpl1->ntips;
    tpl2->nextnode   = tpl1->nextnode;
    tpl2->opt_level  = tpl1->opt_level;
    tpl2->prelabeled = tpl1->prelabeled;
    tpl2->scrNum     = tpl1->scrNum;
    tpl2->tplNum     = tpl1->tplNum;
    tpl2->smoothed   = tpl1->smoothed;
  } /* copyTopol */


boolean restoreTreeRecursive (topol *tpl, tree *tr)
  { /* restoreTree */
    void  hookup();
    boolean  initrav();

    connptr  r;
    nodeptr  p, p0;
    
    int  i;



/*  Copy connections from topology */

    for (r = tpl->links, i = 0; i < tpl->nextlink; r++, i++) {
      hookup(r->p, r->q, r->z);
      }

    tr->likelihood = tpl->likelihood;
    tr->start      = tpl->start;
    tr->ntips      = tpl->ntips;
    
    tr->nextnode   = tpl->nextnode;
    tr->opt_level  = tpl->opt_level;
    tr->prelabeled = tpl->prelabeled;
    tr->smoothed   = tpl->smoothed;

   
    return TRUE;
    
  } /* restoreTree */




boolean restoreTree (topol *tpl, tree *tr)
  { /* restoreTree */
    /*void  hookup();
      boolean  initrav();*/

    connptr  r;
    nodeptr  p, p0;
    
    int  i;

/*  Clear existing connections */

    for (i = 1; i <= 2*(tr->mxtips) - 2; i++) {  /* Uses p = p->next at tip */
      p0 = p = tr->nodep[i];
      do {
        p->back = (nodeptr) NULL;
        p = p->next;
        } while (p != p0);
      }

/*  Copy connections from topology */

    for (r = tpl->links, i = 0; i < tpl->nextlink; r++, i++) {
      hookup(r->p, r->q, r->z);
      }

    tr->likelihood = tpl->likelihood;
    tr->start      = tpl->start;
    tr->ntips      = tpl->ntips;
    
    tr->nextnode   = tpl->nextnode;
    tr->opt_level  = tpl->opt_level;
    tr->prelabeled = tpl->prelabeled;
    tr->smoothed   = tpl->smoothed;

   

    return (initrav(tr, tr->start) && initrav(tr, tr->start->back));
  } /* restoreTree */


boolean restoreTopology (topol *tpl, tree *tr)
  { 

    connptr  r;
    nodeptr  p, p0;
    
    int  i;

    /* TODO can we reomove that code removed here everywhere?*/

    for (r = tpl->links, i = 0; i < tpl->nextlink; r++, i++) 
      {
	hookup(r->p, r->q, r->z);
      }

    tr->likelihood = tpl->likelihood;
    tr->start      = tpl->start;
    tr->ntips      = tpl->ntips;
    
    tr->nextnode   = tpl->nextnode;
    tr->opt_level  = tpl->opt_level;
    tr->prelabeled = tpl->prelabeled;
    tr->smoothed   = tpl->smoothed;

   

    return (initrav(tr, tr->start) && initrav(tr, tr->start->back));
  } /* restoreTree */


int initBestTree (bestlist *bt, int newkeep, int numsp, int sites)
  { /* initBestTree */
    int  i, nlogf;


    bt->nkeep = 0;

    if (bt->ninit <= 0) {
      if (! (bt->start = setupTopol(numsp, sites)))  return  0;
      bt->ninit = -1;
      bt->nvalid = 0;
      bt->numtrees = 0;
      bt->best = unlikely;
      bt->improved = FALSE;
      bt->byScore = (topol **) malloc((newkeep+1) * sizeof(topol *));
      bt->byTopol = (topol **) malloc((newkeep+1) * sizeof(topol *));
      if (! bt->byScore || ! bt->byTopol) {
        printf( "initBestTree: malloc failure\n");
        return 0;
        }
      }
    else if (ABS(newkeep) > bt->ninit) {
      if (newkeep <  0) newkeep = -(bt->ninit);
      else newkeep = bt->ninit;
      }

    if (newkeep < 1) {    /*  Use negative newkeep to clear list  */
      newkeep = -newkeep;
      if (newkeep < 1) newkeep = 1;
      bt->nvalid = 0;
      bt->best = unlikely;
      }

    if (bt->nvalid >= newkeep) {
      bt->nvalid = newkeep;
      bt->worst = bt->byScore[newkeep]->likelihood;
      }
    else 
      {
	bt->worst = unlikely;
      }

    for (i = bt->ninit + 1; i <= newkeep; i++) 
      {    
	if (! (bt->byScore[i] = setupTopol(numsp, 0)))  break;
	bt->byTopol[i] = bt->byScore[i];
	bt->ninit = i;
      }

    return  (bt->nkeep = MIN(newkeep, bt->ninit));
  } /* initBestTree */



int resetBestTree (bestlist *bt)
  { /* resetBestTree */
    bt->best     = unlikely;
    bt->worst    = unlikely;
    bt->nvalid   = 0;
    bt->improved = FALSE;
  } /* resetBestTree */


boolean  freeBestTree(bestlist *bt)
  { /* freeBestTree */
    while (bt->ninit >= 0)  freeTopol(bt->byScore[(bt->ninit)--]);
    freeTopol(bt->start);
    return TRUE;
  } /* freeBestTree */


/*  Compare two trees, assuming that each is in standard order.  Return
 *  -1 if first preceeds second, 0 if they are identical, or +1 if first
 *  follows second in standard order.  Lower number tips preceed higher
 *  number tips.  A tip preceeds a corresponding internal node.  Internal
 *  nodes are ranked by their lowest number tip.
 */

int  cmpSubtopol (connptr p10, connptr p1, connptr p20, connptr p2)
  { /* cmpSubtopol */
    connptr  p1d, p2d;
    int  cmp;

    if (! p1->descend && ! p2->descend)          /* Two tips */
      return cmpTipVal(p1->valptr, p2->valptr);

    if (! p1->descend) return -1;                /* p1 = tip, p2 = node */
    if (! p2->descend) return  1;                /* p2 = tip, p1 = node */

    p1d = p10 + p1->descend;
    p2d = p20 + p2->descend;
    while (1) {                                  /* Two nodes */
      if (cmp = cmpSubtopol(p10, p1d, p20, p2d))  return cmp; /* Subtrees */
      if (! p1d->sibling && ! p2d->sibling)  return  0; /* Lists done */
      if (! p1d->sibling) return -1;             /* One done, other not */
      if (! p2d->sibling) return  1;             /* One done, other not */
      p1d = p10 + p1d->sibling;                  /* Neither done */
      p2d = p20 + p2d->sibling;
      }
  } /* cmpSubtopol */



int  cmpTopol (void *tpl1, void *tpl2)
  { /* cmpTopol */
    connptr  r1, r2;
    int      cmp;

    r1 = ((topol *) tpl1)->links;
    r2 = ((topol *) tpl2)->links;
    cmp = cmpTipVal(tipValPtr(r1->p), tipValPtr(r2->p));
    if (cmp) return cmp;
    return  cmpSubtopol(r1, r1, r2, r2);
  } /* cmpTopol */



int  cmpTplScore (void *tpl1, void *tpl2)
  { /* cmpTplScore */
    double  l1, l2;

    l1 = ((topol *) tpl1)->likelihood;
    l2 = ((topol *) tpl2)->likelihood;
    return  (l1 > l2) ? -1 : ((l1 == l2) ? 0 : 1);
  } /* cmpTplScore */



/*  Find an item in a sorted list of n items.  If the item is in the list,
 *  return its index.  If it is not in the list, return the negative of the
 *  position into which it should be inserted.
 */

int  findInList (void *item, void *list[], int n, int (* cmpFunc)())
  { /* findInList */
    int  mid, hi, lo, cmp;

    if (n < 1) return  -1;                    /*  No match; first index  */

    lo = 1;
    mid = 0;
    hi = n;
    while (lo < hi) {
      mid = (lo + hi) >> 1;
      cmp = (* cmpFunc)(item, list[mid-1]);
      if (cmp) {
        if (cmp < 0) hi = mid;
        else lo = mid + 1;
        }
      else  return  mid;                        /*  Exact match  */
      }

    if (lo != mid) {
       cmp = (* cmpFunc)(item, list[lo-1]);
       if (cmp == 0) return lo;
       }
    if (cmp > 0) lo++;                         /*  Result of step = 0 test  */
    return  -lo;
  } /* findInList */



int  findTreeInList (bestlist *bt, tree *tr)
  { /* findTreeInList */
    topol  *tpl;

    tpl = bt->byScore[0];
    saveTree(tr, tpl);
    return  findInList((void *) tpl, (void **) (& (bt->byTopol[1])),
                       bt->nvalid, cmpTopol);
  } /* findTreeInList */


int  saveBestTree (bestlist *bt, tree *tr)
  { /* saveBestTree */
   
    topol  *tpl, *reuse;
    int  tplNum, scrNum, reuseScrNum, reuseTplNum, i, oldValid, newValid;

    tplNum = findTreeInList(bt, tr);
    tpl = bt->byScore[0];
    oldValid = newValid = bt->nvalid;

    if (tplNum > 0) {                      /* Topology is in list  */
      reuse = bt->byTopol[tplNum];         /* Matching topol  */
      reuseScrNum = reuse->scrNum;
      reuseTplNum = reuse->tplNum;
      }
                                           /* Good enough to keep? */
    else if (tr->likelihood < bt->worst)  return 0;

    else {                                 /* Topology is not in list */
      tplNum = -tplNum;                    /* Add to list (not replace) */
      if (newValid < bt->nkeep) bt->nvalid = ++newValid;
      reuseScrNum = newValid;              /* Take worst tree */
      reuse = bt->byScore[reuseScrNum];
      reuseTplNum = (newValid > oldValid) ? newValid : reuse->tplNum;
      if (tr->likelihood > bt->start->likelihood) bt->improved = TRUE;
      }

    scrNum = findInList((void *) tpl, (void **) (& (bt->byScore[1])),
                         oldValid, cmpTplScore);
    scrNum = ABS(scrNum);

    if (scrNum < reuseScrNum)
      for (i = reuseScrNum; i > scrNum; i--)
        (bt->byScore[i] = bt->byScore[i-1])->scrNum = i;

    else if (scrNum > reuseScrNum) {
      scrNum--;
      for (i = reuseScrNum; i < scrNum; i++)
        (bt->byScore[i] = bt->byScore[i+1])->scrNum = i;
      }

    if (tplNum < reuseTplNum)
      for (i = reuseTplNum; i > tplNum; i--)
        (bt->byTopol[i] = bt->byTopol[i-1])->tplNum = i;

    else if (tplNum > reuseTplNum) {
      tplNum--;
      for (i = reuseTplNum; i < tplNum; i++)
        (bt->byTopol[i] = bt->byTopol[i+1])->tplNum = i;
      }

   

    tpl->scrNum = scrNum;
    tpl->tplNum = tplNum;
    bt->byTopol[tplNum] = bt->byScore[scrNum] = tpl;
    bt->byScore[0] = reuse;

    if (scrNum == 1)  bt->best = tr->likelihood;
    if (newValid == bt->nkeep) bt->worst = bt->byScore[newValid]->likelihood;

    return  scrNum;
  } /* saveBestTree */




int  recallBestTreeRecursive (bestlist *bt, int rank, tree *tr)
  { /* recallBestTree */
    if (rank < 1)  rank = 1;
    if (rank > bt->nvalid)  rank = bt->nvalid;
    if (rank > 0)  if (! restoreTreeRecursive(bt->byScore[rank], tr)) return FALSE;
    return  rank;
  } /* recallBestTree */



int  recallBestTree (bestlist *bt, int rank, tree *tr)
  { /* recallBestTree */
    if (rank < 1)  rank = 1;
    if (rank > bt->nvalid)  rank = bt->nvalid;
    if (rank > 0)  if (! restoreTree(bt->byScore[rank], tr)) return FALSE;
    return  rank;
  } /* recallBestTree */

int  recallBestTopology (bestlist *bt, int rank, tree *tr)
  { /* recallBestTree */
    if (rank < 1)  rank = 1;
    if (rank > bt->nvalid)  rank = bt->nvalid;
    if (rank > 0)  if (! restoreTopology(bt->byScore[rank], tr)) return FALSE;
    return  rank;
  } /* recallBestTree */

/*=======================================================================*/
/*                       End of best tree routines                       */
/*=======================================================================*/


void getnums (rawdata *rdta)
    /* input number of species, number of sites */
  { /* getnums */
    

    if (fscanf(INFILE, "%d %d", & rdta->numsp, & rdta->sites) != 2) 
      {
	printf("ERROR: Problem reading number of species and sites\n");
	exit(-1);
      }
   

    if (rdta->numsp < 4) 
      {
	printf("TOO FEW SPECIES\n");
	exit(-1);
      }

    if (rdta->sites < 1) 
      {
	printf("TOO FEW SITES\n");
	exit(-1);
      }

    return;
  } /* getnums */


boolean digitchar (int ch) 
{
  return (ch >= '0' && ch <= '9'); 
}


boolean whitechar (int ch)
{ 
  return (ch == ' ' || ch == '\n' || ch == '\t');
}


void uppercase (int *chptr)
    /* convert character to upper case -- either ASCII or EBCDIC */
  { /* uppercase */
    int  ch;

    ch = *chptr;
    if ((ch >= 'a' && ch <= 'i') || (ch >= 'j' && ch <= 'r')
                                 || (ch >= 's' && ch <= 'z'))
      *chptr = ch + 'A' - 'a';
  } /* uppercase */


int base36 (int ch)
  { /* base36 */
    if      (ch >= '0' && ch <= '9') return (ch - '0');
    else if (ch >= 'A' && ch <= 'I') return (ch - 'A' + 10);
    else if (ch >= 'J' && ch <= 'R') return (ch - 'J' + 19);
    else if (ch >= 'S' && ch <= 'Z') return (ch - 'S' + 28);
    else if (ch >= 'a' && ch <= 'i') return (ch - 'a' + 10);
    else if (ch >= 'j' && ch <= 'r') return (ch - 'j' + 19);
    else if (ch >= 's' && ch <= 'z') return (ch - 's' + 28);
    else return -1;
  } /* base36 */


int itobase36 (int i)
  { /* itobase36 */
    if      (i <  0) return '?';
    else if (i < 10) return (i      + '0');
    else if (i < 19) return (i - 10 + 'A');
    else if (i < 28) return (i - 19 + 'J');
    else if (i < 36) return (i - 28 + 'S');
    else return '?';
  } /* itobase36 */


int findch (int c)
  { /* findch */
    int ch;

    while ((ch = getc(INFILE)) != EOF && ch != c) ;
    return  ch;
  } /* findch */

void getyspace (rawdata *rdta)
  { /* getyspace */
    long   size;
    int    i;
    char *y0;

    if (! (rdta->y = (char **) malloc((rdta->numsp + 1) * sizeof(char *)))) 
      {
	printf("ERROR: Unable to obtain space for data array pointers\n");
	exit(-1);
      }

    size = 4 * (rdta->sites / 4 + 1);
    if (! (y0 = (char *) malloc((rdta->numsp + 1) * size * sizeof(char)))) 
      {
	printf("ERROR: Unable to obtain space for data array\n");
	exit(-1);
      }

    rdta->y0 = y0;

    for (i = 0; i <= rdta->numsp; i++) 
      {
	rdta->y[i] = y0;
	y0 += size;
      }

    return;
  } /* getyspace */


void freeyspace(rawdata *rdta)
{
  long   size;
  int    i;
  char *y0;
  for (i = 0; i <= rdta->numsp; i++) 
    {
      free(rdta->y[i]);    
    }
}



boolean setupTree (tree *tr, int nsites, analdef *adef)
  { /* setupTree */
    nodeptr  p0, p, q;
    int      i, j, tips, inter, code;

    tips  = tr->mxtips;
    inter = tr->mxtips - 1;

#ifdef PARALLEL_2
    tr->rDesc = (rearrangementDescriptor *)malloc(sizeof(rearrangementDescriptor));
#endif
   

    tr->nameList = (char **)malloc(sizeof(char *) * (tips + 1));
    
    tr->outgr     =           1;
          
    /*tr->rearrOrder = (nodeptr *)malloc(tips * sizeof(nodeptr));*/

    if (!(p0 = (nodeptr) malloc((tips + 3*inter) * sizeof(node)))) {
      printf("ERROR: Unable to obtain sufficient tree memory\n");
      return  FALSE;
      }


    if (!(tr->nodep = (nodeptr *) malloc((2*tr->mxtips) * sizeof(nodeptr)))) {
      printf("ERROR: Unable to obtain sufficient tree memory, too\n");
      return  FALSE;
      }

    
    tr->nodep[0] = (node *) NULL;    /* Use as 1-based array */

    for (i = 1; i <= tips; i++) {    /* Set-up tips */
      p = p0++;
      p->x      =  NULL;
      p->tip    = (char *) NULL;      
      p->number = i;
      p->next   = p;
      p->back   = (node *) NULL;      
      tr->nodep[i] = p;
      }

    for (i = tips + 1; i <= tips + inter; i++) { /* Internal nodes */
      q = (node *) NULL;
      for (j = 1; j <= 3; j++) {
        p = p0++;
        p->x      =  NULL;
        p->tip    = (char *) NULL;
        p->number = i;

        p->next   = q;
        p->back   = (node *) NULL;
        q = p;
        }
      p->next->next->next = p;
      tr->nodep[i] = p;
      }

    tr->likelihood  = unlikely;
    tr->start       = (node *) NULL;
    tr->outgrnode   = tr->nodep[tr->outgr];
    tr->ntips       = 0;
    tr->nextnode    = 0;
    tr->opt_level   = 0;
    tr->prelabeled  = TRUE;
    tr->smoothed    = FALSE;

   
    /* TODO depends on model */

   

    return TRUE;
  } /* setupTree */

void freeTreeNode (nodeptr p)   /* Free tree node (sector) associated data */
  { /* freeTreeNode */
    if (p) {
      if (p->x) 
	{
	  free(p->x);      
        }
      }
  } /* freeTree */



/* AxML modification start*/






void freeTree (tree *tr)
  { /* freeTree */
    nodeptr  p, q;
    int  i, tips, inter;

    tips  = tr->mxtips;
    inter = tr->mxtips - 1;

    for (i = 1; i <= tips; i++) freeTreeNode(tr->nodep[i]);

    for (i = tips + 1; i <= tips + inter; i++) {
      if (p = tr->nodep[i]) {
        if (q = p->next) {
          freeTreeNode(q->next);
          freeTreeNode(q);
          }
        freeTreeNode(p);
        }
      }
    

    free(tr->nodep[1]);       /* Free the actual nodes */
  } /* freeTree */







boolean getdata (analdef *adef, rawdata *rdta, tree *tr)
{
  int   i, j, k, l, basesread, basesnew, ch, my_i;
  int   meaning[256];          /*  meaning of input characters */ 
  char *nameptr;
  boolean  allread, firstpass;
  char buffer[300]; 
  int len;
   
  for (i = 0; i <= 255; i++) 
    meaning[i] = -1;

  if(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT)
    {          
      meaning['A'] =  0;  /* alanine */
      meaning['R'] =  1;  /* arginine */
      meaning['N'] =  2;  /*  asparagine*/
      meaning['D'] =  3;  /* aspartic */
      meaning['C'] =  4;  /* cysteine */
      meaning['Q'] =  5;  /* glutamine */
      meaning['E'] =  6;  /* glutamic */
      meaning['G'] =  7;  /* glycine */
      meaning['H'] =  8;  /* histidine */
      meaning['I'] =  9;  /* isoleucine */
      meaning['L'] =  10; /* leucine */
      meaning['K'] =  11; /* lysine */
      meaning['M'] =  12; /* methionine */
      meaning['F'] =  13; /* phenylalanine */
      meaning['P'] =  14; /* proline */
      meaning['S'] =  15; /* serine */
      meaning['T'] =  16; /* threonine */
      meaning['W'] =  17; /* tryptophan */
      meaning['Y'] =  18; /* tyrosine */
      meaning['V'] =  19; /* valine */
      meaning['B'] =  20; /* asparagine, aspartic 2 and 3*/
      meaning['Z'] =  21; /* glutamine glutamic 5 and 6*/
      meaning['X'] =  meaning['?'] = meaning['*'] = meaning['-'] = 22; /* all = 1.0 */
    }
  else
    {      
      meaning['A'] =  1;
      meaning['B'] = 14;
      meaning['C'] =  2;
      meaning['D'] = 13;
      meaning['G'] =  4;
      meaning['H'] = 11;
      meaning['K'] = 12;
      meaning['M'] =  3;
      meaning['N'] = 15;
      meaning['O'] = 15;
      meaning['R'] =  5;
      meaning['S'] =  6;
      meaning['T'] =  8;
      meaning['U'] =  8;
      meaning['V'] =  7;
      meaning['W'] =  9;
      meaning['X'] = 15;
      meaning['Y'] = 10;
      meaning['?'] = 15;
      meaning['-'] = 15;	
    }
  
  basesread = basesnew = 0;

  allread = FALSE;
  firstpass = TRUE;
  ch = ' ';

  while (! allread) 
    {
      for (i = 1; i <= tr->mxtips; i++) 
	{   	  
	  if (firstpass) 
	    {                      	       
	      ch = getc(INFILE);
	      while(ch == ' ' || ch == '\n' || ch == '\t')
		{
		  ch = getc(INFILE);
		}

	      my_i = 0;

	      do 
		{
		  buffer[my_i] = ch;
		  ch = getc(INFILE);		   
		  my_i++;
		  if(my_i >= nmlngth)
		    {
		      printf("ERROR Names are to long, adapt constant nmlngth in\n");
		      printf("axml.h AND dnapars.c, current setting %d\n", nmlngth);
		      exit(-1);
		    }
		}
	      while(ch !=  ' ' && ch != '\n' && ch != '\t');
		
	      buffer[my_i] = '\0';
			
	      len = strlen(buffer) + 1;
			       
	      tr->nameList[i] = (char *)malloc(sizeof(char) * len);

	      strcpy(tr->nameList[i], buffer);		
	    }


	  j = basesread;
	  while ((j < rdta->sites) && ((ch = getc(INFILE)) != EOF) && ((! adef->interleaved) || (ch != '\n'))) 
	    {
	      uppercase(& ch);
	      if ((meaning[ch] != -1) || ch == '.') 
		{
		  j++;
		  if (ch == '.') 
		    {
		      if (i != 1) ch = rdta->y[1][j];
		      else 
			{
			  printf("ERROR: Dot (.) found at site %d of sequence 1\n", j);
			  return  FALSE;
			}
		    }
		  rdta->y[i][j] = ch;
		}
	      else 
		if(whitechar(ch) || digitchar(ch)) ;
		else 
		  {
		    printf("ERROR: Bad base (%c) at site %d of sequence %d\n",
			   ch, j, i);
		    return  FALSE;
		  }
	    }

	    if (ch == EOF) 
	      {
		printf("ERROR: End-of-file at site %d of sequence %d\n", j, i);
		return  FALSE;
	      }

	    if (! firstpass && (j == basesread)) i--; 
	    else 
	      if (i == 1) basesnew = j;
	      else 
		if (j != basesnew) 
		  {
		    printf("ERROR: Sequences out of alignment\n");		    
		    printf("%d (instead of %d) residues read in sequence %d %s\n",
			   j - basesread, basesnew - basesread, i, tr->nameList[i]);
		    return  FALSE;
		  }
	    while (ch != '\n' && ch != EOF) ch = getc(INFILE);  /* flush line */
	}                                                  /* next sequence */
      firstpass = FALSE;
      basesread = basesnew;
      allread = (basesread >= rdta->sites);
    }

 
    
  for (j = 1; j <= tr->mxtips; j++)
    {
      for (i = 1; i <= rdta->sites; i++) 
	{
	  rdta->y[j][i] = meaning[rdta->y[j][i]];		
	}
    }   	

    return  TRUE;
  } /* getdata */



void inputweights (analdef *adef, rawdata *rdta, cruncheddata *cdta)
    /* input the character weights 0, 1, 2 ... 9, A, B, ... Y, Z */
  { /* inputweights */
    int i, ch, w, fres;
    FILE *weightFile;
    int *wv = (int *)malloc(sizeof(int) *  rdta->sites + 1);
  

    weightFile = fopen(weightFileName, "r");
    if (!weightFile)
      {
	printf( "Could not open weight file: %s\n", weightFileName);
	exit(1);
      }
     
    i = 1;
    
    while((fres = fscanf(weightFile,"%d", &w)) != EOF)
      {
	if(!fres)
	  {
	    printf("error reading weight file probably encountered a non-integer weight value\n");
	    exit(-1);
	  }
	wv[i] = w;
	i++;	
      }
    
   
    if(i != (rdta->sites + 1))
      {
	printf("number %d of weights not equal to number %d of alignment columns\n", i, rdta->sites);
	exit(-1);
      }

    cdta->wgtsum = 0;
    for(i = 1; i <= rdta->sites; i++) 
      {     
	rdta->wgt[i] = wv[i];
	cdta->wgtsum += rdta->wgt[i];
      }

    fclose(weightFile);
    free(wv);
  } /* inputweights */







void getinput (analdef *adef, rawdata *rdta, cruncheddata *cdta, tree *tr)
{ 
  int i,j;

  getnums(rdta);
  tr->outgr          = 1;
  tr->mxtips         = rdta->numsp;
  rdta->wgt          = (int *)    malloc((rdta->sites + 1) * sizeof(int));
  rdta->wgt2         = (int *)    malloc((rdta->sites + 1) * sizeof(int));
  cdta->alias        = (int *)    malloc((rdta->sites + 1) * sizeof(int));
  cdta->reAlias      = (int *)    malloc((rdta->sites + 1) * sizeof(int));
  cdta->aliaswgt     = (int *)    malloc((rdta->sites + 1) * sizeof(int));
  cdta->patrat       = (double *) malloc((rdta->sites + 1) * sizeof(double));
  cdta->patratStored = (double *) malloc((rdta->sites + 1) * sizeof(double));
  cdta->rateCategory = (int *)    malloc((rdta->sites + 1) * sizeof(int));
  cdta->wr           = (double *) malloc((rdta->sites + 1) * sizeof(double));  
  cdta->wr2          = (double *) malloc((rdta->sites + 1) * sizeof(double));   

  if(!adef->useWeightFile)
    {
      for (i = 1; i <= rdta->sites; i++) 
	rdta->wgt[i] = 1;
     
      cdta->wgtsum = rdta->sites;   
    }
  else
    {	
      inputweights(adef, rdta, cdta);
    }
    
  getyspace(rdta);
  setupTree(tr, rdta->sites, adef);
  getdata(adef, rdta, tr);
        
  return;
} 




double randum (long  *seed)
    /* random number generator, modified to use 12 bit chunks */
  { /* randum */
    long  sum, mult0, mult1, seed0, seed1, seed2, newseed0, newseed1, newseed2;
    double res;

    mult0 = 1549;
    seed0 = *seed & 4095;
    sum  = mult0 * seed0;
    newseed0 = sum & 4095;
    sum >>= 12;
    seed1 = (*seed >> 12) & 4095;
    mult1 =  406;
    sum += mult0 * seed1 + mult1 * seed0;
    newseed1 = sum & 4095;
    sum >>= 12;
    seed2 = (*seed >> 24) & 255;
    sum += mult0 * seed2 + mult1 * seed1;
    newseed2 = sum & 255;

    *seed = newseed2 << 24 | newseed1 << 12 | newseed0;
    res = 0.00390625 * (newseed2 + 0.000244140625 * (newseed1 + 0.000244140625 * newseed0));  

    return res;   
  } /* randum */






void makeboot (analdef *adef, rawdata *rdta, cruncheddata *cdta)
  {
    int  i, j, nonzero;
#ifdef PARALLEL
    long seed;
#endif

   
    nonzero = 0;
   
    for (i = 1; i <= rdta->sites; i++)  
      {
	if (rdta->wgt[i] > 0) nonzero++;		
      }

    for (j = 1; j <= nonzero; j++) 
      cdta->aliaswgt[j] = 0;     

#ifdef PARALLEL 
    seed = (long) gettimeSrand();
    for (j = 1; j <= nonzero; j++)
      cdta->aliaswgt[(int) (nonzero*randum(& seed)) + 1]++;
#endif    

#ifndef PARALLEL
    for (j = 1; j <= nonzero; j++)
      cdta->aliaswgt[(int) (nonzero*randum(& adef->boot)) + 1]++;
#endif

    j = 0;
    cdta->wgtsum = 0;
    for (i = 1; i <= rdta->sites; i++) 
      {
	if (rdta->wgt[i] > 0)
	  cdta->wgtsum += (rdta->wgt2[i] = rdta->wgt[i] * cdta->aliaswgt[++j]);
	else
	  rdta->wgt2[i] = 0;
      }        
  } 


void sitesort (rawdata *rdta, cruncheddata *cdta)
    /* Shell sort keeping sites with identical residues and weights in
     * the original order (i.e., a stable sort).
     * The index created in cdta->alias is 1 based.  The
     * sitecombcrunch routine packs it to a 0 based index.
     */
  { /* sitesort */
    int  gap, i, j, jj, jg, k, n, nsp;
    int  *index, *category;
    boolean  flip, tied;
    char  **data;

    index    = cdta->alias;
    data     = rdta->y;
    n        = rdta->sites;
    nsp      = rdta->numsp;

    for (gap = n / 2; gap > 0; gap /= 2) {
      for (i = gap + 1; i <= n; i++) {
        j = i - gap;

        do {
          jj = index[j];
          jg = index[j+gap];
          flip = 0;
          tied = 1;
          for (k = 1; (k <= nsp) && tied; k++) {
            flip = (data[k][jj] >  data[k][jg]);
            tied = (data[k][jj] == data[k][jg]);
            }
          if (flip) {
            index[j]     = jg;
            index[j+gap] = jj;
            j -= gap;
            }
          } while (flip && (j > 0));

        }  /* for (i ...   */
      }    /* for (gap ... */
  } /* sitesort */


void sitecombcrunch (rawdata *rdta, cruncheddata *cdta)
    /* combine sites that have identical patterns (and nonzero weight) */
  { /* sitecombcrunch */
    int  i, sitei, j, sitej, k;
    int gaps = 0;
    boolean  tied;

    i = 0;
    cdta->alias[0] = cdta->alias[1];
    cdta->aliaswgt[0] = 0;   
    
    for (j = 1; j <= rdta->sites; j++) 
      {
	sitei = cdta->alias[i];
	sitej = cdta->alias[j];
	tied = 1;

	for (k = 1; tied && (k <= rdta->numsp); k++)
	  tied = (rdta->y[k][sitei] == rdta->y[k][sitej]);

	if (tied) 
	  {
	    cdta->aliaswgt[i] += rdta->wgt2[sitej];
	  }
	else 
	  {
	    if (cdta->aliaswgt[i] > 0) i++;
	    cdta->aliaswgt[i] = rdta->wgt2[sitej];
	    cdta->alias[i] = sitej;
	  }
      }

    cdta->endsite = i;
    if (cdta->aliaswgt[i] > 0) cdta->endsite++;       
  } /* sitecombcrunch */




boolean makeweights (analdef *adef, rawdata *rdta, cruncheddata *cdta)
    /* make up weights vector to avoid duplicate computations */
  { /* makeweights */
    int  i;

    if (adef->boot)  
      makeboot(adef, rdta, cdta);
    else 
      {
	for (i = 1; i <= rdta->sites; i++)  
	  rdta->wgt2[i] = rdta->wgt[i];
      }

    for (i = 1; i <= rdta->sites; i++)  
      cdta->alias[i] = i;

   

    sitesort(rdta, cdta);
    sitecombcrunch(rdta, cdta);
    return TRUE;
  } /* makeweights */











boolean makevalues (rawdata *rdta, cruncheddata *cdta)
    /* set up fractional likelihoods at tips */
  { /* makevalues */
    double  temp, wtemp;
    int  i, j;
    char buf;      

    for (i = 1; i <= rdta->numsp; i++) 
      {
	for (j = 0; j < cdta->endsite; j++) 
	  {	   
	      rdta->y[i-1][j] = rdta->y[i][cdta->alias[j]];
	  }
      }
               
    for (j = 0; j < cdta->endsite; j++) 
      {
	cdta->patrat[j] = temp = 1.0;
	cdta->patratStored[j] = 1.0;
	cdta->rateCategory[j] = 0;
	cdta->wr[j]  = wtemp = temp * cdta->aliaswgt[j];
	cdta->wr2[j] = temp * wtemp;
      }
 
    return TRUE;
  } /* makevalues */


char *itobase36_str(i)
     int i;
  { 
    static char buf[4];
    int j = 2;
    strcpy(buf,"  0");
    while(i && j >= 0) {
      buf[j--] = itobase36(i % 36);
      i /= 36;
    }
    return buf;
  }

boolean empiricalfreqs (analdef *adef, rawdata *rdta, cruncheddata *cdta)
    /* Get empirical base frequencies from the data */
  { /* empiricalfreqs */
    double  sum, suma, sumc, sumg, sumt, wj, fa, fc, fg, ft, sumb;
    int     i, j, k, code;
    char  *yptr;
    
    for (j = 0; j < cdta->endsite; j++) 
      {
	for (k = 1; k <= 8; k++) 
	  {
	     rdta->freqa = 0.25;
	     rdta->freqc = 0.25;
	     rdta->freqg = 0.25;
	     rdta->freqt = 0.25;
	    suma = 0.0;
	    sumc = 0.0;
	    sumg = 0.0;
	    sumt = 0.0;
	    for (i = 0; i < rdta->numsp; i++) 
	      {
		yptr = rdta->y[i];
		
		code = yptr[j];
		fa = rdta->freqa * ( code       & 1);
		fc = rdta->freqc * ((code >> 1) & 1);
		fg = rdta->freqg * ((code >> 2) & 1);
		ft = rdta->freqt * ((code >> 3) & 1);
		wj = cdta->aliaswgt[j] / (fa + fc + fg + ft);
		suma += wj * fa;
		sumc += wj * fc;
		sumg += wj * fg;
		sumt += wj * ft;
	      }	  
	  }
	sum = suma + sumc + sumg + sumt;
	rdta->freqa = suma / sum;
	rdta->freqc = sumc / sum;
	rdta->freqg = sumg / sum;
	rdta->freqt = sumt / sum;		
      }

    rdta->freqa = 0.25;
    rdta->freqc = 0.25;
    rdta->freqg = 0.25;
    rdta->freqt = 0.25;

    for (k = 1; k <= 8; k++) {
      suma = 0.0;
      sumc = 0.0;
      sumg = 0.0;
      sumt = 0.0;
      for (i = 0; i < rdta->numsp; i++) {
        yptr = rdta->y[i];
        for (j = 0; j < cdta->endsite; j++) {
          code = *yptr++;
          fa = rdta->freqa * ( code       & 1);
          fc = rdta->freqc * ((code >> 1) & 1);
          fg = rdta->freqg * ((code >> 2) & 1);
          ft = rdta->freqt * ((code >> 3) & 1);
          wj = cdta->aliaswgt[j] / (fa + fc + fg + ft);
          suma += wj * fa;
          sumc += wj * fc;
          sumg += wj * fg;
          sumt += wj * ft;
          }
        }
      sum = suma + sumc + sumg + sumt;
      rdta->freqa = suma / sum;
      rdta->freqc = sumc / sum;
      rdta->freqg = sumg / sum;
      rdta->freqt = sumt / sum;
      }

   rdta->freqr = rdta->freqa + rdta->freqg;
    rdta->invfreqr = 1.0/rdta->freqr;
    rdta->freqar = rdta->freqa * rdta->invfreqr;
    rdta->freqgr = rdta->freqg * rdta->invfreqr;
    rdta->freqy = rdta->freqc + rdta->freqt;
    rdta->invfreqy = 1.0/rdta->freqy;
    rdta->freqcy = rdta->freqc * rdta->invfreqy;
    rdta->freqty = rdta->freqt * rdta->invfreqy;
    
    suma = rdta->ttratio*rdta->freqr*rdta->freqy
         - (rdta->freqa*rdta->freqg + rdta->freqc*rdta->freqt);
    sumb = rdta->freqa*rdta->freqgr + rdta->freqc*rdta->freqty;
    rdta->xi = suma/(suma+sumb);
    rdta->xv = 1.0 - rdta->xi;
    if (rdta->xi <= 0.0) {
      /*printf("WARNING: This transition/transversion ratio\n");
      printf("         is impossible with these base frequencies!\n");
      printf("Transition/transversion parameter reset\n\n");*/
      rdta->xi = 0.000001;
      rdta->xv = 1.0 - rdta->xi;
      rdta->ttratio = (sumb * rdta->xi / rdta->xv 
                       + rdta->freqa * rdta->freqg
                       + rdta->freqc * rdta->freqt)
                    / (rdta->freqr * rdta->freqy);
     
      }
   
    rdta->fracchange = 2.0 * rdta->xi * (rdta->freqa * rdta->freqgr
                                       + rdta->freqc * rdta->freqty)
                     + rdta->xv * (1.0 - rdta->freqa * rdta->freqa
                                       - rdta->freqc * rdta->freqc
                                       - rdta->freqg * rdta->freqg
                                       - rdta->freqt * rdta->freqt);
    
    return TRUE;
  } /* empiricalfreqs */






void alterTT(tree *tr, double t_value)
  { /* reportfreqs */
    double  suma, sumb;       
    
    rawdata *rdta = tr->rdta;
    tr->t_value = t_value;
    suma = t_value*rdta->freqr*rdta->freqy
         - (rdta->freqa*rdta->freqg + rdta->freqc*rdta->freqt);
    sumb = rdta->freqa*rdta->freqgr + rdta->freqc*rdta->freqty;
    rdta->xi = suma/(suma+sumb);
    rdta->xv = 1.0 - rdta->xi;
    if (rdta->xi <= 0.0) {
      /*      printf("WARNING: This transition/transversion ratio\n");
      printf("         is impossible with these base frequencies!\n");
      printf("Transition/transversion parameter reset\n\n");*/
      rdta->xi = 0.000001;
      rdta->xv = 1.0 - rdta->xi;
      rdta->ttratio = (sumb * rdta->xi / rdta->xv 
                       + rdta->freqa * rdta->freqg
                       + rdta->freqc * rdta->freqt)
                    / (rdta->freqr * rdta->freqy);
     
      }
   
    rdta->fracchange = 2.0 * rdta->xi * (rdta->freqa * rdta->freqgr
                                       + rdta->freqc * rdta->freqty)
                     + rdta->xv * (1.0 - rdta->freqa * rdta->freqa
                                       - rdta->freqc * rdta->freqc
                                       - rdta->freqg * rdta->freqg
                                       - rdta->freqt * rdta->freqt);
  } 




void linkdata2tree (rawdata *rdta, cruncheddata *cdta, tree *tr)    
{ 
  int  i;
  
  tr->rdta       = rdta;
  tr->cdta       = cdta;  
    
  for (i = 1; i <= tr->mxtips; i++)   
    tr->nodep[i]->tip = &(rdta->y[i-1][0]);      
}







void allocNodex (tree *tr, analdef *adef)
{
  nodeptr  p;
  int  i;
    
  for (i = tr->mxtips + 1; (i <= 2*(tr->mxtips) - 2); i++) 
    {
      p = tr->nodep[i];
      allocCount++;
      if(adef->model == M_GTRGAMMA)
	{	       
	  p->x = malloc(tr->cdta->endsite * sizeof(gammalikelivector));	    	  
	}
      else
	{      	
	  p->x = malloc(tr->cdta->endsite * sizeof(likelivector));	    		  
	}		
    }  
} 


void freeNodex(tree *tr)
{
  nodeptr  p;
  int  i;  
  
  for (i = tr->mxtips + 1; (i <= 2*(tr->mxtips) - 2); i++) 
    {
      p = tr->nodep[i];
      while(!p->x)
	p = p->next;
      allocCount--;
      free(p->x);
      p->x = NULL;
    }

  if(allocCount != 0)
    {
      printf("FATAL error\n");
      exit(-1);
    }

}


void *getxnode (nodeptr p)  
{ 
  nodeptr  s;

  if (! (p->x)) 
    { 
      if ((s = p->next)->x || (s = s->next)->x) 
	{
	  p->x = s->x;
	  s->x = NULL;
        }     
    }
  return  p->x;
}


boolean update (tree *tr, nodeptr p)
  { /* update */
    
    nodeptr  q;
    double   z0, z;
	
    q = p->back;
    z0 = q->z;
        
    if ((z = makenewz(tr, p, q, z0, newzpercycle)) == badZ) 
      {	  
	return FALSE;
      }
    
    p->z = q->z = z;
    if (ABS(z - z0) > deltaz)  tr->smoothed = FALSE;
        
    return TRUE;
  } /* update */



boolean smooth (tree *tr, nodeptr p)
  { /* smooth */
    nodeptr  q;
  
    if (! update(tr, p))               return FALSE; /*  Adjust branch */
    if (! p->tip) 
      {                                  /*  Adjust descendants */
        q = p->next;
        while (q != p) 
	  {
	    if (! smooth(tr, q->back))   return FALSE;
	    q = q->next;
          }	
	if (! newview(tr, p)) return FALSE;	
      }

    return TRUE;
  } /* smooth */


boolean smoothTree (tree *tr, int maxtimes)
  { /* smoothTree */
    nodeptr  p, q;   

    /*printf("smoothTree %d \n", maxtimes);*/

    p = tr->start;

    while (--maxtimes >= 0) 
      {
	tr->smoothed = TRUE;
	if (! smooth(tr, p->back))       return FALSE;
	if (! p->tip) 
	  {
	    q = p->next;
	    while (q != p) 
	      {
		if (! smooth(tr, q->back))   return FALSE;
		q = q->next;
	      }
	  }
	if (tr->smoothed)  break;
      }

    return TRUE;
  } /* smoothTree */


boolean localSmooth (tree *tr, nodeptr p, int maxtimes)
  { /* localSmooth -- Smooth branches around p */
    nodeptr  q;

    if (p->tip) return FALSE;            /* Should be an error */

    while (--maxtimes >= 0) {
      tr->smoothed = TRUE;
      q = p;
      do {
        if (! update(tr, q)) return FALSE;
        q = q->next;
        } while (q != p);
      if (tr->smoothed)  break;
      }

    tr->smoothed = FALSE;             /* Only smooth locally */
    return TRUE;
  } /* localSmooth */


void hookup (nodeptr p, nodeptr q, double z)
  { /* hookup */
    p->back = q;
    q->back = p;
    p->z = q->z = z;
  } /* hookup */


/* Insert node p into branch q <-> q->back */


boolean insert (tree *tr, nodeptr p, nodeptr q, boolean glob)
/* glob -- Smooth tree globally? */

/*                q
                 /.
             add/ .
               /  .
             pn   .
    s ---- p      .remove
             pnn  .
               \  .
             add\ .
                 \.      pn  = p->next;
                  r      pnn = p->next->next;
 */

  { /* insert */
    nodeptr  r, s;

    r = q->back;
    s = p->back;

    if(Thorough)
    { 
      double  zqr, zqs, zrs, lzqr, lzqs, lzrs, lzsum, lzq, lzr, lzs, lzmax;      
     
	  

    
      if ((zqr = makenewz(tr, q, r, q->z, iterations)) == badZ) return FALSE; 
      
      if ((zqs = makenewz(tr, q, s, defaultz, iterations)) == badZ) return FALSE;      
      
      if ((zrs = makenewz(tr, r, s, defaultz, iterations)) == badZ) return FALSE;
    

      lzqr = (zqr > zmin) ? log(zqr) : log(zmin); 
      lzqs = (zqs > zmin) ? log(zqs) : log(zmin);
      lzrs = (zrs > zmin) ? log(zrs) : log(zmin);
      lzsum = 0.5 * (lzqr + lzqs + lzrs);
      
      lzq = lzsum - lzrs;
      lzr = lzsum - lzqs;
      lzs = lzsum - lzqr;
      lzmax = log(zmax);
      
      if      (lzq > lzmax) {lzq = lzmax; lzr = lzqr; lzs = lzqs;} 
      else if (lzr > lzmax) {lzr = lzmax; lzq = lzqr; lzs = lzrs;}
      else if (lzs > lzmax) {lzs = lzmax; lzq = lzqs; lzr = lzrs;}
      
      
      hookup(p->next,       q, exp(lzq));
      hookup(p->next->next, r, exp(lzr));
      hookup(p,             s, exp(lzs));      		  
    }
    else
      { 	
	double  z = sqrt(q->z);
	
	if(z < zmin) 
	  z = zmin;
	if(z > zmax)
	  z = zmax;

	hookup(p->next,       q, z);
	hookup(p->next->next, r, z);
      }


   
    
    if (! newview(tr, p)) return FALSE;
    
   
    tr->opt_level = 0;
   
    if(Thorough)
    { 
      
      if (! localSmooth(tr, p, smoothings)) return FALSE;	
     
    }           
     
    return  TRUE;
  } /* insert */


nodeptr  removeNode (tree *tr, nodeptr p)

/*                q
                 .|
          remove. |
               .  |
             pn   |
    s ---- p      |add
             pnn  |
               .  |
          remove. |
                 .|      pn  = p->next;
                  r      pnn = p->next->next;
 */

    /* remove p and return where it was */
  { /* removeNode */
    double   zqr;
    nodeptr  q, r;
    
    q = p->next->back;
    r = p->next->next->back;
    zqr = q->z * r->z;
   

    
    if ((zqr = makenewz(tr, q, r, zqr, iterations)) == badZ) 
      return (node *) NULL;      

    

    hookup(q, r, zqr);

    p->next->next->back = p->next->back = (node *) NULL;
     
    return  q;
  } /* removeNode */


boolean initrav (tree *tr, nodeptr p)
  { /* initrav */
    nodeptr  q;

    if (! p->tip) 
      {
      
	q = p->next;
      
	do 
	  {
	    if (! initrav(tr, q->back))  return FALSE;	
	    q = q->next;	
	  } 
	while (q != p);
            
	if (! newview(tr, p)) return FALSE;	      
      }

    return TRUE;
  } /* initrav */





boolean initravDIST (tree *tr, nodeptr p, int distance)
  {
    nodeptr  q;

    if (! p->tip && distance > 0) 
      {
      
	q = p->next;
      
	do 
	  {
	    if (! initravDIST(tr, q->back, --distance))  return FALSE;	
	    q = q->next;	
	  } 
	while (q != p);
      
      
	if (! newview(tr, p)) return FALSE;	      
      }

    return TRUE;
  } /* initrav */


boolean readKeyValue (char *string, char *key, char *format, void *value)
  { /* readKeyValue */

    if (!(string = strstr(string, key)))  return FALSE;
    string += strlen(key);
    string = strchr(string, '=');
    if (! (string = strchr(string, '=')))  return FALSE;
    string++;
    return  sscanf(string, format, value);  /* 1 if read, otherwise 0 */
  } /* readKeyValue */


void printTree(FILE *outf, tree *tr, int quoted)
{
  char *str = (char *)malloc(treeStringLength);
  char *outString;
  nodeptr startp;
  startp = tr->start;

  tr->start = tr->nodep[1];
  
  treeString(str, tr, tr->start->back, 1);
  
  tr->start = startp;

  outString = str;
  while(*outString++ != '(');
  fprintf(outf, "("); 
  if(!quoted)
  {
    while(*outString != '\0')
      {
	if(*outString != '\'')
	  fprintf(outf, "%c", *outString); 
	outString++;
      }
  }
      
  free(str);
  fflush(outf);
}


void printResult(tree *tr, int quoted)
{
  char *str = (char *)malloc(treeStringLength);
  char *outString;
  FILE *outf = fopen(resultFileName, "w");
  nodeptr startp;
  startp = tr->start;

  tr->start = tr->nodep[1];

  treeString(str, tr, tr->start->back, 1);  
  tr->start = startp; 
  outString = str;
  while(*outString++ != '(');
  fprintf(outf, "("); 
  if(!quoted)
  {
    while(*outString != '\0')
      {
	if(*outString != '\'')
	  fprintf(outf, "%c", *outString); 
	outString++;
      }
  }
      
  free(str);
  fclose(outf);  
}

void printLog(tree *tr, analdef *adef)
{
  
  #ifdef PARALLEL_2
  if(myId == 0)
    {
  #endif
      if(adef->checkpoints)
	{
	  char thisCheckPointName[2048], str[16];
	  FILE *cpf;
	  
	  printf( "%f %f %d\n", gettime() - masterTime, tr->likelihood, checkPointCounter);
	  fprintf(logFile,"%f %f %d\n", gettime() - masterTime, tr->likelihood, checkPointCounter); 
	  
	  strcpy(thisCheckPointName, checkpointFileName);
	  sprintf(str, "%d", checkPointCounter);
	  strcat(thisCheckPointName, str);
	  checkPointCounter++;
	  cpf = fopen(thisCheckPointName, "w");
	  printTree(cpf, tr, 0);
	  fflush(cpf);
	  fclose(cpf);    
	}
      else
	{
	  printf( "%f %f\n", gettime() - masterTime, tr->likelihood);
	  fprintf(logFile,"%f %f\n", gettime() - masterTime, tr->likelihood); 
	}
      
 
      fflush(logFile);
#ifdef PARALLEL_2
    }
#endif
    
  

}

/*===========  This is a problem if tr->start->back is a tip!  ===========*/
/*  All routines should be contrived so that tr->start->back is not a tip */

void initTreeString(char *treestr, tree *tr, int form)
{
   (void) sprintf(treestr, "[&&%s: version = '%s'",
                                 programName, programVersion);
   while (*treestr) treestr++;

   (void) sprintf(treestr, ", %s = %15.13g",
		  likelihood_key, unlikely);
   while (*treestr) treestr++;
   
   (void) sprintf(treestr, ", %s = %d", ntaxa_key, tr->mxtips);
   while (*treestr) treestr++;
   
   (void) sprintf(treestr,", %s = %d", opt_level_key, 0);
   while (*treestr) treestr++;
   
   (void) sprintf(treestr, ", %s = %d", smoothed_key, 0);
   while (*treestr) treestr++;
   
   (void) sprintf(treestr, "]%s", form == treeProlog ? ", " : " ");
   while (*treestr) treestr++;
}


char *treeString (char *treestr, tree *tr, nodeptr p, int form)
    /* write string with representation of tree */
    /*   form == 1 -> Newick tree */
    /*   form == 2 -> Prolog fact */
    /*   form == 3 -> PHYLIP tree */
  { /* treeString */
    double  x, z;
    char  *nameptr;
    int    c;

    if (p == tr->start->back) {
      if (form != treePHYLIP) {
        if (form == treeProlog) {
          (void) sprintf(treestr, "phylip_tree(");
          while (*treestr) treestr++;            /* move pointer to null */
          }

        (void) sprintf(treestr, "[&&%s: version = '%s'",
                                 programName, programVersion);
        while (*treestr) treestr++;

        (void) sprintf(treestr, ", %s = %15.13g",
                                 likelihood_key, tr->likelihood);
        while (*treestr) treestr++;

        (void) sprintf(treestr, ", %s = %d", ntaxa_key, tr->ntips);
        while (*treestr) treestr++;

        (void) sprintf(treestr,", %s = %d", opt_level_key, tr->opt_level);
        while (*treestr) treestr++;

        (void) sprintf(treestr, ", %s = %d", smoothed_key, tr->smoothed);
        while (*treestr) treestr++;

        (void) sprintf(treestr, "]%s", form == treeProlog ? ", " : " ");
        while (*treestr) treestr++;
        }
    }

    if (p->tip) {
      if (form != treePHYLIP) *treestr++ = '\'';
     
      nameptr = tr->nameList[p->number];
     
      while (c = *nameptr++) {
        if (form != treePHYLIP) {if (c == '\'') *treestr++ = '\'';}
        else if (c == ' ') {c = '_';}
        *treestr++ = c;
        }
      if (form != treePHYLIP) *treestr++ = '\'';
    }

    else {
      *treestr++ = '(';
      treestr = treeString(treestr, tr, p->next->back, form);
      *treestr++ = ',';
      treestr = treeString(treestr, tr, p->next->next->back, form);
      if (p == tr->start->back) {
        *treestr++ = ',';
        treestr = treeString(treestr, tr, p->back, form);
        }
      *treestr++ = ')';
    }

    if (p == tr->start->back) 
      {
	(void) sprintf(treestr, ":0.0%s\n", (form != treeProlog) ? ";" : ").");
      }
    else 
      {
	z = p->z;
	if (z < zmin) z = zmin;      	 
	x = -log(z) * tr->rdta->fracchange;	
	(void) sprintf(treestr, ":%8.6f", x);  /* prolog needs the space */
      }

    while (*treestr) treestr++;     /* move pointer up to null termination */
    return  treestr;
  } /* treeString */

void treeOut (FILE *treefile, tree *tr, int form)
    /* write out file with representation of final tree */
  { /* treeOut */
    int    c;
    char  *cptr, *treestr;

    treestr = (char *) malloc((tr->ntips * (nmlngth+32)) + 256);
    if (! treestr) {
      printf( "treeOut: malloc failure\n");
      exit(1);
      }

    (void) treeString(treestr, tr, tr->start->back, form);
    cptr = treestr;
    while (c = *cptr++) putc(c, treefile);

    printf( "%s\n", treestr);
    printf("%s\n", treestr);

    free(treestr);
  } /* treeOut */


/*=======================================================================*/
/*                         Read a tree from a file                       */
/*=======================================================================*/


/*  1.0.A  Processing of quotation marks in comment removed
 */

int treeFinishCom (FILE *fp, char **strp)
  { /* treeFinishCom */
    int  ch;

    while ((ch = getc(fp)) != EOF && ch != ']') {
      if (strp != NULL) *(*strp)++ = ch;    /* save character  */
      if (ch == '[') {                      /* nested comment; find its end */
        if ((ch = treeFinishCom(fp, strp)) == EOF)  break;
        if (strp != NULL) *(*strp)++ = ch;  /* save closing ]  */
        }
      }

    if (strp != NULL) **strp = '\0';        /* terminate string  */
    return  ch;
  } /* treeFinishCom */


int treeGetCh (FILE *fp)         /* get next nonblank, noncomment character */
  { /* treeGetCh */
    int  ch;

    while ((ch = getc(fp)) != EOF) {
      if (whitechar(ch)) ;
      else if (ch == '[') {                   /* comment; find its end */
        if ((ch = treeFinishCom(fp, (char **) NULL)) == EOF)  break;
        }
      else  break;
      }

    return  ch;
  } /* treeGetCh */


boolean  treeLabelEnd (int ch)
  { /* treeLabelEnd */
    switch (ch) {
        case EOF:  case '\0':  case '\t':  case '\n':  case ' ':
        case ':':  case ',':   case '(':   case ')':   case '[':
        case ';':
          return TRUE;
        default:
          break;
        }
    return FALSE;
  } /* treeLabelEnd */


boolean  treeGetLabel (FILE *fp, char *lblPtr, int maxlen)
  { /* treeGetLabel */
    int      ch;
    boolean  done, quoted, lblfound;

    if (--maxlen < 0) lblPtr = (char *) NULL;  /* reserves space for '\0' */
    else if (lblPtr == NULL) maxlen = 0;

    ch = getc(fp);
    done = treeLabelEnd(ch);

    lblfound = ! done;
    quoted = (ch == '\'');
    if (quoted && ! done) {ch = getc(fp); done = (ch == EOF);}

    while (! done) {
      if (quoted) {
        if (ch == '\'') {ch = getc(fp); if (ch != '\'') break;}
        }

      else if (treeLabelEnd(ch)) break;
      /*
	else if (ch == '_') ch = ' ';*/  /* unquoted _ goes to space */

      if (--maxlen >= 0) *lblPtr++ = ch;
      ch = getc(fp);
      if (ch == EOF) break;
      }

    if (ch != EOF)  (void) ungetc(ch, fp);

    if (lblPtr != NULL) *lblPtr = '\0';

    return lblfound;
  } /* treeGetLabel */


boolean  treeFlushLabel (FILE *fp)
  { /* treeFlushLabel */
    return  treeGetLabel(fp, (char *) NULL, (int) 0);
  } /* treeFlushLabel */


int  treeFindTipByLabel (char  *str, tree *tr)
                     /*  str -- label string pointer */
  { /* treeFindTipByLabel */
    nodeptr  q;
    char    *nameptr;
    int      ch, i, n;
    boolean  found;
    char num[64];
    char *numptr;

    for (n = 1; n <= tr->mxtips; n++) 
      {
	q = tr->nodep[n];

	if (! (q->back)) 
	  {        
	    i = 0;
	    
	    nameptr = tr->nameList[q->number];
       
	    while ((found = (str[i++] == (ch = *nameptr++))) && ch) ;
	    if (found) return n; 	 	

	    /*
	      Uncomment this to read NEXUS-style trees with taxon number instead of names
	      i = 0;
	      numptr = num;
	      sprintf(numptr,"%d", q->number);
	      while((found = (str[i++] == (ch = *numptr++))) && ch) ;
	      if (found) return n; 
	    */
	  }
      
      }

    printf("ERROR: Cannot find tree species: %s\n", str);

    return  0;
  } /* treeFindTipByLabel */


int  treeFindTipName (FILE *fp, tree *tr)
  { /* treeFindTipName */
    char    *nameptr, str[nmlngth+2];
    int      n;

    if (tr->prelabeled) {
      if (treeGetLabel(fp, str, nmlngth+2))
        n = treeFindTipByLabel(str, tr);
      else
        n = 0;
      }

    else if (tr->ntips < tr->mxtips) {
      n = tr->ntips + 1;
      
      nameptr = tr->nameList[n];
   
      if (! treeGetLabel(fp, nameptr, nmlngth+1)) n = 0;
      }

    else {
      n = 0;
      }

    return  n;
  } /* treeFindTipName */


void  treeEchoContext (FILE *fp1, FILE *fp2, int n)
 { /* treeEchoContext */
   int      ch;
   boolean  waswhite;

   waswhite = TRUE;

   while (n > 0 && ((ch = getc(fp1)) != EOF)) {
     if (whitechar(ch)) {
       ch = waswhite ? '\0' : ' ';
       waswhite = TRUE;
       }
     else {
       waswhite = FALSE;
       }

     if (ch > '\0') {putc(ch, fp2); n--;}
     }
 } /* treeEchoContext */


boolean treeProcessLength (FILE *fp, double *dptr)
  { /* treeProcessLength */
    int  ch;

    if ((ch = treeGetCh(fp)) == EOF)  return FALSE;    /*  Skip comments */
    (void) ungetc(ch, fp);

    if (fscanf(fp, "%lf", dptr) != 1) {
      printf("ERROR: treeProcessLength: Problem reading branch length\n");
      treeEchoContext(fp, stdout, 40);
      printf("\n");
      return  FALSE;
      }

    return  TRUE;
  } /* treeProcessLength */


int treeFlushLen (FILE  *fp)
  { /* treeFlushLen */
    double  dummy;
    boolean res;
    int     ch;

    ch = treeGetCh(fp);
    
    if (ch == ':') 
      {
	ch = treeGetCh(fp);
	
	    ungetc(ch, fp);
	    if(!treeProcessLength(fp, & dummy)) return 0;
	    return 1;	  
      }
   
    

    if (ch != EOF) (void) ungetc(ch, fp);
    return 1;
  } /* treeFlushLen */





boolean  treeNeedCh (FILE *fp, int c1, char *where)
  { /* treeNeedCh */
    int  c2;

    if ((c2 = treeGetCh(fp)) == c1)  return TRUE;

    printf("ERROR: Expecting '%c' %s tree; found:", c1, where);
    if (c2 == EOF) {
      printf("End-of-File");
      }
    else {
      ungetc(c2, fp);
      treeEchoContext(fp, stdout, 40);
      }
    putchar('\n');
    return FALSE;
  } /* treeNeedCh */





boolean  addElementLen (FILE *fp, tree *tr, nodeptr p)
  { /* addElementLen */
    double   z, branch;
    nodeptr  q;
    int      n, ch, fres;
    
    if ((ch = treeGetCh(fp)) == '(') {     /*  A new internal node */
      n = (tr->nextnode)++;
      if (n > 2*(tr->mxtips) - 2) {
        if (tr->rooted || n > 2*(tr->mxtips) - 1) {
          printf("ERROR: Too many internal nodes.  Is tree rooted?\n");
          printf("       Deepest splitting should be a trifurcation.\n");
          return FALSE;
          }
        else {
          tr->rooted = TRUE;
          }
        }
      q = tr->nodep[n];
      if (! addElementLen(fp, tr, q->next))        return FALSE;
      if (! treeNeedCh(fp, ',', "in"))             return FALSE;
      if (! addElementLen(fp, tr, q->next->next))  return FALSE;
      if (! treeNeedCh(fp, ')', "in"))             return FALSE;
      (void) treeFlushLabel(fp);
      }

    else {                               /*  A new tip */
      ungetc(ch, fp);
      if ((n = treeFindTipName(fp, tr)) <= 0)          return FALSE;
      q = tr->nodep[n];
      if (tr->start->number > n)  tr->start = q;
      (tr->ntips)++;
      }                                  /* End of tip processing */

  

 
    
    fres = treeFlushLen(fp);
    if(!fres) return FALSE;
    
    hookup(p, q, defaultz);
    return TRUE;
      
    
  } /* addElementLen */






int saveTreeCom (char  **comstrp)
  { /* saveTreeCom */
    int  ch;
    boolean  inquote;

    inquote = FALSE;
    while ((ch = getc(INFILE)) != EOF && (inquote || ch != ']')) {
      *(*comstrp)++ = ch;                        /* save character  */
      if (ch == '[' && ! inquote) {              /* comment; find its end */
        if ((ch = saveTreeCom(comstrp)) == EOF)  break;
        *(*comstrp)++ = ch;                      /* add ] */
        }
      else if (ch == '\'') inquote = ! inquote;  /* start or end of quote */
      }

    return  ch;
  } /* saveTreeCom */


boolean processTreeCom (FILE *fp, tree *tr)
  { /* processTreeCom */
    int   text_started, functor_read, com_open;

    /*  Accept prefatory "phylip_tree(" or "pseudoNewick("  */

    functor_read = text_started = 0;
    (void) fscanf(fp, " p%nhylip_tree(%n", & text_started, & functor_read);
    if (text_started && ! functor_read) {
      (void) fscanf(fp, "seudoNewick(%n", & functor_read);
      if (! functor_read) {
        printf("Start of tree 'p...' not understood.\n");
        return FALSE;
        }
      }

    com_open = 0;
    (void) fscanf(fp, " [%n", & com_open);

    if (com_open) {                                  /* comment; read it */
      char  com[1024], *com_end;

      com_end = com;
      if (treeFinishCom(fp, & com_end) == EOF) {     /* omits enclosing []s */
        printf("Missing end of tree comment\n");
        return FALSE;
        }

      (void) readKeyValue(com, likelihood_key, "%lg",
                               (void *) &(tr->likelihood));
      (void) readKeyValue(com, opt_level_key,  "%d",
                               (void *) &(tr->opt_level));
      (void) readKeyValue(com, smoothed_key,   "%d",
                               (void *) &(tr->smoothed));

      if (functor_read) (void) fscanf(fp, " ,");   /* remove trailing comma */
      }

    return (functor_read > 0);
  } /* processTreeCom */


nodeptr uprootTree (tree *tr, nodeptr p)
  { /* uprootTree */
    nodeptr  q, r, s, start;
    int      n;

    if (p->tip || p->back) {
      printf("ERROR: Unable to uproot tree.\n");
      printf("       Inappropriate node marked for removal.\n");
      return (nodeptr) NULL;
      }

    n = --(tr->nextnode);               /* last internal node added */
    if (n != tr->mxtips + tr->ntips - 1) {
      printf("ERROR: Unable to uproot tree.  Inconsistent\n");
      printf("       number of tips and nodes for rooted tree.\n");
      return (nodeptr) NULL;
      }

    q = p->next->back;                  /* remove p from tree */
    r = p->next->next->back;
    hookup(q, r, defaultz);

    start = (r->tip || (! q->tip)) ? r : r->next->next->back;

    if (tr->ntips > 2 && p->number != n) {
      q = tr->nodep[n];            /* transfer last node's conections to p */
      r = q->next;
      s = q->next->next;
      hookup(p,             q->back, q->z);   /* move connections to p */
      hookup(p->next,       r->back, r->z);
      hookup(p->next->next, s->back, s->z);
      if (start->number == q->number) start = start->back->back;
      q->back = r->back = s->back = (nodeptr) NULL;
      }
    else {
      p->back = p->next->back = p->next->next->back = (nodeptr) NULL;
      }

    tr->rooted = FALSE;
    return  start;
  } /* uprootTree */

boolean treeReadLen (FILE *fp, tree *tr)
  { /* treeReadLen */
    nodeptr  p;
    int      i, ch;
    boolean  is_fact;

    for (i = 1; i <= tr->mxtips; i++) tr->nodep[i]->back = (node *) NULL;
    tr->start       = tr->nodep[tr->mxtips];
    tr->ntips       = 0;
    tr->nextnode    = tr->mxtips + 1;
    tr->opt_level   = 0;
   
    tr->smoothed    = FALSE;
    tr->rooted      = FALSE;

    is_fact = processTreeCom(fp, tr);

    p = tr->nodep[(tr->nextnode)++]; 
    while(ch = treeGetCh(fp) != '(');
      
    
    /*if (! treeNeedCh(fp, '(', "at start of"))       return FALSE;*/
    if (! addElementLen(fp, tr, p))                 return FALSE;
    if (! treeNeedCh(fp, ',', "in"))                return FALSE;
    if (! addElementLen(fp, tr, p->next))           return FALSE;
    if (! tr->rooted) {
      if ((ch = treeGetCh(fp)) == ',') {        /*  An unrooted format */
        if (! addElementLen(fp, tr, p->next->next)) return FALSE;
        }
      else {                                    /*  A rooted format */
        tr->rooted = TRUE;
        if (ch != EOF)  (void) ungetc(ch, fp);
        }
      }
    else {
      p->next->next->back = (nodeptr) NULL;
      }
    if (! treeNeedCh(fp, ')', "in"))                return FALSE;
    (void) treeFlushLabel(fp);
    if (! treeFlushLen(fp))                         return FALSE;
    if (is_fact) {
      if (! treeNeedCh(fp, ')', "at end of"))       return FALSE;
      if (! treeNeedCh(fp, '.', "at end of"))       return FALSE;
      }
    else {
      if (! treeNeedCh(fp, ';', "at end of"))       return FALSE;
      }

    if (tr->rooted) {
      p->next->next->back = (nodeptr) NULL;
      tr->start = uprootTree(tr, p->next->next);
      if (! tr->start)                              return FALSE;
      }
    else {
      tr->start = p->next->next->back;  /* This is start used by treeString */
      }

    return  (initrav(tr, tr->start) && initrav(tr, tr->start->back));
  } /* treeReadLen */

/*=======================================================================*/
/*                        Read a tree from a string                      */
/*=======================================================================*/

int str_treeFinishCom (char **treestrp, char **strp)
                      /* treestrp -- tree string pointer */
                      /* strp -- comment string pointer */
  { /* str_treeFinishCom */
    int  ch;
    while ( (ch=*(*treestrp)++)!='\0' && ch!=']' ) {
      if (strp != NULL) *(*strp)++ = ch;    /* save character  */
      if (ch == '[') {                      /* nested comment; find its end */
        if ((ch=str_treeFinishCom(treestrp,strp)) == '\0')  break;
        if (strp != NULL) *(*strp)++ = ch;  /* save closing ]  */
        }
      }
    if (strp != NULL) **strp = '\0';        /* terminate string  */
    return  ch;
  } /* str_treeFinishCom */


int str_treeGetCh (char **treestrp)
    /* get next nonblank, noncomment character */
  { /* str_treeGetCh */
    int  ch;
    while ((ch=*(*treestrp)++) != '\0') {
      if (whitechar(ch)) ;
      else if (ch == '[') {                  /* comment; find its end */
        if ((ch = str_treeFinishCom(treestrp,(char**)NULL)) == '\0')  break;
        }
      else  break;
      }

    return  ch;
  } /* str_treeGetCh */




boolean  str_treeGetLabel (char **treestrp, char *lblPtr, int maxlen)
  { /* str_treeGetLabel */
    int      ch;
    boolean  done, quoted, lblfound;


    if (--maxlen < 0)
      lblPtr = (char*)NULL;  /* reserves space for '\0' */
    else if(lblPtr == NULL)
      maxlen = 0;

    ch = *(*treestrp)++;
    done = treeLabelEnd(ch);

    lblfound = !done;
    quoted = (ch == '\'');
    if (quoted && ! done) {
      ch = *(*treestrp)++;
      done = (ch == '\0');
    }

    while(!done) {
      if (quoted) 
	{
	  if (ch == '\'') 
	    {
	      ch = *(*treestrp)++;
	      if (ch != '\'') break;
	    }
	}
      else 
	if (treeLabelEnd(ch))
	  break;
      /*else 
	  if (ch == '_')
	    ch = ' ';*/
      if (--maxlen >= 0) *lblPtr++ = ch;
      ch = *(*treestrp)++;
      if (ch == '\0') break;
    }

    (*treestrp)--;

    if (lblPtr != NULL) *lblPtr = '\0';

    return lblfound;
  } /* str_treeGetLabel */


boolean  str_treeFlushLabel (char **treestrp)
  { /* str_treeFlushLabel */
    return  str_treeGetLabel(treestrp, (char*)NULL, (int)0);
  } /* str_treeFlushLabel */


int  str_treeFindTipName (char **treestrp, tree *tr)         /*DKB-orig*/
/*int  str_treeFindTipName (char **treestrp, tree *tr, int ch)*/   /*DKB-change*/
  { /* str_treeFindTipName */
    nodeptr  q;
    char    *nameptr, str[nmlngth+2];
    int      i, n;

    if (tr->prelabeled) {
      if (str_treeGetLabel(treestrp, str, nmlngth+2)) {
        n = treeFindTipByLabel(str, tr);
      }
      else
        n = 0;
      }

    else if (tr->ntips < tr->mxtips) {
      n = tr->ntips + 1;
     
      nameptr = tr->nameList[n];
    
      if (! str_treeGetLabel(treestrp, nameptr, nmlngth+1)) n = 0;
      }

    else {
      n = 0;
      }

    return  n;
  } /* str_treeFindTipName */


boolean str_treeProcessLength (char **treestrp, double *dptr)
  { /* str_treeProcessLength */
    int     used;

    if(!str_treeGetCh(treestrp))  return FALSE;    /*  Skip comments */
    (*treestrp)--;

    if (sscanf(*treestrp, "%lf%n", dptr, &used) != 1) {
      printf("ERROR: str_treeProcessLength: Problem reading branch length\n");
      printf("%40s\n", *treestrp);
      *dptr = 0.0;
      return FALSE;
      }
    else {
      *treestrp += used;
      }

    return  TRUE;
  } /* str_treeProcessLength */


boolean  str_treeFlushLen (char **treestrp)
  { /* str_treeFlushLen */
    int  ch;
    double  x;

    if ((ch = str_treeGetCh(treestrp)) == ':')
    /*return str_treeProcessLength(treestrp, (double*)NULL);*/   /*DKB-orig*/
      return str_treeProcessLength(treestrp, &x);              /*DKB-change*/
    else {
      (*treestrp)--;
      return TRUE;
      }
  } /* str_treeFlushLen */


boolean  str_treeNeedCh (char **treestrp, int c1, char *where)
  { /* str_treeNeedCh */
    int  c2, i;

    if ((c2 = str_treeGetCh(treestrp)) == c1)  return TRUE;

    printf("ERROR: Missing '%c' %s tree; ", c1, where);
    if (c2 == '\0') 
      printf("end-of-string");
    else {
      putchar('"');
      for (i = 24; i-- && (c2 != '\0'); c2 = *(*treestrp)++)  putchar(c2);
      putchar('"');
      }

    printf(" found instead\n");
    return FALSE;
  } /* str_treeNeedCh */


boolean str_processTreeCom(tree *tr, char **treestrp)
  { /* str_processTreeCom */
    char  *com, *com_end;
    int  text_started, functor_read, com_open;

    com = *treestrp;

    /* Comment must begin with either "phylip_tree" or "pseudoNewick".
     * If it is neither, return FALSE. */
    functor_read = text_started = 0;
    sscanf(com, " p%nhylip_tree(%n", &text_started, &functor_read);
    if (functor_read) {
      com += functor_read;
    }
    else if (text_started) {
      com += text_started;
      sscanf(com, "seudoNewick(%n", &functor_read);
      if (! functor_read) {
        printf("Start of tree 'p...' not understood.\n");
        return  FALSE;
      }
      else {
        com += functor_read;
      }
    }

    /* Find opening bracket of comment */
    com_open = 0;
    sscanf(com, " [%n", &com_open);
    com += com_open;

    /* Read comment. */
    if (com_open) {
      if (!(com_end = strchr(com, ']'))) {
        printf("Missing end of tree comment.\n");
        return  FALSE;
      }
      *com_end = 0;
      (void)readKeyValue(com,likelihood_key,"%lg",(void*)&(tr->likelihood));
      (void)readKeyValue(com,opt_level_key, "%d", (void*)&(tr->opt_level));
      (void)readKeyValue(com,smoothed_key,  "%d", (void*)&(tr->smoothed));
      *com_end = ']';
      com_end++;

      /* Remove trailing comma, and return addr of next char in treestp */
      if (functor_read) {
        text_started = 0;
        sscanf(com_end, " ,%n", & text_started);
        com_end += text_started;
      }
      *treestrp = com_end;
    }
    return (functor_read > 0);
  } /* str_processTreeCom */

boolean str_processTreeComMerge(int *ntaxa, char **treestrp)
  { /* str_processTreeCom */
    char  *com, *com_end;
    int  text_started, functor_read, com_open;

    com = *treestrp;

    /* Comment must begin with either "phylip_tree" or "pseudoNewick".
     * If it is neither, return FALSE. */
    functor_read = text_started = 0;
    sscanf(com, " p%nhylip_tree(%n", &text_started, &functor_read);
    if (functor_read) {
      com += functor_read;
    }
    else if (text_started) {
      com += text_started;
      sscanf(com, "seudoNewick(%n", &functor_read);
      if (! functor_read) {
        printf("Start of tree 'p...' not understood.\n");
        return  FALSE;
      }
      else {
        com += functor_read;
      }
    }

    /* Find opening bracket of comment */
    com_open = 0;
    sscanf(com, " [%n", &com_open);
    com += com_open;

    /* Read comment. */
    if (com_open) {
      if (!(com_end = strchr(com, ']'))) {
        printf("Missing end of tree comment.\n");
        return  FALSE;
      }
      *com_end = 0;
      /*(void)readKeyValue(com,likelihood_key,"%lg",(void*)&(tr->likelihood));*/
      (void)readKeyValue(com,ntaxa_key, "%d", (void *)ntaxa);   
      /*      (void)readKeyValue(com,opt_level_key, "%d", (void*)&(tr->opt_level));
	      (void)readKeyValue(com,smoothed_key,  "%d", (void*)&(tr->smoothed));*/
      *com_end = ']';
      com_end++;

      /* Remove trailing comma, and return addr of next char in treestp */
      if (functor_read) {
        text_started = 0;
        sscanf(com_end, " ,%n", & text_started);
        com_end += text_started;
      }
      *treestrp = com_end;
    }
    return (functor_read > 0);
  } /* str_processTreeCom */


boolean  str_addElementLen (char **treestrp, tree *tr, nodeptr p)
  { /* str_addElementLen */
    double   z, branch;
    nodeptr  q;
    int      n, ch;

    if ((ch = str_treeGetCh(treestrp)) == '(') { /*  A new internal node */
      n = (tr->nextnode)++;
      if (n > 2*(tr->mxtips) - 2) {
        if (tr->rooted || n > 2*(tr->mxtips) - 1) {
          printf("ERROR: too many internal nodes.  Is tree rooted?\n");
          printf("Deepest splitting should be a trifurcation.\n");
          return  FALSE;
        }
        else {
          tr->rooted = TRUE;
        }
      }
      q = tr->nodep[n];
      if (! str_addElementLen(treestrp, tr, q->next))          return FALSE;
      if (! str_treeNeedCh(treestrp, ',', "in"))               return FALSE;
      if (! str_addElementLen(treestrp, tr, q->next->next))    return FALSE;
      if (! str_treeNeedCh(treestrp, ')', "in"))               return FALSE;
    /*if (! str_treeFlushLabel(treestrp))                      return FALSE;*//*DKB-orig*/
    }

    else {                           /*  A new tip */
    /*n = str_treeFindTipName(treestrp, tr, ch);*/ /*DKB-orig*/
      if(ch=='\'') (*treestrp)--;                  /*DKB-change*/
      n = str_treeFindTipName(treestrp, tr);       /*DKB-change*/
      if (n <= 0) return FALSE;
      q = tr->nodep[n];
      if (tr->start->number > n)  tr->start = q;
      (tr->ntips)++;
    }

    /*  Master and Slave always use lengths */

    if (! str_treeNeedCh(treestrp, ':', "in"))                 return FALSE;
    if (! str_treeProcessLength(treestrp, &branch))            return FALSE;
   
    z = exp(-branch / tr->rdta->fracchange);
    if (z > zmax)  z = zmax;
    hookup(p, q, z);

    return  TRUE;
  } /* str_addElementLen */




boolean str_treeReadLen (char *treestr, tree *tr)
    /* read string with representation of tree */
  { /* str_treeReadLen */
    nodeptr  p;
    int  i;
    boolean  is_fact, found;

    for(i=1; i<=(tr->mxtips); i++) tr->nodep[i]->back = (node*)NULL;
    tr->start       = tr->nodep[tr->mxtips];
    tr->ntips       = 0;
    tr->nextnode    = tr->mxtips + 1;
    tr->opt_level   = 0;
    tr->smoothed    = 1;/*(myprogtype==DNAML_MASTER); DANGER !!!!!!!!!*/
    tr->rooted      = FALSE;

    is_fact = str_processTreeCom(tr,&treestr);
    p = tr->nodep[(tr->nextnode)++];

    if(!str_treeNeedCh(&treestr, '(', "at start of"))       return FALSE;
    if(!str_addElementLen(&treestr, tr, p))                 return FALSE;
    if(!str_treeNeedCh(&treestr, ',', "in"))                return FALSE;
    if(!str_addElementLen(&treestr, tr, p->next))           return FALSE;
    if(!tr->rooted) {
      if(str_treeGetCh(&treestr) == ',') {        /*  An unrooted format */
        if(!str_addElementLen(&treestr,tr,p->next->next)) return FALSE;
      }
      else {                                       /*  A rooted format */
        p->next->next->back = (nodeptr) NULL;
        tr->rooted = TRUE;
        treestr--;
      }
    }
    if(!str_treeNeedCh(&treestr, ')', "in"))                 return FALSE;
  /*if(!str_treeFlushLabel(&treestr))                        return FALSE;*//*DKB-orig*/
    if(!str_treeFlushLen(&treestr))                          return FALSE;
    if(is_fact) {
      if(!str_treeNeedCh(& treestr, ')', "at end of"))       return FALSE;
      if(!str_treeNeedCh(& treestr, '.', "at end of"))       return FALSE;
    }
    else {
      if(!str_treeNeedCh(& treestr, ';', "at end of"))       return FALSE;
    }

    if(tr->rooted)  if (! uprootTree(tr, p->next->next))     return FALSE;
    tr->start = p->next->next->back;  /* This is start used by treeString */

    return  (initrav(tr,tr->start) && initrav(tr,tr->start->back));
  } /* str_treeReadLen */


boolean treeEvaluate (tree *tr, double smoothFactor)       /* Evaluate a user tree */
  { /* treeEvaluate */
    
    /*double inLH = tr->likelihood;*/

    if (! smoothTree(tr, (int)((double)smoothings * smoothFactor))) 
      {
	return FALSE;      
      }
      
    if (evaluate(tr, tr->start) == badEval)  
      {	
	return FALSE;
      }         

    /*    if(inLH > tr->likelihood)
      {
	printf("FATAL error in treeEvaluate %.20f <-> %.20f factor %d\n", inLH, tr->likelihood, (int)((double)smoothings * smoothFactor));
	}*/

    return TRUE;
  } /* treeEvaluate */



boolean evalFast (tree *tr)
  {
       
    if (! smoothTree(tr, 1)) 
      {
	return FALSE;      
      }
    if (evaluate(tr, tr->start) == badEval)  
      {
	return FALSE;
      }
     
    return TRUE;
  } /* treeEvaluate */


double str_readTreeLikelihood (char *treestr)
  { /* str_readTreeLikelihood */
    double lk1;
    char    *com, *com_end;
    boolean  readKeyValue();

    if ((com = strchr(treestr, '[')) /*&& (com < strchr(treestr, '('))*/
                                     && (com_end = strchr(com, ']'))) 
      {
      com++;
      *com_end = 0;
      if (readKeyValue(com, likelihood_key, "%lg", (void *) &(lk1))) {
        *com_end = ']';
        return lk1;
        }
      }

    printf( "ERROR reading likelihood in receiveTree\n");
    printf("%s \n", treestr);
    return  badEval;
  } /* str_readTreeLikelihood */


void initAdef(analdef *adef)
{
  adef->model = M_GTRCAT;  
  adef->t0 = 3.0;
  adef->repeats = 0;
  adef->max_rearrange = 21;  
  adef->stepwidth = 5;
  adef->initial = 10;
  adef->initialSet = FALSE;
  adef->restart = FALSE;
  adef->mode = BIG_RAPID_MODE;
  adef->categories = 25; 
  adef->boot    =           0;
  adef->interleaved =    TRUE;   
  adef->useWeightFile = FALSE;
  adef->protmodel = M_JTT;
  adef->checkpoints = 0;
  adef->startingTreeOnly = 0;
  adef->rapidParsimony = 0;
  adef->useMixedModel = 0;
  adef->multipleRuns = 0;
  adef->parsimonySeed = 0;
}


void printVersionInfo()
{
  printf("\nThis is %s version %s released by Alexandros Stamatakis in %s\n\n",  programName, programVersion, programDate);
}

void printREADME()
{
  printVersionInfo();  
  printf("\n");
  printf("Please also consult the RAxML-manual\n");
  printf("To report bugs send an email to stamatak@ics.forth.gr\n\n\n");
  


  printf("raxmlHPC[-MPI|-OMP] [-a weightFileName] [-b bootstrapRandomNumberSeed] [-c numberOfCategories] [-f d|e]\n");
  printf("                    [-h] [-i initialRearrangementSetting] [-j] [-m GTRCAT|GTRMIX|GTRGAMMA]\n");
  printf("                    [-t userStartingTree] [-w workingDirectory] [-v] [-y]  [-# numberOfRuns]\n");
  printf("                    -s sequenceFileName  -n outputFileName\n");
  printf("\n");
  printf("	-a 	Specify a column weight file name to assign individual weights to each column of \n");
  printf("		the alignment. Those weights must be integers separated by any type and number \n");
  printf("		of whitespaces whithin a separate file, see file \"example_weights\" for an example.\n");
  printf("\n");
  printf("	-b	Specify an integer number (random seed) and turn on bootstrapping\n");
  printf("\n");
  printf("		DEFAULT: OFF\n");
  printf("\n");
  printf("	-c 	Specify number of distinct rate catgories for RAxML when modelOfEvolution\n");
  printf("		is set to GTRCAT or GTRMIX\n");
  printf("		Individual per-site rates are categorized into numberOfCategories rate \n");
  printf("		categories to accelerate computations. \n");
  printf("\n");
  printf("		DEFAULT: 25\n");
  printf("\n");
  printf("	-f	select algorithm:\n");
  printf("\n");
  printf("		\"-f d\": rapid hill-climbing \n");
  printf("		\"-f e\": optimize model+branch lengths for given input tree under GTRGAMMA only\n"); 
  printf("\n");
  printf("		DEFAULT: rapid hill climbing\n");  
  printf("\n");
  printf("	-i  	Initial rearrangement setting for the subsequent application of topological \n");
  printf("		changes phase\n");
  printf("\n");
  printf("		DEFAULT: determined by program\n");
  printf("\n");
  printf("	-j	Specifies if checkpoints will be written by the program. If checkpoints \n");
  printf("		(intermediate tree topologies) shall be written by the program specify \"-j\"\n");
  printf("\n");
  printf("		DEFAULT: OFF\n");
  printf("\n");
  printf("	-h	Display this help message.\n");  
  printf("\n");
  printf("	-m      Model of Nucleotide Substitution: \n");  
  printf("\n");
  printf("		\"-m GTRCAT\":   GTR + Optimization of substitution rates +  Optimization of site-specific\n");
  printf("                               evolutionary rates which are categorized into numberOfCategories distinct \n");
  printf("                               rate categories for greater computational efficiency\n");
  printf("                               if you do a multiple analysis with  \"-#\" but without bootstrapping the program\n");
  printf("                               will use GTRMIX instead\n");
  printf("		\"-m GTRGAMMA\": GTR + Optimization of substitution rates + GAMMA model of rate \n");
  printf("                               heterogeneity (alpha parameter will be estimated)\n");
  printf("		\"-m GTRMIX\":   Inference of the tree under GTRCAT\n");
  printf("                               and thereafter evaluation of the final tree topology under GTRGAMMA\n");
  printf("\n");
  printf("		DEFAULT: GTRCAT\n");
  printf("\n");
  printf("	-n	Specifies the name of the output file.	\n"); 
  printf("\n"); 
  printf("	-s 	specify the name of the alignment data file in PHYLIP format\n");
  printf("\n");
  printf("	-t      Specify a user starting tree file name in Newick format\n");
  printf("\n");
  printf("	-v      Display version information\n");  
  printf("\n");
  printf("	-w 	Name of the working directory where RAxML will write its output files\n");
  printf("\n");
  printf("		DEFAULT: current directory\n"); 
  printf("\n");
  printf("	-y	If you want to only compute a parsimony starting tree with RAxML specify \"-y\",\n");
  printf("		the program will exit after computation of the starting tree\n");
  printf("\n");
  printf("		DEFAULT: OFF\n");
  printf("\n");  
  printf("	-#	Specify the number of alternative runs on distinct starting trees\n");
  printf("		In combination with the \"-b\" option, this will invoke a multiple boostrap analysis\n");
  printf("\n");
  printf("		DEFAULT: 1 single analysis\n");
  printf("\n\n\n\n");

}





void get_args(int argc, char *argv[], boolean print_usage, analdef *adef, rawdata *rdta)
{ /* get_args */
  int        c;
  boolean    bad_opt=FALSE;
  int        i, time, temperature;
  char       buf[2048];
  char       model[2048] = "";
  char       modelChar;
  int nameSet = 0, 
    alignmentSet = 0, 
    catSet = 0,
    checkpointSet = 0,
    startingTreeOnly = 0,
    multipleRuns = 0,
    parsimonySeed = 0;
  
  run_id[0] = 0;
  workdir[0] = 0;
  seq_file[0] = 0;
  tree_file[0] = 0;
  model[0] = 0;
  weightFileName[0] = 0;
    
  /*while(!bad_opt && ((c=getopt(argc,argv,"a:b:c:d:f:g:i:k:l:m:r:t:w:s:n:e:#:p:vxyjh"))!=-1) ) */
  while(!bad_opt && ((c=getopt(argc,argv,"a:b:c:f:i:m:t:w:s:n:#:p:vyjh"))!=-1) )
    {
    switch(c) 
      {
      case 'p':
	sscanf(optarg,"%d", &parsimonySeed);
	adef->parsimonySeed = parsimonySeed;
	break;
      case '#':
	sscanf(optarg,"%d", &multipleRuns);
	adef->multipleRuns = multipleRuns;
	break;
      case 'v':
	printVersionInfo();
	exit(1);
      case 'y':
	/*sscanf(optarg,"%d", &startingTreeOnly);*/
	adef->startingTreeOnly = 1;
	break;
      case 'h':
	printREADME();
	exit(1);
      case 'j':
	/*sscanf(optarg,"%d", &checkpointSet);
	  if(checkpointSet > 0)*/
	adef->checkpoints = 1;
	/*else
	  adef->checkpoints = 0;*/
	break;
      case 'a':
	strcpy(weightFileName,optarg);
	adef->useWeightFile = TRUE;
        break;            
      case 'b':
	sscanf(optarg,"%ld", &adef->boot);       
	break;
      case 'c':
	sscanf(optarg, "%d", &adef->categories);
	catSet = 1;
	break;	      
	/*case 'd':
	sscanf(optarg, "%d", &temperature);
	adef->t0 = (double)temperature;	
	break;*/
	/*case 'g':
	sscanf(optarg, "%d", &adef->repeats);
	break;*/
	/*case 'x':
	adef->rapidParsimony = 1;
	break;*/
      case 'f': 
	sscanf(optarg, "%c", &modelChar);
	switch(modelChar)
	  {
	  
	  case 'e': adef->mode = TREE_EVALUATION; break;	  
	    /* case 'h': adef->mode = HEAVY_MODE; break;  */
#ifdef PARALLEL_2
	    /*case 'p': adef->mode = PARALLEL_MODE; break;*/
#endif	 

	  case 'd': adef->mode = BIG_RAPID_MODE; break;	 
	    /*case 'g': adef->mode = BIG_SIMULATED_ANNEALING; break;*/
	  default: 
	    {
	      printf("Error select on of the following algorithms:\n");	     	     
	      printf("-f e : optimization of model parameters and branch lengths\n");	      
	      /*	      printf("-f h : hierarchical optimization mode\n");*/
	      printf("-f d : very fast hill-climbing for huge data\n");
	      /*	      printf("-f g : rapid simulated annealing mode\n");*/
	      exit(-1);
	    }
	  }
	break;      
      case 'i':
	sscanf(optarg, "%d", &adef->initial);
	adef->initialSet = TRUE;
	break;     
      case 'n':
        strcpy(run_id,optarg);
	nameSet = 1;
        break;
      case 'w':
        strcpy(workdir,optarg);
        break;                 
      case 't':
	strcpy(tree_file, optarg);
	adef->restart = TRUE;
	break;
	/*case 'r':
	sscanf(optarg, "%d", &adef->stepwidth);	
	if(adef->stepwidth < 2) adef->stepwidth = 2;
	break;    */
	/*case 'k':
	sscanf(optarg, "%d", &adef->max_rearrange);
	if(adef->max_rearrange < 5) adef->max_rearrange = 5;
	break;  */
      case 's':
	strcpy(seq_file, optarg);
	alignmentSet = 1;
	break;
      case 'm':
	strcpy(model,optarg);
	if(strcmp(model, "GTRGAMMA\0") == 0)
	  {
	    adef->model = M_GTRGAMMA;
	    goto l_1;
	  }
	if(strcmp(model, "GTRCAT\0") == 0)
	  {
	    adef->model = M_GTRCAT;
	    goto l_1;
	  }
	if(strcmp(model, "GTRMIX\0") == 0)
	  {
	    adef->model = M_GTRCAT;
	    adef->useMixedModel = 1;
	    goto l_1;
	  }
	printf("Model %s does not exist\n", model);
	printf("please use either GTRCAT or GTRGAMMA for DNA DATA\n");
	exit(-1);
      l_1:
	break;      
      default:
        bad_opt = TRUE;
        break;
    }    
  }     


  if(adef->restart && adef->startingTreeOnly)
    {
      printf("\n Error conflicting options: you want to compute only a parsimony starting tree with -y\n");
      printf(" while you actually specified a starting tree with -t %s\n", tree_file);
      exit(-1);
    }
  
  if(adef->mode == TREE_EVALUATION && (!adef->restart))
    {
      printf("\n Error: please specify a treefile for the tree you want to evaluate with -t\n");
      exit(-1);
    }

  if(adef->mode == TREE_EVALUATION && adef->model == M_GTRCAT)
    {
      printf("\n Error: No tree evaluation with GTRCAT possible\n");
      printf("the GTRCAT likelihood values are instable at present and should not\n");
      printf("be used to compare trees based on ML values\n");
      exit(-1);
    }

  if(!nameSet)
    {
      printf("\n Error: please specify a name for this run with -n\n");
      exit(-1);
    }
    
  if(! alignmentSet)
    {
      printf("\n Error: please specify an alignment for this run with -s\n");
      exit(-1);
    }


#ifdef WIN32
  if(workdir[0]==0 || (workdir[1] != ':' && workdir[2] != '\\'))
    {
      getcwd(buf,sizeof(buf));
      if( buf[strlen(buf)-1] != '\\') strcat(buf,"\\");
      strcat(buf,workdir);
      if( buf[strlen(buf)-1] != '\\') strcat(buf,"\\");
      strcpy(workdir,buf);
    }
#else
  if(workdir[0]==0 || workdir[0] != '/') 
    {
      getcwd(buf,sizeof(buf));
      if( buf[strlen(buf)-1] != '/') strcat(buf,"/");
      strcat(buf,workdir);
      if( buf[strlen(buf)-1] != '/') strcat(buf,"/");
      strcpy(workdir,buf);
    }
#endif

  if(bad_opt) 
    {      
      exit(-1);
    }
    return;
}

boolean str_treeReadLenSpecial (char *treestr, tree *tr)
    /* read string with representation of tree */
  { /* str_treeReadLen */
    nodeptr  p;
    int  i;
    boolean  is_fact, found;

    for (i = 1; i <= tr->mxtips; i++) tr->nodep[i]->back = (node *) NULL;
    tr->start       = tr->nodep[tr->mxtips];
    tr->ntips       = 0;
    tr->nextnode    = tr->mxtips + 1;
    tr->opt_level   = 0;
    tr->smoothed    = FALSE;
    tr->rooted      = FALSE;


    is_fact = str_processTreeCom(tr,&treestr);
    p = tr->nodep[(tr->nextnode)++];

    if(!str_treeNeedCh(&treestr, '(', "at start of"))       return FALSE;
    if(!str_addElementLen(&treestr, tr, p))                 return FALSE;
    if(!str_treeNeedCh(&treestr, ',', "in"))                return FALSE;
    if(!str_addElementLen(&treestr, tr, p->next))           return FALSE;
    if (! tr->rooted) 
      { 
	if(str_treeGetCh(&treestr) == ',')
	  {
	    if (! str_addElementLen(&treestr, tr, p->next->next)) return FALSE;
	  }
	else
	  {
	    tr->rooted = TRUE;
	    if(*treestr != '\n') treestr--;
	  }
      }
    else
      {
	p->next->next->back = (nodeptr) NULL;
      }
      
    if(!str_treeNeedCh(&treestr, ')', "in"))                 return FALSE;
    str_treeFlushLabel(&treestr);           
    if(!str_treeFlushLen(&treestr))                          return FALSE;
    if(is_fact) {
      if(!str_treeNeedCh(& treestr, ')', "at end of"))       return FALSE;
      if(!str_treeNeedCh(& treestr, '.', "at end of"))       return FALSE;
    }
    else 
      {
	if(!str_treeNeedCh(& treestr, ';', "at end of"))       return FALSE;
      }

    if(tr->rooted)
      {
	p->next->next->back = (nodeptr) NULL;
	tr->start =  uprootTree(tr, p->next->next);
	if(!tr->start) return FALSE;
	}
    else
      {
	tr->start = p->next->next->back;
      }

    return  (initrav(tr,tr->start) && initrav(tr,tr->start->back));
  } /* str_treeReadLen */



void optimizeAlpha(tree *tr)
{
 int k;
 
 double currentTT = tr->alpha;
 double startLikelihood = tr->likelihood;
 double maxLikelihoodMinus = tr->likelihood;
 double maxLikelihoodPlus = tr->likelihood;
 double maxTTPlus = currentTT;
 double maxTTMinus = currentTT;
 double spacing = 0.5/optimizeTTinvocations;
 double tree_likelihood = startLikelihood;
 optimizeAlphaInvocations++;
      
  k = 1;
      
  while(tree_likelihood >= maxLikelihoodMinus && (currentTT - spacing * k) > 0) 
    {      
      tr->alpha = currentTT - spacing * k;     
      makeGammaCats(tr);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);            

      tree_likelihood = tr->likelihood;
     
      if(tr->likelihood > maxLikelihoodMinus)
	{
	  maxLikelihoodMinus = tr->likelihood;
	  maxTTMinus = currentTT - spacing * k;
	}    

      k++;
    }
  
  k = 1;
  tree_likelihood = startLikelihood;
  while(tree_likelihood >= maxLikelihoodPlus) 
    {      
      tr->alpha = currentTT + spacing * k;     
      makeGammaCats(tr);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);         

      tree_likelihood = tr->likelihood;

      if(tr->likelihood > maxLikelihoodPlus)
	{
	  maxLikelihoodPlus = tr->likelihood;
	  maxTTPlus = currentTT + spacing * k;
	}	 
      k++;
    }
  
  if(maxLikelihoodPlus > startLikelihood || maxLikelihoodMinus > startLikelihood)
    {
      if(maxLikelihoodPlus > maxLikelihoodMinus)
	{
	  tr->alpha = maxTTPlus;
	}
      else
	{
	  tr->alpha = maxTTMinus;
	}     
      makeGammaCats(tr);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);      
    }
  else
    { 
      tr->alpha = currentTT;
      makeGammaCats(tr);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);      
      evaluate(tr, tr->start);
    }    
}

void optimizeTT(tree *tr)
{
  int k;
 
  double currentTT = tr->t_value;
  double startLikelihood = tr->likelihood;
  double maxLikelihoodMinus = tr->likelihood;
  double maxLikelihoodPlus = tr->likelihood;
  double maxTTPlus = currentTT;
  double maxTTMinus = currentTT;
  double spacing = 0.5/optimizeTTinvocations;
  double tree_likelihood = startLikelihood;
  optimizeTTinvocations++;
      
  k = 1;
      
  while(tree_likelihood >= maxLikelihoodMinus && (currentTT - spacing * k) > 0) 
    {
      
      alterTT(tr, currentTT - spacing * k);     
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);
      
      tree_likelihood = tr->likelihood;
      if(tr->likelihood > maxLikelihoodMinus)
	{
	  maxLikelihoodMinus = tr->likelihood;
	  maxTTMinus = currentTT - spacing * k;
	}
      k++;
    }
  
  k = 1;
  tree_likelihood = startLikelihood;
  while(tree_likelihood >= maxLikelihoodPlus) 
    {      
      alterTT(tr, currentTT + spacing * k);     
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);      

      tree_likelihood = tr->likelihood;
      if(tr->likelihood > maxLikelihoodPlus)
	{
	  maxLikelihoodPlus = tr->likelihood;
	  maxTTPlus = currentTT + spacing * k;
	}	 
      k++;
    }
  
  if(maxLikelihoodPlus > startLikelihood || maxLikelihoodMinus > startLikelihood)
    {
      if(maxLikelihoodPlus > maxLikelihoodMinus)
	{
	  alterTT(tr, maxTTPlus);
	}
      else
	{
	  alterTT(tr, maxTTMinus);
	}     
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);
    }
  else
    { 
      alterTT(tr, currentTT);	
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);      
      evaluate(tr, tr->start);
    }
}


void baseFrequenciesGTR(rawdata *rdta, cruncheddata *cdta)
{
  double  sum, suma, sumc, sumg, sumt, wj, fa, fc, fg, ft, freqa, freqc, freqg, freqt;
  int     i, j, k, code;
  char  *yptr;

  freqa = 0.25;
  freqc = 0.25;
  freqg = 0.25;
  freqt = 0.25;
  for (k = 1; k <= 8; k++) 
    {
      suma = 0.0;
      sumc = 0.0;
      sumg = 0.0;
      sumt = 0.0;
      for (i = 0; i < rdta->numsp; i++) 
	{
	  yptr = rdta->y[i];
	  for (j = 0; j < cdta->endsite; j++) 
	    {
	      code = *yptr++;
	      fa = freqa * ( code       & 1);
	      fc = freqc * ((code >> 1) & 1);
	      fg = freqg * ((code >> 2) & 1);
	      ft = freqt * ((code >> 3) & 1);
	      wj = cdta->aliaswgt[j] / (fa + fc + fg + ft);
	      suma += wj * fa;
	      sumc += wj * fc;
	      sumg += wj * fg;
	      sumt += wj * ft;
	    }
        }
      sum = suma + sumc + sumg + sumt;
      freqa = suma / sum;
      freqc = sumc / sum;
      freqg = sumg / sum;
      freqt = sumt / sum;
    }

  rdta->freqa = freqa;
  rdta->freqc = freqc;
  rdta->freqg = freqg;
  rdta->freqt = freqt;
       
  return;
}

void initReversibleGTR(rawdata *rdta, cruncheddata *cdta, tree *tr)
{
  /*printf("INIT\n");*/
  
  double r[4][4], a[4][4], *initialRates = rdta->initialRates, 
    f[4], e[4], d[4];
    int i, j, k, m;
    
    f[0] = rdta->freqa;
    f[1] = rdta->freqc;
    f[2] = rdta->freqg;
    f[3] = rdta->freqt;

    i = 0;
    for (j = 0; j < 2; j++)
      for (k = j+1; k< 4; k++)
	r[j][k] = initialRates[i++];
    r[2][3] = 1;
    for (j = 0; j < 4; j++) 
	{
	  r[j][j] = 0;
	  for (k = 0; k < j; k++)
	    r[j][k] = r[k][j];
	}
        
    rdta->fracchange = 0;
    for (j = 0; j<4; j++)
      for (k = 0; k<4; k++)
	rdta->fracchange += f[j] * r[j][k] * f[k];
           
    m = 0;
    
      for(i=0; i<4; i++) 
	a[i][i] = 0;
      
      for(i=0; i<4; i++) 
	{
	  for(j=i+1; j<4; j++) 
	    {
	      double factor = ((m>=5) ? 1 : initialRates[m++]);
	      a[i][j] = a[j][i] = factor * sqrt( f[i] * f[j]);
	      a[i][i] -= factor * f[j];
	      a[j][j] -= factor * f[i];
	    }
	}
      
 

      tred2((double *)a,4,4,d,e);       
      tqli(d, e, 4 , 4, (double *)a);
           
      for(i=0; i<4; i++) 
	{
	  for(j=0; j<4; j++) 
	    {
	      a[i][j] *= sqrt(f[j]);
	    }
	}    


     
      for (i=0; i<4; i++)
	{	  
	  if (d[i] > -1e-8) 
	    {
	      
	      if (i != 0) 
		{		    
		  double tmp = d[i], sum=0;
		  d[i] = d[0];
		  d[0] = tmp;
		  for (j=0; j < 4; j++) 
		    {
		      tmp = a[i][j];
		      a[i][j] = a[0][j];
		      sum += (a[0][j] = tmp);
		    }
		  for (j=0; j < 4; j++) 
		    a[0][j] /= sum;
		}
	      break;
	    }
	}
      for (i=0; i< 4; i++) 
	{
	  rdta->EIGN[0][i] = -d[i];
	  for (j=0; j<4; j++)
	    rdta->EIGV[0][i][j] = a[j][i];
	}
     
      rdta->EIGN[1][0] = rdta->EIGN[0][0];
      rdta->EIGN[1][1] = rdta->EIGN[0][1];
      rdta->EIGN[1][2] = rdta->EIGN[0][2];
      rdta->EIGN[1][3] = rdta->EIGN[0][3];

     
      rdta->matread = TRUE;
      
      for (i = 0; i<4; i++) 
      {
	rdta->invfreq[0][i] = 1 / rdta->EIGV[0][i][0];	
      }

      {     	
	double (*EIGV)[4] = &(rdta->EIGV[0][0]);
	double *invfreq = &(rdta->invfreq[0][0]);
	double *EIGN = &(tr->rdta->EIGN[0][0]);
	double *eptr = tr->EV;    
	double *EI = tr->EI;
	int i, j;
	for(i = 0; i < 4; i++)
	  for(j = 0; j < 4; j++)
	    {
	      *eptr++ = EIGV[i][j];	    
	    }       


	EI[0] = tr->EV[1] * invfreq[0];
       	EI[1] = tr->EV[2] * invfreq[0];
	EI[2] = tr->EV[3] * invfreq[0];

	EI[3] = tr->EV[5] * invfreq[1];
	EI[4] = tr->EV[6] * invfreq[1];
	EI[5] = tr->EV[7] * invfreq[1];

	EI[6] = tr->EV[9]  * invfreq[2];
        EI[7] = tr->EV[10] * invfreq[2];
        EI[8] = tr->EV[11] * invfreq[2];

        EI[9] = tr->EV[13] * invfreq[3];
	EI[10] = tr->EV[14] * invfreq[3];
        EI[11] = tr->EV[15] * invfreq[3];	  
      }
      

      /*invfreq[4], EIGV[4][4], EIGN[4]*/

}








void alterRates(tree *tr, int k, analdef *adef)
{
  int i, increasing;
  double granularity = 0.1/((double)optimizeRatesInvocations);
  double bestLikelihood = tr->likelihood, 
    maxLikelihoodMinus = tr->likelihood, maxLikelihoodPlus = tr->likelihood,
    treeLikelihood, originalRate, maxRateMinus, maxRatePlus;
 
  
  i = 1;
  originalRate = tr->rdta->initialRates[k];
  maxRateMinus = maxRatePlus = originalRate;
  treeLikelihood = bestLikelihood;


  while(treeLikelihood >= maxLikelihoodMinus && (originalRate - granularity * i) > 0)
    {
      tr->rdta->initialRates[k] = originalRate - granularity * i;
      initReversibleGTR(tr->rdta, tr->cdta, tr);
      initGlobalTip(tr, adef);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);      
      treeLikelihood = tr->likelihood;
      if(tr->likelihood > maxLikelihoodMinus)
	{
	  increasing = 1;
	  maxLikelihoodMinus = tr->likelihood;
	  maxRateMinus = originalRate - granularity * i;
	}      
      i++;
    }

  i = 1;
  treeLikelihood = bestLikelihood;
  tr->rdta->initialRates[k] = originalRate;
  while(treeLikelihood >= maxLikelihoodPlus)
    {
      tr->rdta->initialRates[k] = originalRate + i * granularity;
      initReversibleGTR(tr->rdta, tr->cdta, tr);
      initGlobalTip(tr, adef);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);      
      treeLikelihood = tr->likelihood;
      if(tr->likelihood > maxLikelihoodPlus)
	{
	  increasing = 1;
	  maxLikelihoodPlus = tr->likelihood;
	  maxRatePlus = originalRate + granularity * i;
	}      
      i++;
    }
  if(maxLikelihoodPlus > bestLikelihood || maxLikelihoodMinus > bestLikelihood)
    {
      if(maxLikelihoodPlus > maxLikelihoodMinus)
	{	  
	  tr->rdta->initialRates[k] = maxRatePlus;
	  initReversibleGTR(tr->rdta, tr->cdta, tr);	  
	}
      else
	{
	 tr->rdta->initialRates[k] = maxRateMinus;
	 initReversibleGTR(tr->rdta, tr->cdta, tr);	  	
	}     

      initGlobalTip(tr, adef);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);     
    }
  else
    {      
      tr->rdta->initialRates[k] = originalRate;
      initReversibleGTR(tr->rdta, tr->cdta, tr);      
      initGlobalTip(tr, adef);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);     
    }

  
}

void optimizeRates(tree *tr, analdef *adef)
{
  int i;
  double initialLikelihood, improvedLikelihood;
 
  for(i = 0; i < 5; i++)
    {	      
      alterRates(tr, i, adef);     
      
    }

  optimizeRatesInvocations++;
}



void printModelAndProgramInfo(tree *tr, analdef *adef, int argc, char *argv[])
{ 
  int i;
  
  infoFile = fopen(infoFileName, "w");

  if(adef->repeats == 0)
    adef->repeats = tr->mxtips;


  switch(adef->mode)
     {       
     case TREE_EVALUATION : 
       printf("RAxML Model Optimization\n\n");
       fprintf(infoFile, "RAxML Model Optimization\n\n");
       break;     
     case  BIG_RAPID_MODE:
       printf("RAxML rapid hill-climbing mode for huge data\n\n"); 
       fprintf(infoFile, "RAxML   rapid hill-climbing mode for huge data\n\n");
       break;
     }
    
  if(adef->boot && adef->multipleRuns == 0)
    {
      printf("Executing ONE non-parametric bootstrap inference\n\n");
      fprintf(infoFile, "Executing ONE non-parametric bootstrap inference\n\n");
    }

  if(adef->boot && adef->multipleRuns > 0)
    {
      printf("Executing %d non-parametric bootstrap inferences\n\n", adef->multipleRuns);
      fprintf(infoFile, "Executing %d non-parametric bootstrap inferences\n\n", adef->multipleRuns);
    }


  if(adef->boot == 0 && adef->multipleRuns > 0)
    {
      printf("Executing %d inferences on the originial alignment using %d distinct randomized MP starting trees\n\n", adef->multipleRuns, adef->multipleRuns);
      fprintf(infoFile, "Executing %d inferences on the originial alignment using %d distinct randomized MP starting trees\n\n", adef->multipleRuns, adef->multipleRuns);
    }


  switch(adef->model)
    {
    
    case M_GTRCAT:
       if(adef->useMixedModel)
	 { 
	   printf( "GTR mixed model of nucleotide substitution\n"); 
	   printf( "All model parameters will be estimated by RAxML\n");
	   printf( "ML estimate of %d per site rate categories\n", adef->categories);      
	   printf( "Likelihood of final tree will be evaluated and optimized under GTR+GAMMA\n");
	   fprintf(infoFile, "GTR mixed model of nucleotide substitution\n"); 
	   fprintf(infoFile, "All model parameters will be estimated by RAxML\n");
	   fprintf(infoFile, "ML estimate of %d per site rate categories\n", adef->categories);
	   fprintf(infoFile, "Likelihood of final tree will be evaluated and optimized under GTR+GAMMA\n");
	 }
       else
	 {
	   printf( "GTR approximation of nucleotide substitution\n"); 
	   printf( "All model parameters will be estimated by RAxML\n");
	   printf( "ML estimate of %d per site rate categories\n", adef->categories);   
	   printf( "WARNING: GTR+CAT likelihood values are numerically instable!\n");
	   printf( "WARNING: For ML-based comparisons of final trees use -m GTRMIX\n");
	   fprintf(infoFile, "GTR approximation of nucleotide substitution\n"); 
	   fprintf(infoFile, "All model parameters will be estimated by RAxML\n");
	   fprintf(infoFile, "ML estimate of %d per site rate categories\n", adef->categories);
	   fprintf(infoFile, "WARNING: GTR+CAT likelihood values are numerically instable!\n");
	   fprintf(infoFile, "WARNING: For ML-based comparisons of final trees use -m GTRMIX\n");
	 }
       break;   
    case M_GTRGAMMA:
      printf( "GTR model of nucleotide substitution\n");  
      printf( "All model parameters will be estimated by RAxML\n");  
      printf( "GAMMA model of rate heteorgeneity, ML estimate of alpha-parameter\n");
      fprintf(infoFile, "GTR model of nucleotide substitution\n");  
      fprintf(infoFile, "All model parameters will be estimated by RAxML\n");  
      fprintf(infoFile, "GAMMA model of rate heterogeneity, ML estimate of alpha-parameter\n");
      break;

    }
  

  if(adef->boot == 0)
    {
      printf("Empirical Base Frequencies:\n");
      
      printf("pi(A): %f ",   tr->rdta->freqa);
      printf("pi(C): %f ",   tr->rdta->freqc);
      printf("pi(G): %f ",   tr->rdta->freqg);
      printf("pi(T): %f \n",   tr->rdta->freqt);  
                 
      fprintf(infoFile, "Empirical Base Frequencies:\n");
      
      fprintf(infoFile, "pi(A): %f ",   tr->rdta->freqa);
      fprintf(infoFile, "pi(C): %f ",   tr->rdta->freqc);
      fprintf(infoFile, "pi(G): %f ",   tr->rdta->freqg);
      fprintf(infoFile, "pi(T): %f \n",   tr->rdta->freqt);  
    }
      
  printf("\n\n");
  fprintf(infoFile, "\n\n");
      
  fprintf(infoFile,"RAxML was called as follows:\n\n");
  for(i = 0; i < argc; i++)
    fprintf(infoFile,"%s ", argv[i]);
  fprintf(infoFile,"\n\n\n");  

  fclose(infoFile);
}










static int infoCompare(const void *p1, const void *p2)
{
 bestInfo *rc1 = (bestInfo *)p1;
 bestInfo *rc2 = (bestInfo *)p2;

 int i = rc1->node;
  int j = rc2->node;
  
  if (i > j)
    return (1);
  if (i < j)
    return (-1);
  return (0);
}


void sortInfoList()
{

  qsort(&(iList.list[0]), iList.n, sizeof(bestInfo), infoCompare);
}

void resetInfoList()
{
  int i;

  iList.valid = 0;

  for(i = 0; i < iList.n; i++)
    {
      iList.list[i].node = 0;
      iList.list[i].likelihood = unlikely;
    }
}

void initInfoList(int n)
{
  int i;

  iList.n = n;
  iList.valid = 0;
  iList.list = (bestInfo *)malloc(sizeof(bestInfo) * n);

  for(i = 0; i < n; i++)
    {
      iList.list[i].node = 0;
      iList.list[i].likelihood = unlikely;
    }
}

void freeInfoList()
{ 
  free(iList.list);   
}

void insertInfoList(int node, double likelihood)
{
  int i;
  int min = 0;
  double min_l =  iList.list[0].likelihood;

  for(i = 1; i < iList.n; i++)
    {
      if(iList.list[i].likelihood < min_l)
	{
	  min = i;
	  min_l = iList.list[i].likelihood;
	}
    }

  if(likelihood > min_l)
    {
      iList.list[min].likelihood = likelihood;
      iList.list[min].node = node;
      iList.valid += 1;
    }
  
  if(iList.valid > iList.n)
    iList.valid = iList.n;

}

void printInfoList()
{
  int i;
  
  for(i = 0; i < iList.n; i++)
    {
      printf("L : %f Node : %d\n", iList.list[i].likelihood, iList.list[i].node);
    }
  
}


boolean smoothRegion (tree *tr, nodeptr p, int region)
  { /* smooth */
    nodeptr  q;
   
    if (! update(tr, p))               return FALSE; /*  Adjust branch */
    if(region > 0)
      {
	if (! p->tip) 
	  {                                 
	    q = p->next;
	    while (q != p) 
	      {
		if (! smoothRegion(tr, q->back, --region))   return FALSE;
		q = q->next;
	      }	
	    if (! newview(tr, p)) return FALSE;	
	  }
      }
    
    return TRUE;
  } /* smooth */

boolean regionalSmooth (tree *tr, nodeptr p, int maxtimes, int region)
  {
    nodeptr  q;

    if (p->tip) return FALSE;            /* Should be an error */

    while (--maxtimes >= 0) 
      {
	tr->smoothed = TRUE;
	q = p;
	do 
	  {
	    if (! smoothRegion(tr, q, region)) return FALSE;
	    q = q->next;
	  } 
	while (q != p);
	if (tr->smoothed)  break;
      }

    tr->smoothed = FALSE;             /* Only smooth locally */
    return TRUE;
  } /* localSmooth */



nodeptr  removeNodeBIG (tree *tr, nodeptr p)
{
  double   zqr;
  nodeptr  q, r;
        
  q = p->next->back;
  r = p->next->next->back;
  zqr = q->z * r->z;        

  if ((zqr = makenewz(tr, q, r, zqr, iterations)) == badZ) return (node *) NULL;      

  tr->zqr = zqr;
    
  hookup(q, r, zqr);

  p->next->next->back = p->next->back = (node *) NULL;
     
  return  q;
}

nodeptr  removeNodeRestoreBIG (tree *tr, nodeptr p)
{
  nodeptr  q, r;
        
  q = p->next->back;
  r = p->next->next->back;  

  newview(tr, q);
  newview(tr, r);
  
  hookup(q, r, tr->currentZQR);

  p->next->next->back = p->next->back = (node *) NULL;
     
  return  q;
}


boolean insertBIG (tree *tr, nodeptr p, nodeptr q, boolean glob)
{
  nodeptr  r, s;
  
  r = q->back;
  s = p->back;

  tr->lzi = q->z;

  if(Thorough)
    { 
      double  zqr, zqs, zrs, lzqr, lzqs, lzrs, lzsum, lzq, lzr, lzs, lzmax;      
            
      if ((zqr = makenewz(tr, q, r, q->z, iterations)) == badZ) return FALSE; 
      
      if ((zqs = makenewz(tr, q, s, defaultz, iterations)) == badZ) return FALSE;      
      
      if ((zrs = makenewz(tr, r, s, defaultz, iterations)) == badZ) return FALSE;
      
      lzqr = (zqr > zmin) ? log(zqr) : log(zmin); 
      lzqs = (zqs > zmin) ? log(zqs) : log(zmin);
      lzrs = (zrs > zmin) ? log(zrs) : log(zmin);
      lzsum = 0.5 * (lzqr + lzqs + lzrs);
      
      lzq = lzsum - lzrs;
      lzr = lzsum - lzqs;
      lzs = lzsum - lzqr;
      lzmax = log(zmax);
      
      if      (lzq > lzmax) {lzq = lzmax; lzr = lzqr; lzs = lzqs;} 
      else if (lzr > lzmax) {lzr = lzmax; lzq = lzqr; lzs = lzrs;}
      else if (lzs > lzmax) {lzs = lzmax; lzq = lzqs; lzr = lzrs;}          
            
      hookup(p->next,       q, exp(lzq));
      hookup(p->next->next, r, exp(lzr));
      hookup(p,             s, exp(lzs));      		  
    }
  else
    { 
      double  z = sqrt(q->z);      

      if(z < zmin) 
	z = zmin;
      if(z > zmax)
	z = zmax;

      hookup(p->next,       q, z);
      hookup(p->next->next, r, z);	         
    }   
    
  if (! newview(tr, p)) return FALSE;    

  if(Thorough)
    {     
      localSmooth(tr, p, smoothings);

      /* regionalSmooth() for special ? */

      /* Did not show to be more efficient */
      tr->lzq = p->next->z;
      tr->lzr = p->next->next->z;
      tr->lzs = p->z;            
    }           
     
  return  TRUE;
}

boolean insertRestoreBIG (tree *tr, nodeptr p, nodeptr q, boolean glob)
{
  nodeptr  r, s;
  
  r = q->back;
  s = p->back;

  if(Thorough)
    {                        
      hookup(p->next,       q, tr->currentLZQ);
      hookup(p->next->next, r, tr->currentLZR);
      hookup(p,             s, tr->currentLZS);      		  
    }
  else
    { 
      
      double  z;
      
      z = sqrt(q->z);     
      if(z < zmin) 
	z = zmin;
      if(z > zmax)
	z = zmax;

      hookup(p->next,       q, z);
      hookup(p->next->next, r, z);
    }   
    
  if (! newview(tr, p)) return FALSE;    
       
  return  TRUE;
}


void restoreTopologyOnly(tree *tr)
{
  nodeptr p = tr->removeNode;
  nodeptr q = tr->insertNode;
  double qz, pz, p1z, p2z;
  nodeptr p1, p2, r, s;
  double currentLH = tr->likelihood;

  p1 = p->next->back;
  p2 = p->next->next->back;

  p1z = p1->z;
  p2z = p2->z;
  
  hookup(p1, p2, tr->currentZQR);

  p->next->next->back = p->next->back = (node *) NULL;  

  /* remove */

  /* testInsert */

  qz = q->z;
  pz = p->z;

  /* testInsert */

  /* insert */

  r = q->back;
  s = p->back;
    
 

  if(Thorough)
    {                        
      hookup(p->next,       q, tr->currentLZQ);
      hookup(p->next->next, r, tr->currentLZR);
      hookup(p,             s, tr->currentLZS);      		  
    }
  else
    { 	
      double  z;	
      z = sqrt(q->z);      
      if(z < zmin)
	z = zmin;
      if(z > zmax)
	z = zmax;
      hookup(p->next,       q, z);
      hookup(p->next->next, r, z);
    }     

  tr->likelihood = tr->bestOfNode;
   
 
  saveBestTree(bt, tr);
 
  tr->likelihood = currentLH;

  hookup(q, r, qz);

  p->next->next->back = p->next->back = (nodeptr) NULL;

  if(Thorough)
    {      
      hookup(p, s, pz);      
    }  

   hookup(p->next,       p1, p1z); 
   hookup(p->next->next, p2, p2z);      
}


boolean testInsertBIG (tree *tr, nodeptr p, nodeptr q)
{
  double  qz, pz;
  nodeptr  r;
  
  r = q->back; 
  qz = q->z;
  pz = p->z;
  
  
  if (! insertBIG(tr, p, q, FALSE))       return FALSE;         
  
  if (evaluate(tr, p->next->next) == badEval) return FALSE;   
   
  if(tr->likelihood > tr->bestOfNode)
    {
      tr->bestOfNode = tr->likelihood;
      tr->insertNode = q;
      tr->removeNode = p;   
      tr->currentZQR = tr->zqr;           
      tr->currentLZR = tr->lzr;
      tr->currentLZQ = tr->lzq;
      tr->currentLZS = tr->lzs;      
    }
  
  if(tr->likelihood > tr->endLH)
    {	           
      tr->insertNode = q;
      tr->removeNode = p;   
      tr->currentZQR = tr->zqr;      
      tr->endLH = tr->likelihood;           
      
      /*printf("IMP %f -> %f, %f, %f\n", tr->lzi, tr->lzq, tr->lzr, (tr->lzq + tr->lzr) / 2.0);
	printf("IMP2 %f -> %f, %f, %f, Prop: %f\n", exp(-tr->lzi), exp(-tr->lzq), exp(-tr->lzr), (exp(-tr->lzq) + exp(-tr->lzr)) / 2.0, -log(sqrt(exp(-tr->lzi))));*/

      /*tr->currentLZR = tr->lzr;
	tr->currentLZQ = tr->lzq;
	tr->currentLZS = tr->lzs;*/
    }    
 
  hookup(q, r, qz);
  
  p->next->next->back = p->next->back = (nodeptr) NULL;

  if(Thorough)
    {
      nodeptr s = p->back;
      hookup(p, s, pz);      
    } 
  
  return TRUE;
}



 
void addTraverseBIG(tree *tr, nodeptr p, nodeptr q, int mintrav, int maxtrav)
{  
  if (--mintrav <= 0) 
    {              
      if (! testInsertBIG(tr, p, q))  return;        
    }
  
  if ((! q->tip) && (--maxtrav > 0)) 
    {    
      addTraverseBIG(tr, p, q->next->back, mintrav, maxtrav);
      addTraverseBIG(tr, p, q->next->next->back, mintrav, maxtrav);    
    }
} 


int  rearrangeBIGEXP(tree *tr, nodeptr p, int mintrav, int maxtrav)
   
  {
    double   p1z, p2z, q1z, q2z;
    nodeptr  p1, p2, q, q1, q2;
    int      mintrav2;
    int result = 0;

    double iniLH;
      
    if (maxtrav < 1 || mintrav > maxtrav)  return 0;
    q = p->back;

    if (! p->tip) 
      {     
	p1 = p->next->back;
	p2 = p->next->next->back;

	if (! p1->tip || ! p2->tip) 
	  {
	    p1z = p1->z;
	    p2z = p2->z;
	   
	    iniLH = tr->endLH;

	    if (! removeNodeBIG(tr, p)) return badRear;
	  
	    if (! p1->tip) 
	      {
		addTraverseBIG(tr, p, p1->next->back,
			       mintrav, maxtrav);         
		addTraverseBIG(tr, p, p1->next->next->back,
			       mintrav, maxtrav);          
	      }

	    if (! p2->tip) 
	      {
		addTraverseBIG(tr, p, p2->next->back,
			       mintrav, maxtrav);
		addTraverseBIG(tr, p, p2->next->next->back,
			       mintrav, maxtrav);          
	      }
	    
	    if(! (tr->endLH > iniLH))
	      {	
		hookup(p->next,       p1, p1z); 
		hookup(p->next->next, p2, p2z);	   	    	    
		initravDIST(tr, p, 1);	   
	      }
	  }
      }  
          
    return  1;
  } 


int  rearrangeBIG(tree *tr, nodeptr p, int mintrav, int maxtrav)
   
  {
    double   p1z, p2z, q1z, q2z;
    nodeptr  p1, p2, q, q1, q2;
    int      mintrav2;
    int result = 0;

    double iniLH;
      
    if (maxtrav < 1 || mintrav > maxtrav)  return 0;
    q = p->back;

    if (! p->tip) 
      {     
	p1 = p->next->back;
	p2 = p->next->next->back;

	if (! p1->tip || ! p2->tip) 
	  {
	    p1z = p1->z;
	    p2z = p2->z;
	   
	    iniLH = tr->endLH;

	    if (! removeNodeBIG(tr, p)) return badRear;
	  
	    if (! p1->tip) 
	      {
		addTraverseBIG(tr, p, p1->next->back,
			       mintrav, maxtrav);         
		addTraverseBIG(tr, p, p1->next->next->back,
			       mintrav, maxtrav);          
	      }

	    if (! p2->tip) 
	      {
		addTraverseBIG(tr, p, p2->next->back,
			       mintrav, maxtrav);
		addTraverseBIG(tr, p, p2->next->next->back,
			       mintrav, maxtrav);          
	      }
	    
	   
	    hookup(p->next,       p1, p1z); 
	    hookup(p->next->next, p2, p2z);	   	    	    
	    initravDIST(tr, p, 1);	   	    
	  }
      }  

   
    
    if (! q->tip && maxtrav > 0) 
      {
	q1 = q->next->back;
	q2 = q->next->next->back;

	if (! q1->tip && (!q1->next->back->tip || !q1->next->next->back->tip) ||
	    ! q2->tip && (!q2->next->back->tip || !q2->next->next->back->tip)) 
	  {
	    q1z = q1->z;
	    q2z = q2->z;

	    if (! removeNodeBIG(tr, q)) return badRear;
	  
	    mintrav2 = mintrav > 2 ? mintrav : 2;

	    if (! q1->tip) 
	      {
		addTraverseBIG(tr, q, q1->next->back,
			       mintrav2 , maxtrav);
		addTraverseBIG(tr, q, q1->next->next->back,
			       mintrav2 , maxtrav);         
	      }

	    if (! q2->tip) 
	      {
		addTraverseBIG(tr, q, q2->next->back,
			       mintrav2 , maxtrav);
		addTraverseBIG(tr, q, q2->next->next->back,
			       mintrav2 , maxtrav);          
	      }	   
	   
	    hookup(q->next,       q1, q1z); 
	    hookup(q->next->next, q2, q2z);
	   
	    initravDIST(tr, q, 1); 	   
	  }
      } 

    return  1;
  } 


boolean testInsertRestoreBIG (tree *tr, nodeptr p, nodeptr q)
  { 
    nodeptr  r;

    r = q->back; 

    if(Thorough)
      {
        if (! insertBIG(tr, p, q, FALSE))       return FALSE;    
	if (evaluate(tr, p->next->next) == badEval) return FALSE;

	/*	
	  if (! insertRestoreBIG(tr, p, q, FALSE))       return FALSE;
    
	  {
	  nodeptr x, y;
	  
	  x = p->next->next;
	  y = p->back;
	  while ((! x->x) || (! y->x)) 
	  {
	  if (! (x->x)) if (! newview(tr, x)) return badEval;
	  if (! (y->x)) if (! newview(tr, y)) return badEval;
	  }
	  }
	  
	  tr->likelihood = tr->endLH;
	*/

      }
    else
      {
	if (! insertRestoreBIG(tr, p, q, FALSE))       return FALSE;
    
	{
	  nodeptr x, y;
	  x = p->next->next;
	  y = p->back;
	 
	

	  if(!x->tip && y->tip)
	    {
	      while ((! x->x)) 
		{
		  if (! (x->x)) if (! newview(tr, x)) return badEval;		     
		}
	    }
	  if(x->tip && !y->tip)
	    {
	      while ((! y->x)) 
		{		  
		  if (! (y->x)) if (! newview(tr, y)) return badEval;
		}
	    }
	  if(!x->tip && !y->tip)
	    {
	      while ((! x->x) || (! y->x)) 
		{
		  if (! (x->x)) if (! newview(tr, x)) return badEval;
		  if (! (y->x)) if (! newview(tr, y)) return badEval;
		}
	    }				      	
	  
	}
	
	tr->likelihood = tr->endLH;
      }
     
    return TRUE;
  } /* testInsert */

void restoreTreeFast(tree *tr)
{
  removeNodeRestoreBIG(tr, tr->removeNode);    
  testInsertRestoreBIG(tr, tr->removeNode, tr->insertNode);
}



void restoreTreeFastEXP(tree *tr)
{
  testInsertRestoreBIG(tr, tr->removeNode, tr->insertNode);
}




int treeOptimizeBIG(tree *tr, int _mintrav, int _maxtrav, analdef *adef)
{
  int i;
  int mintrav = _mintrav;
  int maxtrav = _maxtrav;
  int subsequent = 0;
  int initial = adef->initial;
  double lh_res;
  double optTime;
  int oldThorough;
  int impCount = 0;

  /* NEW */

  nodeptr *list = (nodeptr *)malloc(sizeof(nodeptr) * 3 * tr->mxtips);
  int n = 0;

  for(i = 1; i <= tr->mxtips + tr->mxtips - 2; i++)
    {    
      if(tr->nodep[i]->tip)
	list[n++] = tr->nodep[i]->back;
      else
	{
	  list[n++] = tr->nodep[i];
	  if(!tr->nodep[i]->back->tip)
	    list[n++] = tr->nodep[i]->back;
	}
    }
      /*NEW*/

  optTime = gettime();
  
  if (maxtrav > tr->ntips - 3)  
    maxtrav = tr->ntips - 3;  
  
 
  resetBestTree(bt);
  saveBestTree(bt, tr);

  if(Multiple && !Thorough)
    {
      mintrav = 1;       
      if(maxtrav < initial)	
	maxtrav = initial;
    }
 
  tr->startLH = tr->endLH = tr->likelihood;

  lh_res = tr->likelihood;

  /*for(i = 1; i <= tr->mxtips + tr->mxtips - 2; i++)*/
  for(i = 0; i < n; i++)
    {           
     
      tr->bestOfNode = unlikely;
     
      /*rearrangeBIG(tr, tr->nodep[i], mintrav, maxtrav);*/
      rearrangeBIG(tr, list[i], mintrav, maxtrav);
           
      if(Multiple && (tr->endLH > tr->startLH))
	{		 
	  restoreTreeFast(tr);	        	  	 	  	      
	  tr->startLH = tr->endLH = tr->likelihood;	  	 	  	  	  	  	  
	  subsequent = 1;
	  saveBestTree(bt, tr);	 
	}	    
      else
	{	 
	  if(tr->bestOfNode != unlikely)
	    {	     
	      restoreTopologyOnly(tr);
	    }	    
	}
	       	  	  	
    }

  /*printf("\n %d %d %d %d %f -> %f\n", mintrav, maxtrav, Thorough, subsequent, lh_res, tr->endLH);*/
   
  Multiple = subsequent;

  free(list);

  return subsequent; 
}



void computeBIG (tree *tr, analdef *adef) 
  { 
    int i, rearrangementsMax, rearrangementsMin, impr, subsequent;                
    double lh, previousLh, difference, epsilon;              
    bestlist *bestT;

    bestT = (bestlist *) malloc(sizeof(bestlist));
    bestT->ninit = 0;
    initBestTree(bestT, 1, tr->mxtips, tr->cdta->endsite);   
    
    /*initInfoList(20);*/
 
    difference = 10.0;

    epsilon = 0.001;
    
    Multiple = 1;
    
    Thorough = 0;  
        
    if(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT)
      {
	Thorough = 1;       
	epsilon = 0.1;
      }    
    
    lh = tr->likelihood;           
   
    impr = 1;		  		  
     
    while(1)
      {		
	recallBestTree(bestT, 1, tr);	

	optimizeModel(tr, adef, 1);       
	treeEvaluate(tr, 2);		
			
	lh = tr->likelihood;	
	saveBestTree(bestT, tr);	

	printLog(tr, adef);	       	       
	
	if(impr)
	  {	    
	    printResult(tr, 0);	   
	    rearrangementsMin = 1;
	    rearrangementsMax = adef->stepwidth;	    
	  }			  			
	else
	  {		       
	    if(Thorough == 0)
	      { 		
		Thorough = 1;
		rearrangementsMax = adef->stepwidth;
		rearrangementsMin = 1; 
	      }
	    else
	      { 
		rearrangementsMax += adef->stepwidth;
		rearrangementsMin += adef->stepwidth; 
	      }	    	      
	    if(rearrangementsMax > adef->max_rearrange)	     	     	 
	      goto cleanup; 	   
	  }

	/*subsequent =*/ 
	treeOptimizeBIG(tr, rearrangementsMin, rearrangementsMax, adef);
	
	/*if(subsequent == 0)
	  {	   
	    Multiple = 0;	    
	    }*/
			
	impr = 0;			      	
	previousLh = lh;
		
	for(i = 1; i <= bt->nvalid; i++)
	  {	    		  	   
	    recallBestTree(bt, i, tr);	    
	    treeEvaluate(tr, 0.25);	    
	  
	    difference = ((tr->likelihood > previousLh)? 
			  tr->likelihood - previousLh: 
			  previousLh - tr->likelihood); 	    
	    if(tr->likelihood > lh && difference > epsilon)
	      {
		impr = 1;	       
		lh = tr->likelihood;	       
		saveBestTree(bestT, tr);		
	      }	   	   
	  }	
      }
    
  cleanup: 
    printLog(tr, adef);
    printResult(tr, 0);	     
  }


/*****************************COMPUTE BIG FAST ***************************************************************/

int determineRearrangementSetting(tree *tr,  analdef *adef, bestlist *bestT)
{
  int i, mintrav, maxtrav, bestTrav, impr;
  double rTime, startLH;
  int MaxFast;

  MaxFast = 26;

  startLH = tr->likelihood;
    
  mintrav = 1;
  maxtrav = 5;

  bestTrav = maxtrav = 5;

  impr = 1;

  resetBestTree(bt);

  while(impr && maxtrav < MaxFast)
    {	
      recallBestTree(bestT, 1, tr);

      rTime = gettime();
      if (maxtrav > tr->ntips - 3)  
	maxtrav = tr->ntips - 3;    
 
      tr->startLH = tr->endLH = tr->likelihood;

      for(i = 1; i <= tr->mxtips + tr->mxtips - 2; i++)
	{                
	  tr->bestOfNode = unlikely;
	  rearrangeBIG(tr, tr->nodep[i], mintrav, maxtrav);                 
	  if(tr->endLH > tr->startLH)                 	
	    {		 	 	      
	      restoreTreeFast(tr);	        	  	 	  	      
	      tr->startLH = tr->endLH = tr->likelihood;	  	 	  	  	  	  	  	  	      
	    }	         	       	  	  	       
	}

      treeEvaluate(tr, 0.25);
      saveBestTree(bt, tr);                                    

      /*printf("%d %f\n", maxtrav, tr->likelihood);*/

      if(tr->likelihood > startLH)
	{	 
	  startLH = tr->likelihood; 	  
	  if(adef->multipleRuns == 0)
	    printLog(tr, adef);
	  bestTrav = maxtrav;	 
	  impr = 1;
	}
      else
	{
	  impr = 0;
	}
      maxtrav += 5;
    }

  recallBestTree(bt, 1, tr);   

  return bestTrav;     
}


int determineRearrangementSettingEXP(tree *tr,  analdef *adef, bestlist *bestT)
{
  int i, mintrav, maxtrav, bestTrav, impr;
  double rTime, startLH, bestRatio, ratio;
  int MaxFast;
  nodeptr *list = (nodeptr *)malloc(sizeof(nodeptr) * 3 * tr->mxtips); 
  int n;
 
  MaxFast = 26; /* TODO let the user specify this */

  startLH = tr->likelihood;
    
  mintrav = 1;
  maxtrav = 5;

  bestTrav = maxtrav = 5;

  impr = 1;

  resetBestTree(bt);

  while(impr && maxtrav < MaxFast)
    {	
      recallBestTree(bestT, 1, tr);

      rTime = gettime();
      if (maxtrav > tr->ntips - 3)  
	maxtrav = tr->ntips - 3;    
 
      tr->startLH = tr->endLH = tr->likelihood;


      n = 0;

      for(i = 1; i <= tr->mxtips + tr->mxtips - 2; i++)
	{    
	  if(tr->nodep[i]->tip)
	    list[n++] = tr->nodep[i]->back;
	  else
	    {
	      list[n++] = tr->nodep[i];
	      if(!tr->nodep[i]->back->tip)
		list[n++] = tr->nodep[i]->back;
	    }
	}
      for(i = 0; i < n; i++)      
	{                
	  tr->bestOfNode = unlikely;
	  rearrangeBIGEXP(tr, list[i], mintrav, maxtrav);                 
	  if(tr->endLH > tr->startLH)                 	
	    {		 	 	      
	      restoreTreeFastEXP(tr);	        	  	 	  	      
	      tr->startLH = tr->endLH = tr->likelihood;	  	 	  	  	  	  	  	  	      
	    }	         	       	  	  	       
	}

      treeEvaluate(tr, 0.25);
      saveBestTree(bt, tr);                                    

      /*printf("%d %f\n", maxtrav, tr->likelihood);*/

      if(tr->likelihood > startLH)
	{	 
	  startLH = tr->likelihood; 	  
	  printLog(tr, adef);
	  bestTrav = maxtrav;
	  bestRatio = ratio;
	  impr = 1;
	}
      else
	{
	  impr = 0;
	}
      maxtrav += 5;
    }

  recallBestTree(bt, 1, tr);   
  
  exit(1);

  return bestTrav;     
}



double treeOptimizeQuick(tree *tr, int mintrav, int maxtrav, analdef *adef)
{
  int i;
  double lh_res;
  int oldThorough;     
  
  if (maxtrav > tr->ntips - 3)  
    maxtrav = tr->ntips - 3;  
  
  resetInfoList();
  resetBestTree(bt);
 
  tr->startLH = tr->endLH = tr->likelihood;

  for(i = 1; i <= tr->mxtips + tr->mxtips - 2; i++)
    {           
     
      tr->bestOfNode = unlikely;

      rearrangeBIG(tr, tr->nodep[i], mintrav, maxtrav);
      
      insertInfoList(i, tr->bestOfNode);

      if(tr->endLH > tr->startLH)                 	
	{		 
	  restoreTreeFast(tr);	        	  	 	  	      
	  tr->startLH = tr->endLH = tr->likelihood;	  	 	  	  	  	  	  	  
	}	    	  
    }

  oldThorough = Thorough;
  Thorough = 1;  
 
  for(i = 0; i < iList.n; i++)
    {      
      tr->bestOfNode = unlikely;

      rearrangeBIG(tr, tr->nodep[iList.list[i].node], 1, 10);                 

      if(tr->endLH > tr->startLH)                 	
	{		 	  
	  restoreTreeFast(tr);	 
	  /*printf("IMP %d %f %f %d bets of this %f\n", i, tr->startLH, tr->endLH, tr->nodep[iList.list[i].node]->number,  tr->bestOfNode);*/
	  tr->startLH = tr->endLH = tr->likelihood;	 
	  saveBestTree(bt, tr);
	}
      else
	{ 
	  /*printf("TOPT %d %f %f %d bets of this %f\n", i, tr->startLH, tr->endLH, tr->nodep[iList.list[i].node]->number, tr->bestOfNode);*/
	  if(tr->bestOfNode != unlikely)
	    {	     
	      restoreTopologyOnly(tr);
	    }	
	}      
    }       

  Thorough =  oldThorough;
  
  return tr->startLH;     
}



void computeBIGFAST (tree *tr, analdef *adef) 
{ 
  int i, rearrangementsMax, rearrangementsMin, noImproveCount, impr, goodTrav, bestTrav, which;                
  double lh, previousLh, difference, epsilon, currentLH, bestLH;              
  bestlist *bestT;
  
  bestT = (bestlist *) malloc(sizeof(bestlist));
  bestT->ninit = 0;
  initBestTree(bestT, 1, tr->mxtips, tr->cdta->endsite);
    
  initInfoList(40);
 
  difference = 10.0;
  epsilon = 0.01;    
    
  Thorough = 0;  
        
  if(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT)
    {
      Thorough = 1;       
      epsilon = 0.1;
    }

  optimizeModel(tr, adef, 1);       
  treeEvaluate(tr, 2);
  printLog(tr, adef);   

  saveBestTree(bestT, tr);   	  		      
       
  bestTrav = determineRearrangementSetting(tr, adef, bestT);       
  printLog(tr, adef);  
  printResult(tr, 0);

  impr = 1;

  while(impr)
    {                            	      

      optimizeModel(tr, adef, 1);    
      treeEvaluate(tr, 2);

      currentLH = tr->likelihood;
      
      bestLH = treeOptimizeQuick(tr, 1, bestTrav, adef);
      
      treeEvaluate(tr, 0.25);

      if(bestLH > currentLH)
	{	  
	  printLog(tr, adef);
	  printResult(tr, 0);

	  difference = (bestLH - currentLH); 	    
	  if(difference > epsilon)	      
	    impr = 1;
	  else
	    impr = 0;
	}
      else
	impr = 0;
    }
  
  bestLH = unlikely;

  for(i = 1; i <= bt->nvalid; i++)
    {	    	    	
      recallBestTree(bt, i, tr);			    
      treeEvaluate(tr, 0.25);
      /*printf("%d %f\n", i, tr->likelihood);*/
      if(tr->likelihood > bestLH)
	{
	  which = i;
	  bestLH = tr->likelihood;
	}           
    }

  recallBestTree(bt, which, tr);  
  optimizeModel(tr, adef, 1);
  treeEvaluate(tr, 2);
  printLog(tr, adef);
  printResult(tr, 0);    
}


double treeOptimizeRapid(tree *tr, int mintrav, int maxtrav, analdef *adef)
{
  int i;  
  int oldThorough;     
  
  if (maxtrav > tr->ntips - 3)  
    maxtrav = tr->ntips - 3;  
  
  resetInfoList();
  resetBestTree(bt);
 
  tr->startLH = tr->endLH = tr->likelihood;

 
  /*printf("%d %d %d %d iterations %d\n", mintrav, maxtrav, Multiple, Thorough, iterations);*/
  
  for(i = 1; i <= tr->mxtips + tr->mxtips - 2; i++)
    {           
      tr->bestOfNode = unlikely;

      rearrangeBIG(tr, tr->nodep[i], mintrav, maxtrav);
      /*NEW*/
      if(Thorough)
	{
	  if(tr->endLH > tr->startLH)                 	
	    {			   	     
	      restoreTreeFast(tr);	 	 
	      tr->startLH = tr->endLH = tr->likelihood;	 
	      saveBestTree(bt, tr);
	    }
	  else
	    { 
	      
	      if(tr->bestOfNode != unlikely)
		{	     
		  restoreTopologyOnly(tr);
		}	
	    }
	  /* NEW */
	}
      else
	{
	  insertInfoList(i, tr->bestOfNode);
           
	  if(tr->endLH > tr->startLH)                 	
	    {		      
	      restoreTreeFast(tr);	  	      
	      tr->startLH = tr->endLH = tr->likelihood;	  	 	  	  	  	  	  	  
	    }	    	  
	}
    }


  /* NEW */
  if(!Thorough)
    {
      /* NEW */
      Thorough = 1;  
 
      for(i = 0; i < iList.valid; i++)
	{      
	  tr->bestOfNode = unlikely;
	  
	  rearrangeBIG(tr, tr->nodep[iList.list[i].node], mintrav, maxtrav);                 
	  
	  if(tr->endLH > tr->startLH)                 	
	    {	 	     
	      restoreTreeFast(tr);	 	 
	      tr->startLH = tr->endLH = tr->likelihood;	 
	      saveBestTree(bt, tr);
	    }
	  else
	    { 
	      
	      if(tr->bestOfNode != unlikely)
		{	     
		  restoreTopologyOnly(tr);
		}	
	    }      
	}       
      
      Thorough = 0;
      /* NEW */
    }
  /* NEW */  

  return tr->startLH;     
}


void computeBIGRAPID (tree *tr, analdef *adef) 
{ 
  int i, rearrangementsMax, rearrangementsMin, noImproveCount, impr, goodTrav, bestTrav, which;                
  double lh, previousLh, difference, epsilon, currentLH, bestLH;              
  bestlist *bestT;
  
  bestT = (bestlist *) malloc(sizeof(bestlist));
  bestT->ninit = 0;
  initBestTree(bestT, 1, tr->mxtips, tr->cdta->endsite);
    
  initInfoList(50);
 
  difference = 10.0;
  epsilon = 0.01;    
    
  Thorough = 0;  
        
  if(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT)
    {
      Thorough = 1; 
      epsilon = 0.1;
    }

  optimizeModel(tr, adef, 1);
  treeEvaluate(tr, 2);
  printLog(tr, adef);   

  saveBestTree(bestT, tr);   	  		      
  
  if(!adef->initialSet)
    bestTrav = determineRearrangementSetting(tr, adef, bestT);       
  else
    bestTrav = adef->initial;

  /*printf("bestTrav: %d\n", bestTrav);*/

  saveBestTree(bestT, tr);  
  impr = 1;

  if(!(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT))
    {
      while(impr)
	{     
	  recallBestTree(bestT, 1, tr);	  
	  optimizeModel(tr, adef, 1);    	 
	  treeEvaluate(tr, 2);	 	 
	  saveBestTree(bestT, tr);	 
	  printLog(tr, adef);
	  printResult(tr, 0);
	  saveBestTree(bestT, tr);
	  lh = previousLh = tr->likelihood;
	  
	  treeOptimizeRapid(tr, 1, bestTrav, adef);
	  
	  impr = 0;
	  
	  for(i = 1; i <= bt->nvalid; i++)
	    {	    		  	   
	      recallBestTree(bt, i, tr);	    
	      treeEvaluate(tr, 0.25);	    	 	
	      
	      difference = ((tr->likelihood > previousLh)? 
			    tr->likelihood - previousLh: 
			    previousLh - tr->likelihood); 	    
	      if(tr->likelihood > lh && difference > epsilon)
		{
		  impr = 1;	       
		  lh = tr->likelihood;	       
		  saveBestTree(bestT, tr);		
		}	   	   
	    }	
	}
    }

  Thorough = 1;
  impr = 1;
  
  while(1)
    {		
      recallBestTree(bestT, 1, tr);	
          	     	        
      if(impr)
	{	    
	  printResult(tr, 0);	   
	  rearrangementsMin = 1;
	  rearrangementsMax = adef->stepwidth;	    
	}			  			
      else
	{		       	   
	  rearrangementsMax += adef->stepwidth;
	  rearrangementsMin += adef->stepwidth; 	        	      
	  if(rearrangementsMax > adef->max_rearrange)	     	     	 
	    goto cleanup; 	   
	}
      
      optimizeModel(tr, adef, 1);       
      treeEvaluate(tr, 2.0);					
      previousLh = lh = tr->likelihood;	
      saveBestTree(bestT, tr);	

      printLog(tr, adef);  

      treeOptimizeRapid(tr, rearrangementsMin, rearrangementsMax, adef);
	
      impr = 0;			      	
		
      for(i = 1; i <= bt->nvalid; i++)
	{	

	  recallBestTree(bt, i, tr);	    
	  
	  treeEvaluate(tr, 0.25);	    	 
	  
	  difference = ((tr->likelihood > previousLh)? 
			tr->likelihood - previousLh: 
			previousLh - tr->likelihood); 	    
	  if(tr->likelihood > lh && difference > epsilon)
	    {
	      impr = 1;	       
	      lh = tr->likelihood;	       
	      saveBestTree(bestT, tr);		
	    }	   	   
	}	
    }
    
 cleanup:   
  freeBestTree(bestT);
  free(bestT);
  freeInfoList();
  printLog(tr, adef);
  printResult(tr, 0);	         
}



void computeBIGEXP (tree *tr, analdef *adef) 
{ 
  int i, rearrangementsMax, rearrangementsMin, noImproveCount, impr, goodTrav, bestTrav, which;                
  double lh, previousLh, difference, epsilon, currentLH, bestLH;              
  bestlist *bestT;
  
  bestT = (bestlist *) malloc(sizeof(bestlist));
  bestT->ninit = 0;
  initBestTree(bestT, 1, tr->mxtips, tr->cdta->endsite);
    
  initInfoList(50);
 
  difference = 10.0;
  epsilon = 0.01;    
    
  Thorough = 0;  
        
  if(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT)
    {
      Thorough = 1;       
      epsilon = 0.1;
    }

  optimizeModel(tr, adef, 1);       
  treeEvaluate(tr, 2);
  printLog(tr, adef);   

  saveBestTree(bestT, tr);   	  		      
       
  bestTrav = determineRearrangementSettingEXP(tr, adef, bestT);       
  saveBestTree(bestT, tr);

  impr = 1;

  while(impr)
    {     
      recallBestTree(bestT, 1, tr);
      optimizeModel(tr, adef, 1);    
      treeEvaluate(tr, 2);
      saveBestTree(bestT, tr);

      printLog(tr, adef);
      printResult(tr, 0);
      saveBestTree(bestT, tr);
      lh = previousLh = tr->likelihood;
     
      treeOptimizeRapid(tr, 1, bestTrav, adef);
      
      impr = 0;

      for(i = 1; i <= bt->nvalid; i++)
	{	    		  	   
	  recallBestTree(bt, i, tr);	    
	  treeEvaluate(tr, 0.25);	    	 	

	  difference = ((tr->likelihood > previousLh)? 
			tr->likelihood - previousLh: 
			previousLh - tr->likelihood); 	    
	  if(tr->likelihood > lh && difference > epsilon)
	    {
	      impr = 1;	       
	      lh = tr->likelihood;	       
	      saveBestTree(bestT, tr);		
	    }	   	   
	}	
    }
  
  printf("SWITCHING\n");

  /* NEW */
  Thorough = 1;
  /* NEW */


  impr = 1;
  
  while(1)
    {		
      recallBestTree(bestT, 1, tr);	

      optimizeModel(tr, adef, 1);       
      treeEvaluate(tr, 2);					
      previousLh = lh = tr->likelihood;	
      saveBestTree(bestT, tr);	

      printLog(tr, adef);      	
     	        
      if(impr)
	{	    
	  printResult(tr, 0);	   
	  rearrangementsMin = 1;
	  rearrangementsMax = adef->stepwidth;	    
	}			  			
      else
	{		       	   
	  rearrangementsMax += adef->stepwidth;
	  rearrangementsMin += adef->stepwidth; 	        	      
	  if(rearrangementsMax > adef->max_rearrange)	     	     	 
	    goto cleanup; 	   
	}
      
      treeOptimizeRapid(tr, rearrangementsMin, rearrangementsMax, adef);
	

      impr = 0;			      	

		
      for(i = 1; i <= bt->nvalid; i++)
	{	    		  	   
	  recallBestTree(bt, i, tr);	    
	  treeEvaluate(tr, 0.25);	    
	  
	  difference = ((tr->likelihood > previousLh)? 
			tr->likelihood - previousLh: 
			previousLh - tr->likelihood); 	    
	  if(tr->likelihood > lh && difference > epsilon)
	    {
	      impr = 1;	       
	      lh = tr->likelihood;	       
	      saveBestTree(bestT, tr);		
	    }	   	   
	}	
    }
    
  cleanup: 
    printLog(tr, adef);
    printResult(tr, 0);	         
}





void computeBIGRAPIDFAST (tree *tr, analdef *adef) 
{ 
  int i, mintrav, maxtrav, which;
  double initialLH, epsilon, diff;
  bestlist *bestT;
  
  bestT = (bestlist *) malloc(sizeof(bestlist));
  bestT->ninit = 0;
  initBestTree(bestT, 1, tr->mxtips, tr->cdta->endsite);
 
  epsilon = 0.01;
  
  Thorough = 0;

  if(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT)
    Thorough = 1;

  optimizeModel(tr, adef, 1);
  treeEvaluate(tr, 2);
  saveBestTree(bestT, tr); 

  Multiple = 1;
  mintrav = 1;
  maxtrav = determineRearrangementSetting(tr, adef, bestT);

  optimizeModel(tr, adef, 1);
  treeEvaluate(tr, 2);
  
  if (maxtrav > tr->ntips - 3)  maxtrav = tr->ntips - 3;
          
  tr->opt_level = 0;
  
  do
    {
      printLog(tr, adef);
      printResult(tr, 0);

      initialLH = tr->likelihood;
      tr->startLH = tr->endLH = tr->likelihood;

      for(i = 1; i < (2 * tr->mxtips - 1); i++)
	{ 
	  tr->bestOfNode = unlikely;

	  rearrangeBIG(tr, tr->nodep[i], mintrav, maxtrav);
     	  	  
	  if(tr->endLH > tr->startLH)                 	
	    {		
	      
	      restoreTreeFast(tr);	 	 
	      tr->startLH = tr->endLH = tr->likelihood;	 
	      saveBestTree(bt, tr);
	    }
	  else
	    { 	      
	      if(tr->bestOfNode != unlikely)
		{	     
		  restoreTopologyOnly(tr);
		}	
	    }	  
	}	  	      
      treeEvaluate(tr, 1);            
      diff = (tr->likelihood > initialLH)? tr->likelihood - initialLH: -1.0;            
    }
  while(diff > epsilon);

  initialLH = unlikely;

  for(i = 1; i <= bt->nvalid; i++)
    {	    	    	
      recallBestTree(bt, i, tr);			    
      smoothTree(tr, smoothings/4);
      evaluate(tr, tr->start);		           
      if(tr->likelihood > initialLH)
	{
	  which = i;
	  initialLH = tr->likelihood;
	}           
    }

  recallBestTree(bt, which, tr);
  treeEvaluate(tr, 1); 
  optimizeModel(tr, adef, 1);
  printLog(tr, adef);
  printResult(tr, 0);

}



void simulatedAnnealingRapid(tree *tr, analdef *adef)
{
  double lh1, lh2, best, p1, p2, deltaf, alpha, t0;
  int repeats;
  int node, mintrav, maxtrav, rn, c_imp, selectMove, counter, topologyMove;
  bestlist *consensusList, *bestT;
  int maxR;

  
  /*srand((unsigned int) time(NULL));*/

  bestT = (bestlist *) malloc(sizeof(bestlist));
  bestT->ninit = 0;
  initBestTree(bestT, 1, tr->mxtips, tr->cdta->endsite);

  maxR = adef->max_rearrange; 

  if(buildConsensus)
    {
      consensusList = (bestlist *) malloc(sizeof(bestlist));
      consensusList->ninit = 0;
      initBestTree(consensusList, 100, tr->mxtips, tr->cdta->endsite);
    }
  
  Multiple = 0;

  /* TODO passt noch nicht fuer prots*/

  if(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT)
    {
      Thorough = 1;
    }
  else
    Thorough = 0;
   
  optimizeModel(tr, adef, 1);
  treeEvaluate(tr, 4);

  alpha = 0.9;
  best = unlikely;
  t0 = adef->t0;
  
  if(adef->repeats == 0)
    adef->repeats = tr->mxtips;
  else
    repeats = adef->repeats;

  c_imp = 0;
  counter = 0;

  saveBestTree(bestT, tr);

  while(1)
    {      
      lh1 = tr->likelihood;
      
      if(lh1 > best)
	{
	  best = lh1;
	  printLog(tr, adef);	 
	  printResult(tr, 0);
	}
   
     
      node = 1 + randomInt(2 * tr->mxtips - 2);
      
      
      selectMove = randomInt(100);      



      if(selectMove < 95)
	{
	  topologyMove = 1;
	  mintrav = 1;	  
	  maxtrav = 2 + randomInt(maxR);  
	  
	  tr->startLH = tr->endLH = unlikely;
	  
	  rearrangeBIG(tr, tr->nodep[node], mintrav, maxtrav);	 
	  restoreTreeFast(tr);
	  regionalSmooth(tr, tr->removeNode, 1, 4);
	  evaluate(tr, tr->removeNode);
	  /*evalFast(tr);*/
	  
	  c_imp++;
	}
      else	
	{
	  topologyMove = 0;
	  selectMove = randomInt(100);
	  
	  if(selectMove < 95)
	    {	      
	      nodeptr p;
	      nodeptr q;
	      double z0, z;
	      
	      p = tr->nodep[node];
	      
	      q = p->back;
	      z0 = q->z;		
	      z = makenewz(tr, p, q, z0, newzpercycle); 	  
	      p->z = q->z = z;
	      
	      evaluate(tr, tr->start);	
	    }
	  else
	    {
	      optimizeModel(tr, adef, 1);
	      treeEvaluate(tr, 4);
	    }
	}
      
     
      lh2 = tr->likelihood;

      if(lh1 > lh2)
	{	  
	  deltaf = lh2 - lh1;
	  p1 = exp( deltaf / t0);
	  rn = randomInt(10000);
	  if(rn == 0) 
	    p2 = 0;
	  else 
	    p2 = ((double)rn)/10000.0;	  	  

	  if(buildConsensus && topologyMove && p1 < p2)
	    saveBestTree(consensusList, tr);	
	  if(p1 < p2)	  
	    {
	      recallBestTopology(bestT, 1, tr);	     
	    }
	  else
	    {
	      resetBestTree(bestT);
	      saveBestTree(bestT, tr);
	      /*printf("BACKWARD %f -> %f \n ", lh1, lh2);*/	     	     
	    }
	}    
      else
	{
	  /*resetBestTree(bestT);*/
	  saveBestTree(bestT, tr);

	  if(topologyMove)
	    {	     
	      if(buildConsensus)
		saveBestTree(consensusList, tr);
	    }
	 
	}

      if(c_imp >= repeats)
	{
	  t0 = t0 * alpha;	  
	  c_imp = 0;
	}
     
      counter++;     

      if((counter % (2 * tr->mxtips)) == 0) 	
	{
	  if(buildConsensus) 
	    consensusTree(consensusList, tr);
	
	  counter = 0;
	}          
    }          
}








/********************** COMPUTE BIG ***************************************************************************/










/********************** COMPUTE PARALLEL ***************************************************************************/

#ifdef PARALLEL_2

  
nodeptr  removeNodePARALLEL (tree *tr, nodeptr p)
{
  double   zqr;
  nodeptr  q, r;
        
  q = p->next->back;
  r = p->next->next->back;
  zqr = q->z * r->z;
       
  tr->removeNumberP = p->number;
  tr->removeNumberQ = q->number;
  tr->removeNumberR = r->number;
  

  if ((zqr = makenewz(tr, q, r, zqr, iterations)) == badZ) return (node *) NULL;      

  tr->zqr = zqr;
    
  hookup(q, r, zqr);

  p->next->next->back = p->next->back = (node *) NULL;
     
  return  q;
}



boolean insertPARALLEL (tree *tr, nodeptr p, nodeptr q, boolean glob)
{
  nodeptr  r, s;
  
  r = q->back;
  s = p->back;

  

  if(Thorough)
    { 
      double  zqr, zqs, zrs, lzqr, lzqs, lzrs, lzsum, lzq, lzr, lzs, lzmax;      
            
       

      if ((zqr = makenewz(tr, q, r, q->z, iterations)) == badZ) return FALSE; 
     
      if ((zqs = makenewz(tr, q, s, defaultz, iterations)) == badZ) return FALSE;      
     
      if ((zrs = makenewz(tr, r, s, defaultz, iterations)) == badZ) return FALSE;
     

      lzqr = (zqr > zmin) ? log(zqr) : log(zmin); 
      lzqs = (zqs > zmin) ? log(zqs) : log(zmin);
      lzrs = (zrs > zmin) ? log(zrs) : log(zmin);
      lzsum = 0.5 * (lzqr + lzqs + lzrs);
      
      lzq = lzsum - lzrs;
      lzr = lzsum - lzqs;
      lzs = lzsum - lzqr;
      lzmax = log(zmax);
      
      if      (lzq > lzmax) {lzq = lzmax; lzr = lzqr; lzs = lzqs;} 
      else if (lzr > lzmax) {lzr = lzmax; lzq = lzqr; lzs = lzrs;}
      else if (lzs > lzmax) {lzs = lzmax; lzq = lzqs; lzr = lzrs;}          
            
      hookup(p->next,       q, exp(lzq));
      hookup(p->next->next, r, exp(lzr));
      hookup(p,             s, exp(lzs));          
    }
  else
    { 
      
      double  z;
           
      z = sqrt(q->z);      

       if(z < zmin) 
	z = zmin;
      if(z > zmax)
	z = zmax;

      hookup(p->next,       q, z);
      hookup(p->next->next, r, z);
    }   
    
  if (! newview(tr, p)) return FALSE;    

  if(Thorough)
    {     
     
      localSmooth(tr, p, smoothings);
      

      /* regionalSmooth() for special ? */

      /* Did not show to be more efficient*/
      tr->lzq = p->next->z;
      tr->lzr = p->next->next->z;
      tr->lzs = p->z;
      
    }           
     
  return  TRUE;
}




boolean testInsertPARALLEL (tree *tr, nodeptr p, nodeptr q)
{
  double  qz, pz;
  nodeptr  r;
  
  r = q->back; 

  qz = q->z;
  pz = p->z;
  
  
  if (! insertPARALLEL(tr, p, q, FALSE))       return FALSE;         
  
  if (evaluate(tr, p->next->next) == badEval) return FALSE;   
   
  if(tr->likelihood > tr->bestOfNode)
    {
      tr->bestOfNode = tr->likelihood;
      tr->insertNode = q;
      tr->removeNode = p;   
      tr->currentZQR = tr->zqr;           
      tr->currentLZR = tr->lzr;
      tr->currentLZQ = tr->lzq;
      tr->currentLZS = tr->lzs;
    }
  
  if(tr->likelihood > tr->endLH)
    {	           
      tr->insertNode = q;
      tr->removeNode = p;   
      tr->currentZQR = tr->zqr;      
      tr->endLH = tr->likelihood;           
    
      tr->removedNumberP = tr->removeNumberP;
      tr->removedNumberQ = tr->removeNumberQ;
      tr->removedNumberR = tr->removeNumberR;
      tr->insertNumberQ = q->number;
      tr->insertNumberR = r->number;
    }    
  
  hookup(q, r, qz);
  
  p->next->next->back = p->next->back = (nodeptr) NULL;

  if(Thorough)
    {
      nodeptr s = p->back;
      hookup(p, s, pz);      
    } 
  
  return TRUE;
}



 
void addTraversePARALLEL(tree *tr, nodeptr p, nodeptr q, int mintrav, int maxtrav)
{  
  if (--mintrav <= 0) 
    {              
      if (! testInsertPARALLEL(tr, p, q))  return;        
    }
  
  if ((! q->tip) && (--maxtrav > 0)) 
    {    
      addTraversePARALLEL(tr, p, q->next->back, mintrav, maxtrav);
      addTraversePARALLEL(tr, p, q->next->next->back, mintrav, maxtrav);    
    }
} 


int  rearrangePARALLEL(tree *tr, nodeptr p, int mintrav, int maxtrav)    
  { 
    double   p1z, p2z, q1z, q2z;
    nodeptr  p1, p2, q, q1, q2;
    int      mintrav2;
    int result = 0;
   
    if (maxtrav < 1 || mintrav > maxtrav)  return 0;
    
    if (! p->tip) 
      {     
	p1 = p->next->back;
	p2 = p->next->next->back;

	if (! p1->tip || ! p2->tip) 
	  {
	    p1z = p1->z;
	    p2z = p2->z;
	    if (! removeNodePARALLEL(tr, p)) return badRear;
	  
	    if (! p1->tip) 
	      {
		addTraversePARALLEL(tr, p, p1->next->back,
			       mintrav, maxtrav);         
		addTraversePARALLEL(tr, p, p1->next->next->back,
			       mintrav, maxtrav);          
	      }

	    if (! p2->tip) 
	      {
		addTraversePARALLEL(tr, p, p2->next->back,
			       mintrav, maxtrav);
		addTraversePARALLEL(tr, p, p2->next->next->back,
			       mintrav, maxtrav);          
	      }

	    hookup(p->next,       p1, p1z);
	    hookup(p->next->next, p2, p2z);

	    initravDIST(tr, p, 1); 	   
	  }
      }  

    q = p->back;
    
    if (! q->tip && maxtrav > 1) 
      {
	q1 = q->next->back;
	q2 = q->next->next->back;

	if (! q1->tip && (!q1->next->back->tip || !q1->next->next->back->tip) ||
	    ! q2->tip && (!q2->next->back->tip || !q2->next->next->back->tip)) 
	  {
	    q1z = q1->z;
	    q2z = q2->z;
	    if (! removeNodePARALLEL(tr, q)) return badRear;
	  
	    mintrav2 = mintrav > 2 ? mintrav : 2;

	    if (! q1->tip) 
	      {
		addTraversePARALLEL(tr, q, q1->next->back,
			       mintrav2 , maxtrav);
		addTraversePARALLEL(tr, q, q1->next->next->back,
			       mintrav2 , maxtrav);         
	      }

	    if (! q2->tip) 
	      {
		addTraversePARALLEL(tr, q, q2->next->back,
			       mintrav2 , maxtrav);
		addTraversePARALLEL(tr, q, q2->next->next->back,
			       mintrav2 , maxtrav);          
	      }

	    hookup(q->next,       q1, q1z); 
	    hookup(q->next->next, q2, q2z);

	    initravDIST(tr, q, 1); 	    
	  }
      } 

    return  1;
  } 


nodeptr verifyRemoveNode(tree *tr)
/* moves node p (remove Node/removed Subtree) to the correct sector and checks 
   if it can be removed in the current topology*/
{
  nodeptr p = tr->nodep[tr->removedNumberP]; 
  nodeptr correctSector = (nodeptr) NULL;

  if(p->tip)
    {
      printf("ERROR in  verifyRemoveNode() is a TIP\n");
      exit(-1);
    } 

  if(p->back->number == tr->removedNumberQ && p->next->back->number == tr->removedNumberR)    
    correctSector = p->next->next;          
  else
    {
      if(p->next->back->number == tr->removedNumberQ && p->next->next->back->number == tr->removedNumberR)	
	correctSector = p;	
      else
	{
	  if(p->next->next->back->number == tr->removedNumberQ && p->back->number == tr->removedNumberR)	    
	    correctSector = p->next;	    
	}
    } 

  return correctSector;
}


nodeptr verifyInsertNode(tree *tr)
/* moves node q (insert Node/insertion Branch) to the correct sector and checks 
   if it can be used as insertion point in the current topology*/
{
  nodeptr q = tr->nodep[tr->insertNumberQ];
  nodeptr r;
  nodeptr insertSector = (nodeptr) NULL;

  q = tr->nodep[tr->insertNumberQ];
  
  if(q->tip)
    {      
      if(q->back->number == tr->insertNumberR && q->back->back->number == tr->insertNumberQ);
      insertSector = q;    
    }
  else
    {
      if(q->back->number == tr->insertNumberR && q->back->back->number == tr->insertNumberQ)    
	insertSector = q;    
      else
	{
	  if(q->next->back->number == tr->insertNumberR && q->next->back->back->number == tr->insertNumberQ)
	    insertSector = q->next;
	  else
	    {
	      if(q->next->next->back->number == tr->insertNumberR && q->next->next->back->back->number == tr->insertNumberQ)
		insertSector = q->next->next;	      
	    }
	}     
    }
  return  insertSector;
}







int inSub(nodeptr p,  int insertSectorNumber)
/* checks if LZR received by other worker can be 
   performed in local topology*/
{
  nodeptr q;
  int result = 0;

  if(p->tip)
    {
      if(p->number == insertSectorNumber) 
	return 1;
      else
	return 0;
    }
  else
    {
       if(p->number == insertSectorNumber) 
	 return 1;
       else
	 {
	   q = p->next;
	   while(q != p)
	     {
	       result = (result || inSub(q->back, insertSectorNumber));
	       q = q->next;
	     }
	     
	   return result;
	 }
    }
      
}
   
 
int inSubtree(tree *tr, nodeptr q, int insertSectorNumber)
/* recursive call to inSub for right and left children*/
{
  int r1, r2;
  nodeptr r = q->back;

  r1 = inSub(q,  insertSectorNumber);
  r2 = inSub(r,  insertSectorNumber);

  return (r1 || r2);
}


/* TODO-Michael make this dynamic*/

#define STEPS 99
#define LENGTH 1002 /* == 2 + 100 * 10 */

int applyTopologicalChanges(tree *tr)
/* applies topological changes received from other workers,
   regardless if likelihood increases or decreases
   returns 0 if move could not be performed due to topological restrictions*/
{
  nodeptr removeSector, insertSector; 
  nodeptr p1, p2, r, s;
  double p1z, p2z, zqr, qz, ps, z, zqs, zrs, lzqr, lzqs, lzrs, lzsum, lzq, lzr, lzs, lzmax;   
  int r1, r2;

  removeSector = verifyRemoveNode(tr);  /* can we remove that node? */

 
  if(!removeSector) 
    return 0;
  
  insertSector = verifyInsertNode(tr);  /* can we really insert at insertSector*/

  if(!insertSector) 
    return 0;
  
  p1 = removeSector->next->back;
  p2 = removeSector->next->next->back;
 
  r1 = inSub(p1,  insertSector->number); /* check if in proper subtree*/
  r2 = inSub(p2,  insertSector->number); /* check if in proper subtree*/

  if((!removeSector->tip) && (! p1->tip || ! p2->tip) && ((r1 || r2) != 0)) 
    /* if all this is true then we can perform the move */
    {                      
      hookup(p1, p2, tr->currentZQR);		

      removeSector->next->next->back = removeSector->next->back = (node *) NULL;    
                   	
      r = insertSector->back;
      qz = insertSector->z;
          
      s = removeSector->back;

      if(Thorough)
	{ 	     	  	        
	  hookup(removeSector->next,       insertSector, tr->currentLZQ);
	  hookup(removeSector->next->next, r, tr->currentLZR);
	  hookup(removeSector,             s, tr->currentLZS);      		  
	}
      else
	{ 	  	  
	  z = sqrt(insertSector->z);

	  if(z < zmin) 
	    z = zmin;
	  if(z > zmax)
	    z = zmax;
	  
	  hookup(removeSector->next,       insertSector, z);
	  hookup(removeSector->next->next, r, z);
	}      	      	                    	
    }
  else
    return 0;

  return 1;
}

void packRearrangementInfo(tree *tr, int *improved, double *rSpec)
/* packs the rearrangement information to be sent to other workers */
{
  double *rSpecPtr;

  rSpecPtr = &rSpec[2 + (*improved) * 10];
	      
  rSpecPtr[0] = (double) Thorough;
  rSpecPtr[1] = (double) tr->removedNumberP;
  rSpecPtr[2] = (double) tr->removedNumberQ;
  rSpecPtr[3] = (double) tr->removedNumberR;
  rSpecPtr[4] = (double) tr->insertNumberQ;
  rSpecPtr[5] = (double) tr->insertNumberR;
  rSpecPtr[6] = tr->currentZQR;
  rSpecPtr[7] = tr->currentLZQ;
  rSpecPtr[8] = tr->currentLZR;
  rSpecPtr[9] = tr->currentLZS;

  *improved = *improved + 1;
}


void workerProcess(tree *tr, analdef *adef)
{
  int mintrav;
  int maxtrav;
  /* TODO-Michael make rSpec and rSpecIn dynamic*/
  double rSpec[LENGTH], rSpecIn[LENGTH], *rSpecPtr,  *rSpecInPtr;
  int node, i, msgFrom, receivedSomething, improved, upper, loopCounter;
  MPI_Request  msgRequest;
  MPI_Status msgStatus; 
  int applicationsOfTopologicalChanges;
  
  double iniLH;/* TODO-Michael remove this, used for testing only */

  if (maxtrav > tr->ntips - 3)  
    maxtrav = tr->ntips - 3;  
      
  srand((unsigned int) gettimeSrand());/* TODO-Michael remove this, used for testing only */

  tr->startLH = tr->endLH = tr->likelihood;
  
  mintrav = 1; /* TODO-Michael adjust mintrav, maxtrav, Thorough according to the new search algorithm */
  maxtrav = 10;
  Thorough = 0;

  while(1)
    {	 


      applicationsOfTopologicalChanges = 0;
      MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &receivedSomething, &msgStatus);    
            
      while(receivedSomething)
	{
	  switch(msgStatus.MPI_TAG)
	    {
	    case REARRANGEMENT_SPECIFICATION:
	      msgFrom = msgStatus.MPI_SOURCE;
	      	     	      
	      MPI_Recv(rSpecIn, LENGTH, MPI_DOUBLE, msgFrom, REARRANGEMENT_SPECIFICATION, MPI_COMM_WORLD, &msgStatus);          	  	      	     

	      printf("Worker %d received %d LZRs from worker %d\n", myId,  (int)(rSpecIn[0]), msgFrom);
	     	     
	      improved = rSpecIn[0];	     

	      for(i = 0; i < improved; i++)
		{
		  rSpecInPtr = &rSpecIn[2 + 10 * i];
		
		  Thorough = (int) rSpecInPtr[0];
		  tr->removedNumberP = (int) rSpecInPtr[1];
		  tr->removedNumberQ = (int) rSpecInPtr[2];
		  tr->removedNumberR = (int) rSpecInPtr[3];
		  tr->insertNumberQ = (int) rSpecInPtr[4];
		  tr->insertNumberR = (int) rSpecInPtr[5];
		  tr->currentZQR =  rSpecInPtr[6];
		  tr->currentLZQ =  rSpecInPtr[7];
		  tr->currentLZR =  rSpecInPtr[8];
		  tr->currentLZS =  rSpecInPtr[9];	      
		
		  if(applyTopologicalChanges(tr))
		    applicationsOfTopologicalChanges++;
		}	      
	          	

	      break;
	    }
	  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &receivedSomething, &msgStatus);
	}  

      if(applicationsOfTopologicalChanges > 0)
	{
	  initrav(tr, tr->start);
	  initrav(tr, tr->start->back);
	  evaluate(tr, tr->start);
	  printf("Worker %d after top change %f\n", myId, tr->likelihood);
	}


      node = 1 + randomInt(2 * tr->mxtips - 2); /* TODO-Michael replace by a less-random LZR-move distribution strategy*/ 
	
      improved = 0; /* count the number of improving LSR moves */

      upper = node + STEPS; /* TODO-Michael make STEPS dynamic as a function of alignment size and number of workers */

      if(upper > 2 * tr->mxtips - 2)
	upper =  2 * tr->mxtips - 2;

      iniLH = tr->startLH = tr->endLH = tr->likelihood;


      for(i = node; i <= upper; i++)
	{	 
	  tr->bestOfNode = unlikely;
	  
	  rearrangePARALLEL(tr, tr->nodep[i], mintrav, maxtrav);
	  
	  if(Thorough)
	    {
	      if(tr->endLH > tr->startLH)                 	
		{		 	  
		  restoreTreeFast(tr);	 	 
		  tr->startLH = tr->endLH = tr->likelihood;
		  
		  packRearrangementInfo(tr, &improved, rSpec);	/* if tree improves store it in data structure and send it later on */
		  
		  saveBestTree(bt, tr);
		}
	      else
		{ 		    
		  if(tr->bestOfNode != unlikely)
		    {			      
		      restoreTopologyOnly(tr);		      
		    }	
		}	  
	    }
	  else
	    {
	      insertInfoList(i, tr->bestOfNode);
	      
	      if(tr->endLH > tr->startLH)                 	
		{		 
		  restoreTreeFast(tr);	        	  	 	  	      
		  tr->startLH = tr->endLH = tr->likelihood;	  	 	  	  	  	  	  	  
		  
		  packRearrangementInfo(tr, &improved, rSpec);	/* if tree improves store it in data structure and send it later on */	   
		}	    	  
	    }	  	   
	}

	
      evaluate(tr, tr->start); /* TODO-Michael, remove that later-on, used for testing only */
      printf("ID %d: %f -> %f %f\n", myId, iniLH, tr->endLH, tr->likelihood); /* TODO-Michael used for testing only */
	
      if(improved)
	/* if worker found ikmproving LSR moves it will send the asynchronously to all other workers*/
	{
	  rSpec[0] = improved;
	  rSpec[1] = tr->endLH;
	  
	  printf("Worker %d sending %d LZRs to all others\n", myId,  improved);
	  
	  for(i = 0; i < numOfWorkers; i++)
	    {
	      if(i != myId)
		MPI_Isend(rSpec, LENGTH, MPI_DOUBLE, i, REARRANGEMENT_SPECIFICATION, MPI_COMM_WORLD, &msgRequest);	  	     
	    }	  	 
	}           
    }       

  /* TODO-Michael:
     -neuen suchalgorithmus in dem geruest implementieren
     -problem mit der parallelisierung der bestimmung des rearrangement-settings
     -ModellOptimierung fuer alle DNA-modelle loesen
     -speicherung der baeume in den verschiedenen listen pro worker parallelisierin
     -asynchron zeitpunkt bestimmen, an dem alle worker die kantenlaengen der baeume 
     in ihrer bestlist() optimieren
     -asuwaehlen der besten topologie
  */
     
}




void computePARALLEL (tree *tr, analdef *adef) 
  {       
    initInfoList(20);        /* TODO-Michael im sequentiellen programm ist das ja auf 50 gesetzt, wie machen wir das im parallelen*/
    	
    optimizeModel(tr, adef, 1);       /* TODO-Michael optimizeModel() an einen dedizierten worker, fuer Gamma ist die model-pack/unpack
					 funktion noch nicht implementiert*/
    treeEvaluate(tr, 2);	      	

    printf("Worker %d\n", myId);
		      
    workerProcess(tr, adef);         /* wir sollten versuchen ohne master auszukommen */
  }
#endif










/************************************ Compute PARALLEL *********************************************/






double calcPartialLikelihood(tree *tr, int i, double ki)
{     
  /*initravPartial(tr, tr->start, i, ki);
    initravPartial(tr, tr->start->back, i, ki);*/ 

  return evaluatePartial(tr, tr->start, i, ki);
}





static int doublecompare2(const void *p1, const void *p2)
{
  double i = *((double *)p1);

  double j = *((double *)p2);
  
  if (i > j)
    return (1);
  if (i < j)
    return (-1);
  return (0);
}

static int catCompare(const void *p1, const void *p2)
{
 rateCategorize *rc1 = (rateCategorize *)p1;
 rateCategorize *rc2 = (rateCategorize *)p2;

  double i = rc1->accumulatedSiteLikelihood;
  double j = rc2->accumulatedSiteLikelihood;
  
  if (i > j)
    return (1);
  if (i < j)
    return (-1);
  return (0);
}



void categorize(tree *tr, rateCategorize *rc)
{
  int i, k, found;
  double temp, diff, min;

  for (i = 0; i < tr->cdta->endsite; i++) 
      {
	temp = tr->cdta->patrat[i];
	found = 0;
	for(k = 0; k < tr->NumberOfCategories; k++)
	  {
	    if(temp == rc[k].rate || (fabs(temp - rc[k].rate) < 0.001))
	      {
		found = 1;
		tr->cdta->rateCategory[i] = k;				
		break;
	      }
	  }
	if(!found)
	  {
	    min = fabs(temp - rc[0].rate);
	    tr->cdta->rateCategory[i] = 0;

	    for(k = 1; k < tr->NumberOfCategories; k++)
	    {
	      diff = fabs(temp - rc[k].rate);
	      if(diff < min)
		{
		  min = diff;
		  tr->cdta->rateCategory[i] = k;
		}
	    }
	  }
      }

  for(k = 0; k < tr->NumberOfCategories; k++)
    tr->cdta->patrat[k] = rc[k].rate; 

}




void optimizeRateCategories(tree *tr, int categorized, int _maxCategories)
{
  int i, k;
  double initialRate, initialLikelihood, v, leftRate, rightRate, leftLH, rightLH, temp, wtemp;   
  double lower_spacing, upper_spacing;
  int maxCategories = _maxCategories;
  double initialLH = tr->likelihood;
  double *oldRat = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  double *ratStored = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  double *oldwr = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  double *oldwr2 = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  double *lhs = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  int *oldCategory = (int *)malloc(sizeof(int) * tr->cdta->endsite);  
  int oldNumber; 
  double epsilon = 0.00001;
  double optt = gettime();
  double help;
  double accT = 0;

  if(optimizeRateCategoryInvocations == 1)
    {
      lower_spacing = 0.5 / ((double)optimizeRateCategoryInvocations);
      upper_spacing = 1.0 / ((double)optimizeRateCategoryInvocations);
    }
  else
    {
      lower_spacing = 0.05 / ((double)optimizeRateCategoryInvocations);
      upper_spacing = 0.1 / ((double)optimizeRateCategoryInvocations);
    }

  if(lower_spacing < 0.001)
    lower_spacing = 0.001;

  if(upper_spacing < 0.001)
    upper_spacing = 0.001;

  optimizeRateCategoryInvocations++;
  
  oldNumber = tr->NumberOfCategories;


  for(i = 0; i < tr->cdta->endsite; i++)
    {    
      oldCategory[i] = tr->cdta->rateCategory[i];
      ratStored[i] = tr->cdta->patratStored[i];    
      oldRat[i] = tr->cdta->patrat[i];
      oldwr[i] =  tr->cdta->wr[i];
      oldwr2[i] =  tr->cdta->wr2[i];
    }

  


 
#pragma omp parallel for private(initialRate, initialLikelihood, leftLH, rightLH, initialLikelihood, leftRate, rightRate, initialRate, k, v) schedule(dynamic)
  for(i = 0; i < tr->cdta->endsite; i++)
    {    
      tr->cdta->patrat[i] = tr->cdta->patratStored[i];     
      initialRate = tr->cdta->patrat[i];
      
      initialLikelihood = calcPartialLikelihood(tr, i, initialRate);
      
      
      leftLH = rightLH = initialLikelihood;
      leftRate = rightRate = initialRate;
       
      k = 1;
      while((initialRate - k * lower_spacing > 0.0001) && 
	    ((v = calcPartialLikelihood(tr, i, initialRate - k * lower_spacing)) > leftLH) && 
	    (fabs(leftLH - v) > epsilon))
	{	  
	  leftLH = v;
	  leftRate = initialRate - k * lower_spacing;
	  k++;	  
	}      
   
      k = 1;
      while(((v = calcPartialLikelihood(tr, i, initialRate + k * upper_spacing)) > rightLH) &&
	    (fabs(rightLH - v) > epsilon))
	{
	  rightLH = v;
	  rightRate = initialRate + k * upper_spacing;	 
	  k++;
	}           

      if(rightLH > initialLikelihood || leftLH > initialLikelihood)
	{
	  if(rightLH > leftLH)	    
	    {	     
	      tr->cdta->patrat[i] = rightRate;
	      lhs[i] = rightLH;
	    }
	  else
	    {	      
	      tr->cdta->patrat[i] = leftRate;
	      lhs[i] = leftLH;
	    }
	}
      else
	lhs[i] = initialLikelihood;

      tr->cdta->patratStored[i] = tr->cdta->patrat[i];     
      
    }
     

  {     
    rateCategorize *rc = (rateCategorize *)malloc(sizeof(rateCategorize) * tr->cdta->endsite);
    int where;
    int found = 0;
    for (i = 0; i < tr->cdta->endsite; i++)
      {
	rc[i].accumulatedSiteLikelihood = 0;
	rc[i].rate = 0;
      }
      
    where = 1;   
    rc[0].accumulatedSiteLikelihood = lhs[0];/*calcPartialLikelihood(tr, 0, tr->cdta->patrat[0]);*/
    rc[0].rate = tr->cdta->patrat[0];
    tr->cdta->rateCategory[0] = 0;
    
    for (i = 1; i < tr->cdta->endsite; i++) 
      {
	temp = tr->cdta->patrat[i];
	found = 0;
	for(k = 0; k < where; k++)
	  {
	    if(temp == rc[k].rate || (fabs(temp - rc[k].rate) < 0.001))
	      {
		found = 1;						
		rc[k].accumulatedSiteLikelihood += lhs[i];	
		break;
	      }
	  }
	if(!found)
	  {	    
	    rc[where].rate = temp;	    
	    rc[where].accumulatedSiteLikelihood += lhs[i];	    
	    where++;
	  }
	}

    qsort(rc, where, sizeof(rateCategorize), catCompare);

    /*{
      int k;

      for(k = 0; k < where; k++)
	printf("%d %f\n", k, rc[k].rate, rc[k].accumulatedSiteLikelihood);
      
    }*/



    if(where < maxCategories)
      {
	tr->NumberOfCategories = where;
	categorize(tr, rc);
      }
    else
      {
	tr->NumberOfCategories = maxCategories;	
	categorize(tr, rc);
      }
      
    free(rc);
  
    for (i = 0; i < tr->cdta->endsite; i++) 
      {
	/*printf("%d %f\n", i, tr->cdta->patrat[tr->cdta->rateCategory[i]]);*/
	temp = tr->cdta->patrat[tr->cdta->rateCategory[i]];

	tr->cdta->wr[i]  = wtemp = temp * tr->cdta->aliaswgt[i];
	tr->cdta->wr2[i] = temp * wtemp;
      }     

    initrav(tr, tr->start);    
    initrav(tr, tr->start->back);   
    evaluate(tr, tr->start);   

    if(tr->likelihood < initialLH)
      {
	/*printf("No improvement->restoring %f < %f\n", tr->likelihood, initialLH);*/
	tr->NumberOfCategories = oldNumber;
	for (i = 0; i < tr->cdta->endsite; i++)
	    {
	      tr->cdta->patratStored[i] = ratStored[i]; 
	      tr->cdta->rateCategory[i] = oldCategory[i];
	      tr->cdta->patrat[i] = oldRat[i];	    
	      tr->cdta->wr[i]  = oldwr[i];
	      tr->cdta->wr2[i] = oldwr2[i];
	    }       

	initrav(tr, tr->start);
	initrav(tr, tr->start->back);
        evaluate(tr, tr->start);

      }
    }
  
  free(oldCategory);
  free(oldRat);
  free(ratStored);
  free(oldwr);
  free(oldwr2); 
  free(lhs);

  /*  printf("OPT-Time: %f CATS %d LH %f INITIME %f\n", gettime() - optt, tr->NumberOfCategories, tr->likelihood, accT); */
}
  








int optimizeModel (tree *tr, analdef *adef, int finalOptimization)
{
  double startLH = tr->likelihood;
  double initialLH;
  int oldInv;

  switch(adef->model)
    {    
    case M_GTRCAT:
       {	 
	 oldInv = optimizeRatesInvocations;
	 do
	   {	    
	     initialLH = tr->likelihood;
	     optimizeRates(tr, adef);       
	   }
	 while(tr->likelihood > initialLH && optimizeRatesInvocations < oldInv + 10);  	     

	 oldInv = optimizeRateCategoryInvocations;	 
	 optimizeRateCategories(tr, finalOptimization, adef->categories);       	
       }                    
      break;    
    case M_GTRGAMMA:
       {
	 oldInv = optimizeRatesInvocations;

	 do
	   {	    
	     initialLH = tr->likelihood;
	     optimizeRates(tr, adef);       
	   }
	 while(tr->likelihood > initialLH && optimizeRatesInvocations < oldInv + 10);

	 oldInv = optimizeAlphaInvocations;
	
	 do
	   {  
	     initialLH = tr->likelihood;
	     optimizeAlpha(tr);
	   }
	   while(tr->likelihood > initialLH && optimizeAlphaInvocations < oldInv + 10);

	
       }
      break;             
    }

  if(optimizeRatesInvocations > 90)
    optimizeRatesInvocations = 90;
  if(optimizeTTinvocations > 90)
    optimizeTTinvocations = 90;
  if(optimizeRateCategoryInvocations > 90)
    optimizeRateCategoryInvocations = 90;
  if(optimizeAlphaInvocations > 90)
    optimizeAlphaInvocations = 90;

  if(startLH > tr->likelihood) return 0;
  else return 1;
}

void consensusTree(bestlist *consensusList, tree *tr)
{
  
  FILE *temporary;  
  char consensusFileName[1024], str[16], infileName[1024],
    consensusCommand[1024];
  int i;
  bestlist *bestT;

  strcpy(infileName,workdir);
  strcat(infileName,"RAxML_TreesForConsensus.");
  strcat(infileName,run_id);
  
  strcpy(consensusFileName,workdir);
  strcat(consensusFileName,"RAxML_consensus.");
  strcat(consensusFileName,run_id);
  strcat(consensusFileName,".");
  sprintf(str, "%d", consensusTreeCounter);
  strcat(consensusFileName, str);
  printf("Consensus Tree written to RAxML_consensus.%s.%s\n",
	  run_id, str);

  
  strcpy(consensusCommand,"./consense ");
  strcat(consensusCommand, infileName);
  strcat(consensusCommand," "); 
  strcat(consensusCommand, consensusFileName);

  temporary = fopen(infileName, "w");

  bestT = (bestlist *) malloc(sizeof(bestlist));
  bestT->ninit = 0;
  initBestTree(bestT, 1, tr->mxtips, tr->cdta->endsite);
  saveBestTree(bestT, tr);
  
  infoFile = fopen(infoFileName, "a");

  for(i = 1; i <= consensusList->nvalid; i++)
    {
      recallBestTree(consensusList, i, tr);
      if(i == 1)
	fprintf(infoFile, "CONSENSUS(%d): best LH %f ",consensusTreeCounter, tr->likelihood);
      if(i == consensusList->nvalid)
	fprintf(infoFile, " <-> worst LH %f out of %d trees\n",tr->likelihood, consensusList->nvalid);
      tr->start = tr->nodep[1];
      printTree(temporary, tr, 0);      
    }

  fclose(infoFile);

  fflush(temporary);
  fclose(temporary);

  system(consensusCommand);

  recallBestTree(bestT, 1, tr);
  freeBestTree(bestT);
  consensusTreeCounter++;
}









void treeEvaluation(tree *tr, analdef *adef)
{
  int i;
  
  printResult(tr, 0);

  for(i = 0; i < 10; i++)
    {      
      optimizeModel(tr, adef, 1);      
      if(i < 9)
	treeEvaluate(tr, 1);
      else
	treeEvaluate(tr, 4);           
      
      printf("%f %f\n",  gettime() - masterTime, tr->likelihood);
    } 
  printLog(tr, adef);

  printf("Time %f\n", gettime() -masterTime);
  exit(1);
}





double LnGamma (double alpha)
{
/* returns ln(gamma(alpha)) for alpha>0, accurate to 10 decimal places.  
   Stirling's formula is used for the central polynomial part of the procedure.
   Pike MC & Hill ID (1966) Algorithm 291: Logarithm of the gamma function.
   Communications of the Association for Computing Machinery, 9:684
*/
  double x=alpha, f=0, z, result;

   if (x<7) 
     {
       f=1;  
       z=x-1;
       while (++z<7)  f*=z;
       x=z;   
       f=-log(f);
     }

   z = 1/(x*x);
   
   result = f + (x-0.5)*log(x) - x + .918938533204673 
	  + (((-.000595238095238*z+.000793650793651)*z-.002777777777778)*z
	       +.083333333333333)/x;  

   return result;
}


double IncompleteGamma (double x, double alpha, double ln_gamma_alpha)
{
/* returns the incomplete gamma ratio I(x,alpha) where x is the upper 
	   limit of the integration and alpha is the shape parameter.
   returns (-1) if in error
   ln_gamma_alpha = ln(Gamma(alpha)), is almost redundant.
   (1) series expansion     if (alpha>x || x<=1)
   (2) continued fraction   otherwise
   RATNEST FORTRAN by
   Bhattacharjee GP (1970) The incomplete gamma integral.  Applied Statistics,
   19: 285-287 (AS32)
*/
   int i;
   double p=alpha, g=ln_gamma_alpha;
   double accurate=1e-8, overflow=1e30;
   double factor, gin=0, rn=0, a=0,b=0,an=0,dif=0, term=0, pn[6];

   if (x==0) return (0);
   if (x<0 || p<=0) return (-1);

   factor=exp(p*log(x)-x-g);   
   if (x>1 && x>=p) goto l30;
   /* (1) series expansion */
   gin=1;  term=1;  rn=p;
 l20:
   rn++;
   term*=x/rn;   gin+=term;

   if (term > accurate) goto l20;
   gin*=factor/p;
   goto l50;
 l30:
   /* (2) continued fraction */
   a=1-p;   b=a+x+1;  term=0;
   pn[0]=1;  pn[1]=x;  pn[2]=x+1;  pn[3]=x*b;
   gin=pn[2]/pn[3];
 l32:
   a++;  b+=2;  term++;   an=a*term;
   for (i=0; i<2; i++) pn[i+4]=b*pn[i+2]-an*pn[i];
   if (pn[5] == 0) goto l35;
   rn=pn[4]/pn[5];   dif=fabs(gin-rn);
   if (dif>accurate) goto l34;
   if (dif<=accurate*rn) goto l42;
 l34:
   gin=rn;
 l35:
   for (i=0; i<4; i++) pn[i]=pn[i+2];
   if (fabs(pn[4]) < overflow) goto l32;
   for (i=0; i<4; i++) pn[i]/=overflow;
   goto l32;
 l42:
   gin=1-factor*gin;

 l50:
   return (gin);
}

double PointNormal (double prob)
{
/* returns z so that Prob{x<z}=prob where x ~ N(0,1) and (1e-12)<prob<1-(1e-12)
   returns (-9999) if in error
   Odeh RE & Evans JO (1974) The percentage points of the normal distribution.
   Applied Statistics 22: 96-97 (AS70)

   Newer methods:
     Wichura MJ (1988) Algorithm AS 241: the percentage points of the
       normal distribution.  37: 477-484.
     Beasley JD & Springer SG  (1977).  Algorithm AS 111: the percentage 
       points of the normal distribution.  26: 118-121.

*/
   double a0=-.322232431088, a1=-1, a2=-.342242088547, a3=-.0204231210245;
   double a4=-.453642210148e-4, b0=.0993484626060, b1=.588581570495;
   double b2=.531103462366, b3=.103537752850, b4=.0038560700634;
   double y, z=0, p=prob, p1;

   p1 = (p<0.5 ? p : 1-p);
   if (p1<1e-20) return (-9999);

   y = sqrt (log(1/(p1*p1)));   
   z = y + ((((y*a4+a3)*y+a2)*y+a1)*y+a0) / ((((y*b4+b3)*y+b2)*y+b1)*y+b0);
   return (p<0.5 ? -z : z);
}


double PointChi2 (double prob, double v)
{
/* returns z so that Prob{x<z}=prob where x is Chi2 distributed with df=v
   returns -1 if in error.   0.000002<prob<0.999998
   RATNEST FORTRAN by
       Best DJ & Roberts DE (1975) The percentage points of the 
       Chi2 distribution.  Applied Statistics 24: 385-388.  (AS91)
   Converted into C by Ziheng Yang, Oct. 1993.
*/
   double e=.5e-6, aa=.6931471805, p=prob, g;
   double xx, c, ch, a=0,q=0,p1=0,p2=0,t=0,x=0,b=0,s1,s2,s3,s4,s5,s6;

   if (p<.000002 || p>.999998 || v<=0) return (-1);

   g = LnGamma(v/2);

   xx=v/2;   c=xx-1;
   if (v >= -1.24*log(p)) goto l1;

   ch=pow((p*xx*exp(g+xx*aa)), 1/xx);
   if (ch-e<0) return (ch);
   goto l4;
l1:
   if (v>.32) goto l3;
   ch=0.4;   a=log(1-p);
l2:
   q=ch;  p1=1+ch*(4.67+ch);  p2=ch*(6.73+ch*(6.66+ch));
   t=-0.5+(4.67+2*ch)/p1 - (6.73+ch*(13.32+3*ch))/p2;
   ch-=(1-exp(a+g+.5*ch+c*aa)*p2/p1)/t;
   if (fabs(q/ch-1)-.01 <= 0) goto l4;
   else                       goto l2;
  
l3: 
   x=PointNormal (p);
   p1=0.222222/v;   ch=v*pow((x*sqrt(p1)+1-p1), 3.0);
   if (ch>2.2*v+6)  ch=-2*(log(1-p)-c*log(.5*ch)+g);
l4:
   q=ch;   p1=.5*ch;
   if ((t=IncompleteGamma (p1, xx, g))<0) {
      printf ("\nerr IncompleteGamma");
      return (-1);
   }
   p2=p-t;
   t=p2*exp(xx*aa+g+p1-c*log(ch));   
   b=t/ch;  a=0.5*t-b*c;

   s1=(210+a*(140+a*(105+a*(84+a*(70+60*a))))) / 420;
   s2=(420+a*(735+a*(966+a*(1141+1278*a))))/2520;
   s3=(210+a*(462+a*(707+932*a)))/2520;
   s4=(252+a*(672+1182*a)+c*(294+a*(889+1740*a)))/5040;
   s5=(84+264*a+c*(175+606*a))/2520;
   s6=(120+c*(346+127*c))/5040;
   ch+=t*(1+0.5*t*s1-b*c*(s1-b*(s2-b*(s3-b*(s4-b*(s5-b*s6))))));
   if (fabs(q/ch-1) > e) goto l4;

   return (ch);
}

#define PointGamma(prob,alpha,beta)  PointChi2(prob,2.0*(alpha))/(2.0*(beta))




void makeGammaCats(tree *tr)
{
  int i, K;
  double gap05, t, factor, lnga1, alfa, beta;

  K = 4;
  alfa = beta = tr->alpha;
  gap05 = 1.0 / (2.0 * K);
  factor = alfa / beta * K;
  
  if(K==1) 
     {
       printf("ERROR in makeGammaCats\n");
       exit(-1);
     }
  
   lnga1=LnGamma(alfa+1);

   for (i=0; i<K-1; i++)
     tr->gammaProbs[i]=PointGamma((i+1.0)/K, alfa, beta);

   for (i=0; i<K-1; i++)
     tr->gammaProbs[i]=IncompleteGamma(tr->gammaProbs[i]*beta, alfa+1, lnga1);

   tr->gammaRates[0] = tr->gammaProbs[0]*factor;

   tr->gammaRates[K-1] = (1-tr->gammaProbs[K-2])*factor;

   for (i=1; i<K-1; i++)  
     tr->gammaRates[i] = (tr->gammaProbs[i]-tr->gammaProbs[i-1])*factor;

   for (i=0; i<K; i++) tr->gammaProbs[i]=1.0/K;

   /*for(i = 0; i < K; i++)
     printf("%d rate %f prob %f\n", i,  tr->gammaRates[i], tr->gammaProbs[i]);*/
   
   return;  
}

void initModel(tree *tr, rawdata *rdta, cruncheddata *cdta, analdef *adef)
{
  rdta->ttratio =         2.0;
  tr->t_value = 2.0;
  switch(adef->model)
    {       
    case M_GTRGAMMA:
      newview = newviewGTRGAMMA;
      makenewz = makenewzGTRGAMMA;
      evaluate = evaluateGTRGAMMA;      
      baseFrequenciesGTR(rdta, cdta);       
      rdta->initialRates[0] = 0.5;
      rdta->initialRates[1] =  0.5;
      rdta->initialRates[2] = 0.5;
      rdta->initialRates[3] = 0.5;
      rdta->initialRates[4] = 0.5;      
      initReversibleGTR(rdta, cdta, tr);
      tr->alpha = 1.0;                  
      makeGammaCats(tr);      
      break;
    case M_GTRCAT:
      newview = newviewGTRCAT;
      makenewz = makenewzGTRCAT;
      evaluate = evaluateGTRCAT;
      evaluatePartial = evaluatePartialGTRCAT;
      newviewPartial =  newviewPartialGTRCAT;    
      baseFrequenciesGTR(rdta, cdta);       
      tr->NumberOfCategories = 1;
      rdta->initialRates[0] = 0.5;
      rdta->initialRates[1] =  0.5;
      rdta->initialRates[2] = 0.5;
      rdta->initialRates[3] = 0.5;
      rdta->initialRates[4] = 0.5;      
      initReversibleGTR(rdta, cdta, tr);            
      break;
    }
}


void initModel2(tree *tr, rawdata *rdta, cruncheddata *cdta, analdef *adef)
{
 
      newview = newviewGTRGAMMA;
      makenewz = makenewzGTRGAMMA;
      evaluate = evaluateGTRGAMMA;                
      /*rdta->initialRates[0] = 0.5;
      rdta->initialRates[1] =  0.5;
      rdta->initialRates[2] = 0.5;
      rdta->initialRates[3] = 0.5;
      rdta->initialRates[4] = 0.5;      */
      initReversibleGTR(rdta, cdta, tr);
      tr->alpha = 1.0;                  
      makeGammaCats(tr);      
      
}


#ifdef PARALLEL_2

void initModelPack(tree * tr, analdef * adef)
{
  switch (adef->model)
    {   
    case M_GTRCAT:
      tr->modelParameters = 9 + 5 * tr->cdta->endsite;
      tr->modelData = (double *) malloc(sizeof(double) * tr->modelParameters);
      break;
    case M_GTRGAMMA: /*TODO*/break;   
  }
}

void packCategories(tree * tr, double *m)
{
  int i;
  double *m_ptr = m;

  for (i = 0; i < tr->cdta->endsite; i++)
    *m_ptr++ = tr->cdta->rateCategory[i];

  for (i = 0; i < tr->cdta->endsite; i++)
    *m_ptr++ = tr->cdta->patratStored[i];

  for (i = 0; i < tr->cdta->endsite; i++)
    *m_ptr++ = tr->cdta->patrat[i];

  for (i = 0; i < tr->cdta->endsite; i++)
    *m_ptr++ = tr->cdta->wr[i];

  for (i = 0; i < tr->cdta->endsite; i++)
    *m_ptr++ = tr->cdta->wr2[i];

}

void unpackCategories(tree * tr, double *m)
{
  int i;
  double *m_ptr = m;

  for (i = 0; i < tr->cdta->endsite; i++)
    tr->cdta->rateCategory[i] = *m_ptr++;

  for (i = 0; i < tr->cdta->endsite; i++)
    tr->cdta->patratStored[i] = *m_ptr++;

  for (i = 0; i < tr->cdta->endsite; i++)
    tr->cdta->patrat[i] = *m_ptr++;

  for (i = 0; i < tr->cdta->endsite; i++)
    tr->cdta->wr[i] = *m_ptr++;

  for (i = 0; i < tr->cdta->endsite; i++)
    tr->cdta->wr2[i] = *m_ptr++;

}

void packModel(tree * tr, analdef * adef)
{
  int i;
  printf("pack model LH %f\n", tr->likelihood);

  tr->modelData[0] = (double) optimizeTTinvocations;
  tr->modelData[1] = (double) optimizeRatesInvocations;
  tr->modelData[2] = (double) optimizeRateCategoryInvocations;

  switch (adef->model)
  {    
    case M_GTRCAT:
      for (i = 3; i < 8; i++)
        tr->modelData[i] = tr->rdta->initialRates[i - 3];
      tr->modelData[8] = (double) tr->NumberOfCategories;
      packCategories(tr, &tr->modelData[9]);
      break;
  case M_GTRGAMMA: /*TODO */ break;
  }
}

void unpackModel(tree * tr, analdef * adef)
{
  int i;

  optimizeTTinvocations = (int) tr->modelData[0];
  optimizeRatesInvocations = (int) tr->modelData[1];
  optimizeRateCategoryInvocations = (int) tr->modelData[2];

  switch (adef->model)
    {   
    case M_GTRCAT:
      for (i = 3; i < 8; i++)
        tr->rdta->initialRates[i - 3] = tr->modelData[i];
      tr->NumberOfCategories = (int) tr->modelData[8];
      initReversibleGTR(tr->rdta, tr->cdta, tr);
      unpackCategories(tr, &tr->modelData[9]);
      break;
    case M_GTRGAMMA:
      /*TODO*/break;
  }
  initrav(tr, tr->start);
  initrav(tr, tr->start->back);
  evaluate(tr, tr->start);
  printf("unpack model LH %f\n", tr->likelihood);
}

#endif

int filexists(char *filename)
{
  FILE *fp;
  int res;
  fp = fopen(filename,"r");
  


  if(fp) 
    {
      res = 1;
      fclose(fp);
    }
  else 
    res = 0;
       
  return res;
} 







void initGlobalTip(tree *tr, analdef *adef)
{  
  /* TODO simplify */

  if(adef->model == M_GTRCAT)
    {     
      int code, i, j, jj;
      double (*EIGV)[4] = &(tr->rdta->EIGV[0][0]);
      for(i=0; i < 16; i++)
	{
	  code = i;
	  tr->gtrTip[i].a = 0;
	  tr->gtrTip[i].c = 0;
	  tr->gtrTip[i].g = 0;
	  tr->gtrTip[i].t = 0;
	  for (j = 0; j < 4; j++) 
	    {
	      if ((code >> j) & 1) 
		{
		  int jj = "0213"[j] - '0';		
		  tr->gtrTip[i].a += EIGV[jj][0];
		  tr->gtrTip[i].c += EIGV[jj][1];
		  tr->gtrTip[i].g += EIGV[jj][2];
		  tr->gtrTip[i].t += EIGV[jj][3];
		}
	    }		
	}      
    }   
  if(adef->model == M_GTRGAMMA)
    {     
      int code, i, j, jj;
      double (*EIGV)[4] = &(tr->rdta->EIGV[0][0]);
      for(i=0; i < 16; i++)
	{
	  code = i;
	  tr->GTRGammaTip[i].a = 0;
	  tr->GTRGammaTip[i].c = 0;
	  tr->GTRGammaTip[i].g = 0;
	  tr->GTRGammaTip[i].t = 0;
	  for (j = 0; j < 4; j++) 
	    {
	      if ((code >> j) & 1) 
		{
		  int jj = "0213"[j] - '0';		
		  tr->GTRGammaTip[i].a += EIGV[jj][0];
		  tr->GTRGammaTip[i].c += EIGV[jj][1];
		  tr->GTRGammaTip[i].g += EIGV[jj][2];
		  tr->GTRGammaTip[i].t += EIGV[jj][3];
		}
	    }		
	}      
    }   
}



main (int argc, char *argv[]) 
  {   
    rawdata      *rdta;
    cruncheddata *cdta;
    tree         *tr;         
    analdef      *adef;
    masterTime = gettime();

#ifdef PARALLEL
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &myId);    
    MPI_Comm_size(MPI_COMM_WORLD, &numOfWorkers);
#endif          

    adef = (analdef *)malloc(sizeof(analdef));
    rdta = (rawdata *)malloc(sizeof(rawdata));
    cdta = (cruncheddata *)malloc(sizeof(cruncheddata));     
    tr   = (tree *)malloc(sizeof(tree));

#ifdef PROCESSOR_AFFINITY
    {     
      if (numa_available() >= 0)
	{	 
#pragma omp parallel
	  {
	    numa_run_on_node(omp_get_thread_num());
	  }
	} 
      else
	{
	  printf("NUMA policies not available \n");
	  printf("Either install libnuma or remove flag -DPROCESSOR_AFFINITY from  Makefile.pgi\n");
	  printf("Please read the comments about processor affinity in the RAxML manual\n");
	  printf("Exiting ....\n");
	  exit(1);
	}
    }
#endif

    initAdef(adef);
    get_args(argc,argv,(boolean)1, adef, rdta);              
   	    	         
    INFILE = fopen(seq_file, "r");
    if (!INFILE)
      {
	printf( "Could not open sequence file: %s\n", seq_file);
	return 1;
      }   
 
    getinput(adef, rdta, cdta, tr);   
         
    fclose(INFILE);   


    if(adef->boot && adef->multipleRuns)
      {
	doBootstrap(tr, adef, rdta, cdta, argc, argv);
	return 1;
      }
    else
      {
	if(adef->multipleRuns && !adef->boot)
	  {
	    doMultiple(tr, adef, rdta, cdta, argc, argv);
	    return 1;
	  }
      }
	
    optimizeTTinvocations = 1;
    optimizeRatesInvocations = 1;
    optimizeRateCategoryInvocations = 1;
    optimizeAlphaInvocations = 1;
                
    if (! makeweights(adef, rdta, cdta))                             return 1;        
    if (! makevalues(rdta, cdta))                                    return 1;   
    linkdata2tree(rdta, cdta, tr);    
    initModel(tr, rdta, cdta, adef);                                 
    
    treeStringLength = (tr->mxtips * (nmlngth+32)) + 256 + (tr->mxtips * 2);       
	  
    bt = (bestlist *) malloc(sizeof(bestlist));
    bt->ninit = 0;
    initBestTree(bt, 20, tr->mxtips, tr->cdta->endsite);
    
    strcpy(permFileName, workdir);    
    strcpy(resultFileName, workdir);
    strcpy(logFileName, workdir);
    strcpy(checkpointFileName, workdir);
    strcpy(infoFileName, workdir);
    
    strcat(permFileName, "RAxML_parsimonyTree.");
    strcat(resultFileName, "RAxML_result.");
    strcat(logFileName, "RAxML_log.");
    strcat(checkpointFileName, "RAxML_checkpoint.");
    strcat(infoFileName, "RAxML_info.");
    
    strcat(infoFileName, run_id);
    strcat(permFileName, run_id);
    strcat(resultFileName, run_id);
    strcat(logFileName, run_id);
    
    strcat(checkpointFileName, run_id);
    strcat(checkpointFileName,".");
        
    if(filexists(infoFileName))
      {
	printf("RAxML output files with the run ID <%s> already exist \n", run_id);
	printf("in directory %s ...... exiting\n", workdir);
	exit(-1);
      }
    
    logFile = fopen(logFileName, "w");
    
    printModelAndProgramInfo(tr, adef, argc, argv);      	
             
    initGlobalTip(tr, adef);
     
    if (adef->restart) 
      {	 	
	allocNodex(tr, adef); 
	
	INFILE = fopen(tree_file, "r");	
	if (!INFILE)
	  {
	    printf( "Could not open input tree: %s\n", tree_file);
	    return FALSE;
	  }              
	
	if (! treeReadLen(INFILE, tr))          return FALSE;	       	
	if(adef->mode != TREE_EVALUATION)
	  treeEvaluate(tr, 1); 

	fclose(INFILE);
      }
    else
      {       
	makeParsimonyTree(tr, adef);
	fprintf(infoFile, "Time required to compute MP tree: %f\n", gettime() - masterTime);
	permutationFile = fopen(permFileName,"w");
	printTree(permutationFile, tr, 0);
	fclose(permutationFile);
	allocNodex(tr, adef); 	
	initrav(tr, tr->start);
	initrav(tr, tr->start->back);       	
	treeEvaluate(tr, 1); 
      }         
   
    if(adef->startingTreeOnly)
      exit(1);           

    switch(adef->mode)
      {          
      case TREE_EVALUATION         :      
	modOpt(tr, adef);
	printLog(tr, adef);
	printf("Final GTR+GAMMA  likelihood: %f\n", tr->likelihood);
	printf("Final alpha shape parameter: %f\n", tr->alpha);
	infoFile = fopen(infoFileName, "a");
	fprintf(infoFile, "Final GTR+GAMMA  likelihood: %f\n", tr->likelihood);
	fprintf(infoFile, "Final alpha shape parameter: %f\n", tr->alpha);			
	fclose(infoFile);
	break;
      case BIG_RAPID_MODE          : computeBIGRAPID(tr, adef); break;	    
      case BIG_SIMULATED_ANNEALING : simulatedAnnealingRapid(tr, adef);break;
      }

    if(adef->model == M_GTRGAMMA && adef->mode != TREE_EVALUATION)
      {
	printf("Final Model Optimization for GTR+GAMMA\n");

	modOpt(tr, adef);
	printLog(tr, adef);
	printf("Final GTR+GAMMA  likelihood: %f\n", tr->likelihood);
	printf("Final alpha shape parameter: %f\n", tr->alpha);
	infoFile = fopen(infoFileName, "a");
	fprintf(infoFile, "Final GTR+GAMMA  likelihood: %f\n", tr->likelihood);
	fprintf(infoFile, "Final alpha shape parameter: %f\n", tr->alpha);			
	fclose(infoFile);
	adef->mode = BIG_RAPID_MODE;
      }   

    if(adef->model == M_GTRCAT && adef->useMixedModel == 1 && adef->mode != TREE_EVALUATION) 
      {
	int i;	       

	freeNodex(tr);
	adef->model = M_GTRGAMMA;
	allocNodex(tr, adef); 
	initModel2(tr, rdta, cdta, adef);
	optimizeRatesInvocations = 1;  
	optimizeTTinvocations = 1;  
	optimizeRateCategoryInvocations = 1;  
	optimizeAlphaInvocations = 1;
	initGlobalTip(tr, adef);
	
	for(i = 0; i < tr->cdta->endsite; i++)
	  tr->cdta->wr[i] = tr->cdta->aliaswgt[i];

	modOpt(tr, adef);
	printLog(tr, adef);
	printf("Final GTR+GAMMA  likelihood: %f\n", tr->likelihood);
	printf("Final alpha shape parameter: %f\n", tr->alpha);
	infoFile = fopen(infoFileName, "a");
	fprintf(infoFile, "Final GTR+GAMMA  likelihood: %f\n", tr->likelihood);
	fprintf(infoFile, "Final alpha shape parameter: %f\n", tr->alpha);			
	fclose(infoFile);
	adef->mode = BIG_RAPID_MODE;
      }


    
    fclose(logFile);

    printf("\n\n");

    if(adef->checkpoints)
      {	
	if(checkPointCounter > 1)
	  printf("Checkpoints written to:                 %s0-%d\n", checkpointFileName, (checkPointCounter - 1));
	else
	  printf("checkpoint written to:                  %s0\n", checkpointFileName); 
      }
    
    if((!adef->restart) && (adef->mode != TREE_EVALUATION))
          printf("Parsimony starting tree written to:     %s\n", permFileName);   	          
    if(adef->mode != TREE_EVALUATION)
          printf("Final tree written to:                  %s\n", resultFileName);  
    printf("Execution LogFile written to:           %s\n", logFileName);   
    printf("Execution information file written to:  %s\n", infoFileName);
  }
